#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

#define LEFT	1
#define NEAR	2
#define RESIST  3	/* XXX */
#define RIGHT	4
#define FAR	5
#define UPPER	6
#define LOWER	7

#define F_LEFT	0x01
#define F_RIGHT	0x02
#define F_NEAR	0x04
#define F_FAR	0x08

#define X_CENTER  1.00
#define XL_MARGIN 0.20
#define XR_MARGIN 1.80
#define X_LEN (XL_MARGIN+XR_MARGIN)

#define Y_CENTER  1.00
#define YL_MARGIN 0.20
#define YR_MARGIN 1.80
#define Y_LEN (YL_MARGIN+YR_MARGIN)

#define H_SIZE	(0.30)
#define H_CORONA (H_SIZE + YL_MARGIN)
#define H_LOCUS 1.00
#define H_LOCUS1 1.25

#define SLICE 0.0400001

#define RL_HOLE		0x00
#define RL_LNEAR	0x01
#define RL_RNEAR	0x02
#define RL_FAR		0x03

#define N_OF_RESIST	1.65

int	lightsrcno;
int	lightnodeno;

int	whichsubdo(const struct ement_t *);
int	whichbound(const struct node_t *);
void	make_femmatrix(void);
struct duplex_t expon(const double x);
int	bonlocus(double x);
int	cksubdo(const double x, const double y);

void	impose_bcs(void);
void	set_const3rd(void);


struct vector_t *mux;
struct vector_t *muy;
struct vector_t *epsilon;

double	wave_k;
double	lambda;

struct rule {
	double	rl_min, rl_max;
	int	rl_flag;
	struct rule *rl_next;
};
void		 rl_print(void);
void		 mkrule(const double x, const double radius);
struct rule	*rl_findlast(void);
struct rule	*rl_alloc(void);
struct rule	*rl_findme(const double x);
int		 rl_getsubdo(const double x, const double y);


struct rule *rules;

void
problem(void)
{
	lambda = 0.365;
	wave_k = 2.0 * M_PI/lambda;

	mk_nearements();
	mkrule(1.00, 0.100);
#if 0
	mkrule(1.00, 0.24999);
	mkrule(1.50, 0.1999);
	mkrule(1.30, 0.15);
#endif
	set_subdomain(whichsubdo);
	set_boundary(whichbound);
	put_2dgrids("grids.mtv", 0);
	rl_print();

	set_const3rd();
	put_gridandvector(mux, "mux");
	put_gridandvector(muy, "muy");
	
	vec_null(meshes->m_bvector);
	sp_null(meshes->m_kmatrix);

	make_femmatrix();
	impose_bcs();
	sp_symcheck(meshes->m_kmatrix);
	solveall();

	put_gridandvector(meshes->m_bvector, "a.mtv");

	return;
}

int
whichsubdo(const struct ement_t *ep)
{
	struct proto_t *proto;
	struct node_t *np;
	nodno_t *nnp;
	int	noofnodes;
	float	means[3];
	int	i, j, flag, ret;

	proto = ep->e_proto;
	nnp = ep->e_nodes;
	noofnodes = ep->e_nodesno;

	for (i=0;i<meshes->m_spacedim;i++)
		means[i] = 0.0; /* initialization */

	for (i=0;i<proto->p_nodsno;i++) {
		np = meshes->m_nodes[nnp[i]]; /* for surrounding nodes*/
		for (j=0; j < meshes->m_spacedim; j++)
			means[j] += np->n_coor[j];
	}

	for (i=0;i < meshes->m_spacedim;i++)
		means[i] /= (float) noofnodes; /* initialization */

	if (fabs(means[1] - Y_CENTER) < SLICE) {
		ret = rl_getsubdo(means[0], means[1]);
		if (ret)
			return ret;
	}

	flag = 0;
	if (means[0] <= XL_MARGIN)
		flag |= F_LEFT;
	if (means[0] >= XR_MARGIN)
		flag |= F_RIGHT;

	if (means[1] <= YL_MARGIN)
		flag |= F_NEAR;
	if (means[1] >= YR_MARGIN)
		flag |= F_FAR;

	switch (flag) {
		case F_LEFT:
			if (means[1] < 1.0)
				return LEFT;
			break;
		case F_RIGHT:
			if (means[1] < 1.0)
				return RIGHT;
			break;
		case F_NEAR :
			return NEAR ;
			break;
		case F_FAR:
			return FAR;
			break;
		case F_LEFT | F_NEAR:
			if (means[0] < means[1])
				return LEFT;
			else
				return NEAR;
			break;
		case F_RIGHT | F_NEAR:
			if ((X_LEN-means[0]) < means[1])
				return RIGHT;
			else
				return NEAR;
			break;
		case F_LEFT | F_FAR:
			return FAR;
			if (means[0] > (Y_LEN-means[1]))
				return FAR;
			else
				return LEFT;
			break;
		case F_RIGHT | F_FAR:
			return FAR;
			if ((X_LEN-means[0]) > (Y_LEN-means[1]))
				return FAR;
			else
				return RIGHT;
			break;
	}
	if (means[1] < Y_CENTER)
		return RESIST;

	return 0;
}

int
whichbound(const struct node_t *np)
{
	float	x, y;

	x = np->n_coor[0];
	y = np->n_coor[1];

	if (fabs(x - 2.0) < 0.0001 && (y  <= 1.00)) /* right lower */
		return 4;
	if (fabs(x) < 0.0001 && (y  <= 1.00)) /* left lower */
		return 4;
	return -1;
}

void
make_femmatrix(void)
{
	int     i;
	struct matrix_t    *retmatf, *bmatf;
	struct proto_t  *prevproto = NULL;

	currp = NULL;
	for (i=0; i <  meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			 create_smalls(currp, &retmatf, &bmatf);
		}
		integration(i, retmatf, bmatf);
		bassemble(i, bmatf, meshes->m_bvector);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		prevproto = currp;
	}
	fprintf(stderr, "no of the light source=%d\n", lightsrcno);
	if (lightsrcno == 0)
		fatal("%s(): no of the light source=%d", __F, lightsrcno);
	m_mfree(2, retmatf, bmatf);
	return;
}

void
integration(const int elemno, struct matrix_t *kmatf, struct matrix_t *bmatf)
{
	struct ement_t *ep;
	int	i, j, k;
	int	subdomain;
	struct duplex_t **retmat, **bmat;
	struct integral_t *intg;
	double	**unit, **deriv;
	struct duplex_t	phase;
	double	*jacob, *weight;
	double	rightside;

	struct duplex_t axx, ayy, beta;
	struct duplex_t dxtmp, dx1, dx2, dx3;
	float	x, y;
	float	deltax;
	int	gotit;

	m_mnull(2, kmatf, bmatf); /* XXX */
	ep = meshes->m_ements[elemno];
	intg   = currp->p_intg;
	proto_einit(elemno);
	jacob  = intg->intg_jacob;
	weight = intg->intg_weights;

	unit   = dmat(intg->intg_trans);

	retmat = zmat(kmatf);
	bmat   = zmat(bmatf);

	subdomain = ep->e_sdomain;
	x = get_centercoor(elemno, 0);
	y = get_centercoor(elemno, 1);

	if (fabs(y-1.605) < 0.005) {
		deltax = x - X_CENTER;
		if (bonlocus(deltax)) {
			gotit = 1;
			lightsrcno++;

			phase = duplexv(0.0, +1.0);
		} else {
			gotit = 0;
			phase = duplex_null;
		}
	} else {
		gotit = 0;
		phase = duplex_null;
	}

	rightside = 1.0;
	for (k=0; k < currp->p_intgno; k++) {	/* for integration points */
		deriv = dmat(intg->intg_deriv[k]);
		dx1 = zgetat(elemno, k, muy);
		dx2 = zgetat(elemno, k, mux);
		dx3 = zgetat(elemno, k, epsilon);

		zp_inverse(&axx, &dx1);
		zp_inverse(&ayy, &dx2);
		zp_crossd(&beta, &dx3, -1.0 * wave_k * wave_k);

		for (i=0; i< currp->p_nodsno;i++) {	/* XXX */
			for (j=0; j< currp->p_nodsno;j++) {
				zp_crossd(&dx1, &axx, deriv[0][i]*deriv[0][j]);
				zp_crossd(&dx2, &ayy, deriv[1][i]*deriv[1][j]);

				zp_crossd(&dx3, &beta, unit[k][i]*unit[k][j]);

				zp_3plus(&dxtmp, &dx1, &dx2, &dx3);
				zp_crossd(&dxtmp, &dxtmp, jacob[k] * weight[k]);

				zp_pplus(&retmat[i][j], &dxtmp);
			}
			if (gotit) {
				zp_crossd(&dx2, &phase,
					1.0/ (double) currp->p_nodsno);
				zp_pplus(&bmat[i][0], &dx2);
			}
		}
	}
	return;
}


void
impose_bcs(void)
{
#if 1
	zdirichlet(0, duplexv(0.0, 0.0)); /* XXX */
	zdirichlet(2, duplexv(0.0, 0.0)); /* XXX */
	zdirichlet(4, duplexv(0.0, 0.0)); /* XXX */
#else
	zdirichlet(0, duplexv(0.0, 0.0));
	zdirichlet(1, duplexv(0.0, 0.0));
	zdirichlet(3, duplexv(0.0, 0.0));
	zdirichlet(11, duplexv(1.0, 0.0)); /* XXX */
	zdirichlet(10, duplexv(1.0, 0.0));
#endif
	return;
}

struct duplex_t
expon(const double x)
{
	double	argum;

	argum = x * wave_k;
	return duplexv(cos(argum), sin(argum));
}

int
bonlocus(double x)
{
	return 1;
	x = fabs(x);
#if 0
	if (x < 0.01)
		return 1;
	ddiv = (int) (x/lambda);
	residue = x - lambda * (double) ddiv;

	if (fabs(residue)< 0.01)
		return 1;
#else
	if (fabs(x) < (X_CENTER - XL_MARGIN-0.011))
		return 1;
#endif
	return 0;
}

/*
 * for refraction
 */
int
cksubdo(const double x, const double y)
{
	double	fx;

	fx = y - 0.40 * x - 0.40;
	if (fx < 0)
		return 1;
	return 0;
}


struct rule *
rl_findlast(void)
{
	struct rule *rl;

	rl = rules;
	while (rl->rl_next)
		rl = rl->rl_next;
	return rl;
}

struct rule *
rl_alloc(void)
{
	double	alloced;
	struct rule *ret;

	alloced = sizeof(struct rule);
	ret = (struct rule *) malloc(alloced);
	if (ret == NULL)
		memerr("%s(): ret == NULL", __F);
	memset(ret, 0x00, alloced);
	return ret;
}

void
mkrule(const double x, const double radius)
{
	struct rule *oldone, *newone[4];
	int	i;

	for (i=0; i<4;i++)
		newone[i] = rl_alloc();

	if (rules == NULL) {
		rules = rl_alloc();
		rules->rl_min = 0;
		rules->rl_max = Y_LEN;
		rules->rl_flag = RL_FAR;
		oldone = rules;
	} else {
		oldone = rl_findlast();
	}

	oldone->rl_next = newone[0];
	oldone->rl_max  = x - radius - SLICE;
	if (oldone->rl_max <= oldone->rl_min)
		fatal("%s(): oldone->rl_max <= oldone->rl_min", __F);

	newone[0]->rl_min = x - radius - SLICE;
	newone[0]->rl_max = x - radius;
	newone[0]->rl_flag= RL_LNEAR;
	newone[0]->rl_next = newone[1];

	newone[1]->rl_min = x - radius;
	newone[1]->rl_max = x + radius;
	newone[1]->rl_flag= RL_HOLE;
	newone[1]->rl_next = newone[2];

	newone[2]->rl_min = x + radius;
	newone[2]->rl_max = x + radius + SLICE;
	newone[2]->rl_flag= RL_RNEAR;
	newone[2]->rl_next = newone[3];

	newone[3]->rl_min = x + radius + SLICE;
	newone[3]->rl_max = Y_LEN;
	newone[3]->rl_flag= RL_FAR;
	newone[3]->rl_next = NULL;
	
	return;
}

void
rl_print(void)
{
	static char *rlnames[] = {"HOLE", "LEFT_SIDE", "RIGHT_SIDE", "FAR"};
	struct rule *rl;
	int	i;

	for (i = 0, rl = rules; rl; rl = rl->rl_next, i++)
		printf("++[%2d] %7.3f  -  %7.3f  :  %s\n",
			i, rl->rl_min, rl->rl_max, rlnames[rl->rl_flag]);

	return;
}

struct rule *
rl_findme(const double x)
{
	struct rule *rl;

	for (rl = rules; rl; rl = rl->rl_next)
		if ((rl->rl_min <= x) && (x < rl->rl_max))
			break;
	if (rl == NULL)
		fatal("%s(): rl == NULL and x = %7.4f", __F, x);
	return rl;
}

int
rl_getsubdo(const double x, const double y)
{
	struct rule *rl;
	int	which;
	double	ygap, xgap;

	if (fabs(y-Y_CENTER) > SLICE)
		return 0;
	rl = rl_findme(x);

	which = (y > Y_CENTER) ? UPPER : LOWER;
	ygap = fabs(y-Y_CENTER);
	xgap = x - rl->rl_min;
	
	switch (rl->rl_flag) {
		case RL_HOLE:
			return 0;	/* no subdomain */
#if 0	/* XXX */
		default:
			return which;
#else
		case RL_FAR:
			return which;
		case RL_LNEAR:
			if (ygap < (SLICE - xgap))
				return which;
			break;
		case RL_RNEAR:
			if (ygap < xgap)
				return which;
			break;
#endif
	}
	return 0;
}

void
set_const3rd(void)
{
	int	i, j, subdomain;
	struct node_t *np;
	nodno_t	*nnp;

	double	xcoor, ycoor, delta;
	double	squared, howmany;
	double	factor;
	double	slfactor;
	struct duplex_t ztmp0, ztmp1;
	struct duplex_t *muxp, *muyp, *epsp;

	printf("start of %s\n", __F);
	factor = 20.0 / (YL_MARGIN * wave_k); 	/* XXX */
	slfactor = 20.0 / (SLICE * wave_k); 	/* XXX */

	mux = vector(meshes->m_nodesno, "mux", SP_DUPLEX);
	muy = vector(meshes->m_nodesno, "muy", SP_DUPLEX);
	epsilon = vector(meshes->m_nodesno, "epsilon", SP_DUPLEX);

	muxp = zvec(mux);
	muyp = zvec(muy);
	epsp = zvec(epsilon);

	for (i=0; i < meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		xcoor = np->n_coor[0];
		ycoor = np->n_coor[1];
		nnp = np->n_ements;

		for (j=0; j < np->n_ementsno; j++) {
			subdomain = meshes->m_ements[nnp[j]]->e_sdomain;
			switch (subdomain) {
				case LEFT:
				case RIGHT:
					if (subdomain == LEFT)
						delta = fabs(xcoor-XL_MARGIN);
					else
						delta = fabs(xcoor-XR_MARGIN);

					squared = (delta/XL_MARGIN);
					squared *= squared;
					squared *= 3.0 * factor;

					ztmp0 = duplexv(N_OF_RESIST,
						-1.0 * squared);
					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp1);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case NEAR:
				case FAR:
					if (subdomain == NEAR)
						delta = fabs(ycoor-YL_MARGIN);
					else
						delta = fabs(ycoor-YR_MARGIN);

					squared = (delta/YL_MARGIN);
					squared *= squared;
					squared *= 3.0 * factor;

					if (subdomain == NEAR)
						ztmp0 = duplexv(N_OF_RESIST,
							-1.0 * squared);
					else
						ztmp0 = duplexv(1.0,
							-1.0 * squared);

					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp1);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case UPPER:
				case LOWER:
					if (subdomain == UPPER)
						delta = fabs(ycoor-(1.0+SLICE));
					else
						delta = fabs(ycoor-(1.0-SLICE));

					squared = (delta/SLICE);
					squared *= squared;
					squared *= 3.0 * slfactor; /* XXX */

					if (subdomain == UPPER)
						ztmp0 = duplexv(1.0,
							-1.0*squared);
					else
						ztmp0 = duplexv(N_OF_RESIST,
							-1.0*squared);

					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp1);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case RESIST:
					ztmp0 = duplexv(N_OF_RESIST, 0.02);
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				default:
					ztmp0 = duplexv(1.0, 0.0);
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
			}
		}
		howmany = 1.0/(double) j;

		zp_crossdme(&muxp[i], howmany);
		zp_crossdme(&muyp[i], howmany);
		zp_crossdme(&epsp[i], howmany);
	}

	printf("end of %s\n", __F);
	return;
}
