#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include "common.h"
#include "mesh.h"
#include "hash.h"

/*
 * In the case of the mesh which has a good symmetry, I want to use that
 * property in obtaining the result of this simulation.  Especially I want to
 * use that symmetry property in obtaining the result graphically.  In order to
 * use this periodicity property, I should make hash lookup tables of nodes
 * and hash lookup tables of element information, i.e., meshes->n_hashes and
 * meshes->m_ehashes respectively.
 */
#define ALLOC_UNIT	16
#define H_TOLERANCE	1.0e-15
#define NODEP_UNIT	50

/*
 * Only used at get_nodeinter() and get_ementinter();
 */
int
nodep_compare(nodno_t *first, nodno_t *second)
{
	nodno_t	one, two;

	one = *first;
	two = *second;
	
	if (one > two)
		return 1;
	else if (one < two)
		return -1;
	else
		return 0;
}

int
h_equal(float one, float two)
{
	if (fabs(one - two) < H_TOLERANCE)
		return 1;
	return 0;
}

void
mesh_analysis(void)
{
	int	i, j, k, hashval;
	struct node_t *np;
	struct ement_t *ep;
	struct hash_t *hp, *last;
	float	coor, triple[3];
	nodno_t  *nnp;

	/*
	 * Get the max and min value of the coor's of all nodes
	 */

	for (i=0;i<3;i++) {
		meshes->m_coormax[i] = -1.0e10;
		meshes->m_coormin[i] =  1.0e10;
	}

	/*
	 * Make the hash structure of node coordinate information
	 */
	for (i=0;i<meshes->m_nodeno;i++) {	/* for all nodes */
		np = meshes->m_nodes[i];

		for (j=0;j < meshes->m_spacedim; j++) {	/* for each coor */
			coor = np->n_coor[j];
			if (coor > meshes->m_coormax[j])
				meshes->m_coormax[j] = coor; 
			else if (coor < meshes->m_coormax[j])
				meshes->m_coormin[j] = coor; 

			hashval = gethashval(coor);
			hp = meshes->m_nhashes[j][hashval];

			for (; hp; hp = hp->h_next) {
				if (h_equal(hp->h_value, coor))
					break;
				last = hp;
			}

			if (hp == NULL) {
				hp = (struct hash_t *)
					malloc((size_t)sizeof(struct hash_t));
				last->h_next  = hp;
				hp->h_alloced = 0;
				hp->h_cnt     = 0;
				hp->h_nodes   = NULL;
				hp->h_next    = NULL;
				hp->h_value   = coor;
			}

			if (hp->h_alloced == hp->h_cnt) {
				hp->h_alloced += ALLOC_UNIT;
				if (hp->h_nodes)
					hp->h_nodes = (nodno_t *)
						realloc(hp->h_nodes,
						hp->h_alloced*sizeof(nodno_t));
				else
					hp->h_nodes = (nodno_t *)
						malloc(hp->h_alloced *
							sizeof(nodno_t));
			}
			hp->h_nodes[hp->h_cnt] = i;	/* insert this node */
			hp->h_cnt++;
		}
	}

	/*
	 * End of the analysis of nodes
	 * Make the hash structure of element coordinate information
	 */
	for (i=0;i<meshes->m_elemno;i++) {	/* for ements nodes */
		ep  = meshes->m_ements[i];
		nnp = ep->e_nodes;

		/* Let's get the coordinates of the center of this element */
		for (j=0;j < 3; j++)
			triple[j] = 0.0;

		for (j=0; j< ep->e_proto->p_nodes; j++) { /* for surrounds */
			for (k=0; k < meshes->m_spacedim; k++)
				triple[k] += meshes->m_nodes[nnp[j]]->n_coor[k];
		}

		for (j=0;j < meshes->m_spacedim; j++)
			triple[j] /= (float) ep->e_proto->p_nodes;

		for (j=0;j < meshes->m_spacedim; j++) {	/* for each coor */
			coor = triple[j];

			hashval = gethashval(coor);
			hp = meshes->m_ehashes[j][hashval];

			for (; hp; hp = hp->h_next) {
				if (h_equal(hp->h_value, coor))
					break;
				last = hp;
			}

			if (hp == NULL) {
				hp = (struct hash_t *)
					malloc((size_t)sizeof(struct hash_t));
				last->h_next  = hp;
				hp->h_alloced = 0;
				hp->h_cnt     = 0;
				hp->h_nodes   = NULL;
				hp->h_next    = NULL;
				hp->h_value   = coor;
			}

			if (hp->h_alloced == hp->h_cnt) {
				hp->h_alloced += ALLOC_UNIT;
				if (hp->h_nodes)
					hp->h_nodes = (nodno_t *)
						realloc(hp->h_nodes,
						hp->h_alloced*sizeof(nodno_t));
				else
					hp->h_nodes = (nodno_t *)
						malloc(hp->h_alloced *
							sizeof(nodno_t));
			}
			hp->h_nodes[hp->h_cnt] = i;	/* insert this node */
			hp->h_cnt++;
		}
	}
	/*
	 * End of the analysis of elements
	 */
	return;
}

int
gethashval(float coor)
{
	int	i;
        unsigned int val = 0, g;
        char   *p;
	char	line[32];

	sprintf(line, "%10.5f", coor);
	p = (char *) line;
        for (i = 0; i < 4; p++, i++) {
                val = (val << 4) + (*p);
                if ((g = val & 0xF0000000)) {
                        val = val ^ (g >> 24);
                        val = val ^ g;
                }
        }
        return val % HASHSIZE;
}


/*
 * Private data structure of nodes
 */

struct nodesset *
getnodes_at(int flag, float xmin, float xmax,
	float ymin, float ymax, float zmin, float zmax)
{
	int	i, j, cnt, size;
	int	coorsno[3];	/* for a given axis.. how many differen?? */
	struct hash_t *hp, **hpp;
	float	coor, limits[3][2];
	struct node_private *npriv[3], *npr, *prevnp;
	int	check[3];
	struct nodesset	*retnodeset;
	int	(*compare)();

	if (flag & ~XYZ_ALL)
		fatal("getnodes_at(%8x, ...): not good flag", flag);
	for (i=0;i<3;i++) {
		check[i] = 0;
		coorsno[i] = 0;
	}

	if (flag & X_ONLY)
		check[0] = 1;
	if (flag & Y_ONLY)
		check[1] = 1;
	if (flag & Z_ONLY)
		check[2] = 1;

	limits[0][0] = xmin;
	limits[0][1] = xmax;
	limits[1][0] = ymin;
	limits[1][1] = ymax;
	limits[2][0] = zmin;
	limits[2][1] = zmax;

	compare = nodep_compare;
	for (i=0; i < 3; i++) {
		npriv[i] = NULL;
		if (check[i] == 0)
			continue;
		npr = npriv[i] = (struct node_private *)
			malloc(sizeof(struct node_private));
		if (npriv[i] == NULL)
			memerr("getnodes_at(): at malloc npriv[%d]", i);
		npr->nodep_nodes  = NULL;
		npr->nodep_cnt    = 0;
		npr->nodep_alloced= 0;
	}

	cnt = 0;
	for (i=0; i < 3; i++) {
		if (check[i] == 0)
			continue;

		cnt++;	/* how many coors are checked? */
		npr = npriv[i];
		hpp = meshes->m_nhashes[i];	/* for i'th-coordinate */

		for (j=0;j < HASHSIZE; j++) {
			for (hp = hpp[j]; hp; hp = hp->h_next) {
				coor = hp->h_value;
				if (limits[i][0] < coor && coor < limits[i][1]){
					if (hp->h_nodes == NULL)
						continue; /*there are no nodes*/
					nodep_copy(npr, hp);
					coorsno[i]++;
				}
			}
		}
		if (npr->nodep_cnt == 0)
			goto nothing;
	}

	retnodeset = (struct nodesset *)malloc(sizeof(struct nodesset));
	if (retnodeset == NULL)
		memerr("getnodes_at(): mallocing retnodeset");

	/*
	 * Get the intersection
	 */
	prevnp = NULL;
	for (i=0; i < 3;i++) {
		if (check[i] == 0)
			continue;
		if (prevnp == NULL) {
			prevnp = npriv[i];
			continue;
		}
		npr = get_nodeinter(prevnp, npriv[i]);
		free((void *) npriv[i]->nodep_nodes);
		free((void *) npriv[i]);
		free((void *) prevnp->nodep_nodes);
		free((void *) prevnp);

		prevnp = npr;
	}

	/*
	 * prevnp is the final result, so let's make the prevnp neat form and
	 * transform into struct nodesset form
	 */
	qsort((int *) prevnp->nodep_nodes, prevnp->nodep_cnt,
		sizeof(nodno_t), compare);

	size = sizeof(nodno_t) * prevnp->nodep_cnt;

	retnodeset->ns_cnt = prevnp->nodep_cnt;
	retnodeset->ns_nodes = (nodno_t *)malloc(size);

	memcpy((void *)retnodeset->ns_nodes, prevnp->nodep_nodes, size);
	free((void *) prevnp->nodep_nodes);
	free((void *) prevnp);

	for (i=0;i<3;i++)
		retnodeset->ns_coorsno[i] = coorsno[i];

	retnodeset->ns_flag = flag;
	return retnodeset;
nothing:
	for (i=0;i<3;i++) {
		if (npriv[i] == NULL)
			continue;
		if (npriv[i]->nodep_nodes)
			free((void *) npriv[i]->nodep_nodes);
		free((void *) npriv[i]);
	}
	return NULL;
}

void
nodep_copy(struct node_private *npr, struct hash_t *hp)
{
	int	i;
	nodno_t	*nnp;

	nnp = hp->h_nodes;
	for (i=0; i < hp->h_cnt; i++) {
		if (npr->nodep_alloced == npr->nodep_cnt) {
			npr->nodep_alloced += NODEP_UNIT;
			if (npr->nodep_nodes)
				npr->nodep_nodes = (nodno_t *)
				    realloc(npr->nodep_nodes,
				    npr->nodep_alloced * sizeof(nodno_t));
			else
				npr->nodep_nodes = (nodno_t *)
					malloc(npr->nodep_alloced *
						sizeof(nodno_t));
		}
		npr->nodep_nodes[npr->nodep_cnt] = nnp[i];
		npr->nodep_cnt++;
	}
	return;
}


/*
 * From two set of node numbers, I want to search the numbers of the
 * intersection of these sets
 */

struct node_private *
get_nodeinter(struct node_private *one, struct node_private *two)
{
	int	(*compare)();
	nodno_t	*npone, *nptwo, *npget;
	int	alloced, allocunit, gotit, space;
	int	cnt[2];
	struct node_private *node_priv;

	/*
	 * To facilitate the intersection procedure, I want to make the
	 * node_private->nodep_nodes in the sorted form
	 */
	compare = nodep_compare;
	qsort((int *) one->nodep_nodes, one->nodep_cnt, sizeof(nodno_t),
		compare);
	qsort((int *) two->nodep_nodes, two->nodep_cnt, sizeof(nodno_t),
		compare);

	npone = one->nodep_nodes;
	nptwo = two->nodep_nodes;

	cnt[0] = cnt[1] = 0;

	gotit = 0;
	alloced = allocunit = 10;
	npget = (nodno_t *) malloc(sizeof(nodno_t) * alloced);
	while (cnt[0] < one->nodep_cnt && cnt[1] < two->nodep_cnt) {
		switch (INTCMP(npone[cnt[0]], nptwo[cnt[1]])) {
			case  1:
				cnt[1]++;       /* Hahaha.. you are a loser */
				break;
			case  -1:
				cnt[0]++;       /* I can not match you */
				break;
			case 0:                 /* we have equal power */
				if (alloced == gotit) {
					alloced += allocunit;
					npget = (nodno_t *) realloc(npget,
						sizeof(nodno_t) * alloced);
				}
				npget[gotit] = npone[cnt[0]];
				cnt[0]++;
				cnt[1]++;
				gotit++;
				break;
		}
	}

	node_priv = (struct node_private *)
		malloc(sizeof(struct node_private));
	if (node_priv == NULL)
		fatal("get_2inter(): malloc of node_priv");


	node_priv->nodep_cnt = gotit;

	space = sizeof(nodno_t) * node_priv->nodep_cnt;
	node_priv->nodep_nodes = (nodno_t *) malloc(space);

	if (node_priv->nodep_nodes == NULL)
		fatal("get_2inter(): malloc of node_priv->nodep_nodes");

	memcpy((void *) node_priv->nodep_nodes, (void *) npget, space);
	free((void *) npget);
	return node_priv;
}


void
nset_print(const struct nodesset *nsp)
{
	int	i;
	nodno_t *nnp;

	nnp = nsp->ns_nodes;
	for (i=0;i<nsp->ns_cnt;i++) {
		printf("%3d ", nnp[i]);
	}
	printf("+++\n");

	return;
}

void
nset_free(struct nodesset *nsp)
{
	free((void *) nsp->ns_nodes);
	free((void *) nsp);
	return;
}

void
nodep_print(const struct node_private *np)
{
	int	i, cnt;

	for (i=0, cnt=0; i < np->nodep_cnt;i++, cnt++) {
		printf("%3d ", np->nodep_nodes[i]);
		if (((cnt+1) % 10) == 0)
			printf("\n");
	}
	printf("+++\n");
	return;
}



/*
 * FOR ELEMENT MANIPULATION
 * Private data structure of element
 */


struct ementsset *
getements_at(int flag, float xmin, float xmax,
	float ymin, float ymax, float zmin, float zmax)
{
	int	i, j, cnt, size;
	struct hash_t *hp, **hpp;
	float	coor, limits[3][2];
	struct ement_private *epriv[3], *epr, *prevep;
	int	check[3];
	struct ementsset	*retementset;
	int	(*compare)();

	if (flag & ~XYZ_ALL)
		fatal("getements_at(%8x, ...): not good flag", flag);
	for (i=0;i<3;i++)
		check[i] = 0;

	if (flag & X_ONLY)
		check[0] = 1;
	if (flag & Y_ONLY)
		check[1] = 1;
	if (flag & Z_ONLY)
		check[2] = 1;

	limits[0][0] = xmin;
	limits[0][1] = xmax;
	limits[1][0] = ymin;
	limits[1][1] = ymax;
	limits[2][0] = zmin;
	limits[2][1] = zmax;

	compare = nodep_compare;	/* same as that of nodes */
	for (i=0; i < 3; i++) {
		epriv[i] = NULL;
		if (check[i] == 0)
			continue;
		epr = epriv[i] = (struct ement_private *)
			malloc(sizeof(struct ement_private));
		if (epriv[i] == NULL)
			memerr("getements_at(): at malloc epriv[%d]", i);
		epr->ementp_ements  = NULL;
		epr->ementp_cnt    = 0;
		epr->ementp_alloced= 0;
	}

	cnt = 0;
	for (i=0; i < 3; i++) {
		if (check[i] == 0)
			continue;

		cnt++;	/* how many coors are checked? */
		epr = epriv[i];
		hpp = meshes->m_ehashes[i];	/* for i'th-coordinate */

		for (j=0;j < HASHSIZE; j++) {
			for (hp = hpp[j]; hp; hp = hp->h_next) {
				coor = hp->h_value;
				if (limits[i][0] < coor && coor < limits[i][1]){
					if (hp->h_nodes == NULL)
						continue;/*there are no ements*/
					ementp_copy(epr, hp);
				}
			}
		}
		if (epr->ementp_cnt == 0)
			goto nothing;
	}

	retementset = (struct ementsset *)malloc(sizeof(struct ementsset));
	if (retementset == NULL)
		memerr("getements_at(): mallocing retementset");

	/*
	 * Get the intersection
	 */
	prevep = NULL;
	for (i=0; i < 3;i++) {
		if (check[i] == 0)
			continue;
		if (prevep == NULL) {
			prevep = epriv[i];
			continue;
		}
		epr = get_ementinter(prevep, epriv[i]);
		free((void *) epriv[i]->ementp_ements);
		free((void *) epriv[i]);
		free((void *) prevep->ementp_ements);
		free((void *) prevep);

		prevep = epr;
	}

	/*
	 * prevep is the final result, so let's make the prevep neat form and
	 * transform into struct ementsset form
	 */
	qsort((int *) prevep->ementp_ements, prevep->ementp_cnt,
		sizeof(nodno_t), compare);

	size = sizeof(nodno_t) * prevep->ementp_cnt;

	retementset->es_cnt = prevep->ementp_cnt;
	retementset->es_ements = (nodno_t *)malloc(size);

	memcpy((void *)retementset->es_ements, prevep->ementp_ements, size);
	free((void *) prevep->ementp_ements);
	free((void *) prevep);

	retementset->es_flag = flag;
	return retementset;
nothing:
	for (i=0;i<3;i++) {
		if (epriv[i] == NULL)
			continue;
		if (epriv[i]->ementp_ements)
			free((void *) epriv[i]->ementp_ements);
		free((void *) epriv[i]);
	}
	return NULL;
}

void
ementp_copy(struct ement_private *epr, struct hash_t *hp)
{
	int	i;
	nodno_t	*nnp;

	nnp = hp->h_nodes;
	for (i=0; i < hp->h_cnt; i++) {
		if (epr->ementp_alloced == epr->ementp_cnt) {
			epr->ementp_alloced += NODEP_UNIT;
			if (epr->ementp_ements)
				epr->ementp_ements = (nodno_t *)
				    realloc(epr->ementp_ements,
				    epr->ementp_alloced * sizeof(nodno_t));
			else
				epr->ementp_ements = (nodno_t *)
					malloc(epr->ementp_alloced *
						sizeof(nodno_t));
		}
		epr->ementp_ements[epr->ementp_cnt] = nnp[i];
		epr->ementp_cnt++;
	}
	return;
}


/*
 * From two set of node numbers, I want to search the numbers of the
 * intersection of these sets
 */

struct ement_private *
get_ementinter(struct ement_private *one, struct ement_private *two)
{
	int	(*compare)();
	nodno_t	*npone, *nptwo, *npget;
	int	alloced, allocunit, gotit, space;
	int	cnt[2];
	struct ement_private *ement_priv;

	/*
	 * To facilitate the intersection procedure, I want to make the
	 * ement_private->ementp_nodes in the sorted form
	 */
	compare = nodep_compare;
	qsort((int *) one->ementp_ements, one->ementp_cnt, sizeof(nodno_t),
		compare);
	qsort((int *) two->ementp_ements, two->ementp_cnt, sizeof(nodno_t),
		compare);

	npone = one->ementp_ements;
	nptwo = two->ementp_ements;

	cnt[0] = cnt[1] = 0;

	gotit = 0;
	alloced = allocunit = 10;
	npget = (nodno_t *) malloc(sizeof(nodno_t) * alloced);
	while (cnt[0] < one->ementp_cnt && cnt[1] < two->ementp_cnt) {
		switch (INTCMP(npone[cnt[0]], nptwo[cnt[1]])) {
			case  1:
				cnt[1]++;       /* Hahaha.. you are a loser */
				break;
			case  -1:
				cnt[0]++;       /* I can not match you */
				break;
			case 0:                 /* we have equal power */
				if (alloced == gotit) {
					alloced += allocunit;
					npget = (nodno_t *) realloc(npget,
						sizeof(nodno_t) * alloced);
				}
				npget[gotit] = npone[cnt[0]];
				cnt[0]++;
				cnt[1]++;
				gotit++;
				break;
		}
	}

	ement_priv = (struct ement_private *)
		malloc(sizeof(struct ement_private));
	if (ement_priv == NULL)
		fatal("get_2inter(): malloc of ement_priv");

	ement_priv->ementp_cnt = gotit;
	space = sizeof(nodno_t) * ement_priv->ementp_cnt;
	ement_priv->ementp_ements = (nodno_t *) malloc(space);
	if (ement_priv->ementp_ements == NULL)
		fatal("get_2inter(): malloc of ement_priv->ementp_ements");

	memcpy((void *) ement_priv->ementp_ements, (void *) npget, space);
	free((void *) npget);
	return ement_priv;
}


void
eset_print(const struct ementsset *esp)
{
	int	i;
	elmno_t *enp;

	enp = esp->es_ements;
	for (i=0;i < esp->es_cnt;i++) {
		printf("%3d ", enp[i]);
	}
	printf("+++\n");

	return;
}
void
eset_free(struct ementsset *esp)
{
	free((void *) esp->es_ements);
	free((void *) esp);
	return;
}

void
ementp_print(const struct ement_private *ep)
{
	int	i, cnt;

	for (i=0, cnt=0; i < ep->ementp_cnt; i++, cnt++) {
		printf("%3d ", ep->ementp_ements[i]);
		if (((cnt+1) % 10)==0)
			printf("\n");
	}
	printf("+++\n");
	return;
}
