#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "common.h"
#include "vector.h"
#include "matrix.h"
#include "sparse.h"
#include "finite.h"

void	set_const3rd(const double);
void	radial(const struct vector_t *, const char *fname);
double	getradius(const struct node_t *np);
double	bessj0(double);
double	bessy0(double);
void	get_exact(struct vector_t *);
void	get_exact2nd(struct vector_t *ret, const double incre);
double	diff_awave(const struct vector_t *vp);


#define LEFT	1
#define NEAR	2
#define RESIST  3	/* XXX */
#define RIGHT	4
#define FAR	5
#define UPPER	6
#define LOWER	7

#define F_LEFT	0x01
#define F_RIGHT	0x02
#define F_NEAR	0x04
#define F_FAR	0x08

#define X_CENTER  0.50
#define XL_MARGIN 0.075
#define XR_MARGIN 0.925
#define X_LEN (XL_MARGIN+XR_MARGIN)

#define Y_CENTER  0.50
#define YL_MARGIN 0.075
#define YR_MARGIN 0.925
#define Y_LEN (YL_MARGIN+YR_MARGIN)

int	lighteno_1st;
int	lighteno_2nd;

float	lightx, lighty;
struct node_t *lightnode;

void	vec_zprint(const struct vector_t *vp, const char *fname);
int	whichsubdo(const struct ement_t *);
int	whichbound(const struct node_t *);
void	make_femmatrix(void);

void	impose_bcs(void);

struct vector_t *mux;
struct vector_t *muy;
struct vector_t *epsilon;

double	wave_k;
double	lambda;

void
problem(void)
{
	double	incre;
	struct vector_t *exact, *differ;
	double	gap;

	lambda = 0.15;
	wave_k = 2.0 * M_PI/lambda;

	lighteno_1st = find_nearestement(0.5, 0.5);
	lighteno_2nd = find_nearestement(0.5, 0.5);

	lightx = get_centercoor(lighteno_1st, 0);
	lighty = get_centercoor(lighteno_1st, 1);

	printf("lighteno_1st = %4d\n", lighteno_1st);
	printf("lightx = %7.4f\n", lightx);
	printf("lighty = %7.4f\n", lighty);

	exact = vec_puredup(meshes->m_bvector);
	differ = vec_puredup(meshes->m_bvector);

	mk_nearements();
	set_subdomain(whichsubdo);
	set_boundary(whichbound);

	put_2dgrids("grids.mtv", 0);

	set_const3rd(1.0);

	if (0) {
		put_gridandvector(mux, "mux");
		put_gridandvector(mux, "muy");
	}

	vec_null(meshes->m_bvector);
	sp_null(meshes->m_kmatrix);

	make_femmatrix();
	impose_bcs();
	solveall();

	put_gridandvector(meshes->m_bvector, "a");
	get_exact2nd(exact, 0.0);

	vec_minus(differ, exact, meshes->m_bvector);
	put_gridandvector(differ, "differ");
	put_gridandvector(exact, "exact");

	radial(meshes->m_bvector, "bmat");
	radial(exact, "exat");
	radial(differ, "rdiffer");

	gap = diff_awave(differ);
	printf("%12.5e %12.5e\n", incre, gap);

	return;
}

void
get_exact2nd(struct vector_t *ret, const double incre)
{
	int	i;
	struct node_t *np;
	struct duplex_t *dx;
	float	len;

	dx = zvec(ret);
	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		len = wave_k * (getradius(np) + incre);
		dx[i] = duplexv(-bessy0(len), -bessj0(len));
	}

	return;
}

void
get_exact(struct vector_t *ret)
{
	int	i;
	struct node_t *np;
	struct duplex_t *dx;
	float	len;

	dx = zvec(ret);
	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		len = wave_k * getradius(np);
		dx[i] = duplexv(-bessy0(len), -bessj0(len));
	}

	return;
}

int
whichsubdo(const struct ement_t *ep)
{
	struct proto_t *proto;
	struct node_t *np;
	nodno_t *nnp;
	int	noofnodes;
	float	means[3];
	int	i, j, flag;
	double	x, y;

	proto = ep->e_proto;
	nnp = ep->e_nodes;
	noofnodes = ep->e_nodesno;

	for (i=0;i<meshes->m_spacedim;i++)
		means[i] = 0.0; /* initialization */

	for (i=0;i<proto->p_nodsno;i++) {
		np = meshes->m_nodes[nnp[i]]; /* for surrounding nodes*/
		for (j=0; j < meshes->m_spacedim; j++)
			means[j] += np->n_coor[j];
	}

	for (i=0;i < meshes->m_spacedim;i++)
		means[i] /= (float) noofnodes; /* initialization */

	flag = 0;
	x = means[0];
	y = means[1];

#if 0
	if (x <= XL_MARGIN && (y > YL_MARGIN && y < YR_MARGIN))
		return LEFT;
	if (x >= XR_MARGIN && (y > YL_MARGIN && y < YR_MARGIN))
		return RIGHT;

	if (y <= YL_MARGIN && (x > XL_MARGIN && x < XR_MARGIN))
		return NEAR;
	if (y >= YR_MARGIN && (x > XL_MARGIN && x < XR_MARGIN))
		return FAR;

#else
	if (x <= XL_MARGIN)
		flag |= F_LEFT;
	if (x >= XR_MARGIN)
		flag |= F_RIGHT;

	if (y <= YL_MARGIN)
		flag |= F_NEAR;
	if (y >= YR_MARGIN)
		flag |= F_FAR;
#endif

	switch (flag) {
		case F_LEFT:	return LEFT; break;
		case F_RIGHT:	return RIGHT; break;
		case F_NEAR :	return NEAR; break;
		case F_FAR:	return FAR; break;
		case F_LEFT | F_NEAR:
			if (means[0] < means[1])
				return LEFT;
			else
				return NEAR;
			break;
		case F_RIGHT | F_NEAR:
			if ((X_LEN-means[0]) < means[1])
				return RIGHT;
			else
				return NEAR;
			break;
		case F_LEFT | F_FAR:
			if (means[0] > (Y_LEN-means[1]))
				return FAR;
			else
				return LEFT;
			break;
		case F_RIGHT | F_FAR:
			if ((X_LEN-means[0]) > (Y_LEN-means[1]))
				return FAR;
			else
				return RIGHT;
			break;
	}

	return 0;
}

int
whichbound(const struct node_t *np)
{
	float	x, y;

	x = np->n_coor[0];
	y = np->n_coor[1];

#if XXX
	if (fabs(x - 2.0) < 0.0001 && (y  <= 1.00)) /* right lower */
		return 4;
	if (fabs(x) < 0.0001 && (y  <= 1.00)) /* left lower */
		return 4;
	if (fabs(y - 1.0) < 0.001) {
		rl = rl_findme(x);
		if (rl && (rl->rl_flag == RL_FAR))
			return 4;
	}
#endif
	return -1;
}

void
make_femmatrix(void)
{
	int     i;
	struct matrix_t    *retmatf, *bmatf;
	struct proto_t  *prevproto = NULL;

	currp = NULL;
	sp_null(meshes->m_kmatrix);
	vec_null(meshes->m_bvector);

	for (i=0; i <  meshes->m_ementsno; i++) {
		/*
		 * currp is only set at this location. You must not set currp
		 * at other places.
		 */
		currp = meshes->m_ements[i]->e_proto;
		if (prevproto != currp) {
			if (prevproto != NULL)
				m_mfree(2, retmatf, bmatf);/* no longer used */
			 create_smalls(currp, &retmatf, &bmatf);
		}
		integration(i, retmatf, bmatf);
		bassemble(i, bmatf, meshes->m_bvector);
		Kassemble(i, retmatf, meshes->m_kmatrix);
		prevproto = currp;
	}
	m_mfree(2, retmatf, bmatf);
	return;
}

void
integration(const int elemno, struct matrix_t *kmatf, struct matrix_t *bmatf)
{
	struct ement_t *ep;
	int	i, j, k;
	struct duplex_t **retmat, **bmat;
	struct integral_t *intg;
	double	**unit, **deriv;
	struct duplex_t	phase;
	double	*jacob, *weight;
	double	rightside;
	nodno_t	*nnp;

	struct duplex_t axx, ayy, beta;
	struct duplex_t dxtmp, dx1, dx2, dx3;

	m_mnull(2, kmatf, bmatf); /* XXX */
	ep = meshes->m_ements[elemno];
	intg   = currp->p_intg;
	nnp    = ep->e_nodes;
	proto_einit(elemno);
	jacob  = intg->intg_jacob;
	weight = intg->intg_weights;

	unit   = dmat(intg->intg_trans);

	retmat = zmat(kmatf);
	bmat   = zmat(bmatf);

	rightside = 1.0;
	phase = duplexv(1.0/4.0, 0.0);
	for (k=0; k < currp->p_intgno; k++) {	/* for integration points */
		deriv = dmat(intg->intg_deriv[k]);

		dx1 = zgetat(elemno, k, muy);
		dx2 = zgetat(elemno, k, mux);
		dx3 = zgetat(elemno, k, epsilon);
		if (zp_isnull(&dx1))
			dx1 = zgetat(elemno, k, muy);
		if (zp_isnull(&dx2))
			dx2 = zgetat(elemno, k, mux);

		zp_inverse(&axx, &dx1);
		zp_inverse(&ayy, &dx2);
		zp_crossd(&beta, &dx3, -1.0 * wave_k * wave_k);

		for (i=0; i< currp->p_nodsno;i++) {	/* XXX */
			for (j=0; j< currp->p_nodsno;j++) {
				zp_crossd(&dx1, &axx, deriv[0][i]*deriv[0][j]);
				zp_crossd(&dx2, &ayy, deriv[1][i]*deriv[1][j]);

				zp_crossd(&dx3, &beta, unit[k][i]*unit[k][j]);

				zp_3plus(&dxtmp, &dx1, &dx2, &dx3);
				zp_crossd(&dxtmp, &dxtmp, jacob[k] * weight[k]);

				zp_pplus(&retmat[i][j], &dxtmp);
			}
			if (elemno == lighteno_1st || elemno == lighteno_2nd) {
				zp_pplus(&bmat[i][0], &phase);
				if (0)
				printf("light element no = %4d\n", elemno);
			}
		}
	}
#if 0
	if ((elemno%200) == 0)
		printf("%5d\n", elemno);
#endif
	return;
}


void
impose_bcs(void)
{
	zdirichlet(0, duplex_null);
	zdirichlet(1, duplex_null);
	zdirichlet(2, duplex_null);
	zdirichlet(3, duplex_null);
#if 0
	int	i;
	struct node_t *np;
	struct duplex_t dx;
	float	len;

	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		if (np->n_bdflag) {
			len = wave_k * getradius(np);
			dx = duplexv(4.0*bessj0(len), -4.0*bessy0(len));
			di_zimpose(i, dx);
		}
	}
#endif

	return;
}

void
vec_zprint(const struct vector_t *vp, const char *fname)
{
	int i;
	FILE	*fp;

	struct duplex_t *zp;

	fp = fopen(fname, "w");
	if (fp == NULL)
		fatal("%s():", __F);
	zp = zvec(vp);
	for (i=0;i < vp->vt_size; i++)
		(void) fprintf(fp, "%13.5e %13.5e\n",
			zp[i].real, zp[i].imag);
	fclose(fp);
	return;
}

/*
 * Following does not multiply with cosine
 */

double
get_beta(const double order, const double delta, const double totlen)
{
	double	retval;
	double	factor;

	factor = 14.0 / (totlen * wave_k); 	/* XXX */

	retval = (delta/totlen);
	retval = pow(retval, order);
	retval *= (order + 1.0) * factor;
	retval *= -1.0;

	return retval;
}

/*
 *
@ARTICLE{sacks,
        author="Zachary S. Sacks,
		David M. Kingsland, Robert lee and Jin-Fa Lee",
        title="``{A Perfectly Matched Anisotropic Absorber
                for Use as an Absorbing Boundary Condition}''",
        journal="IEEE Transactions on Antennas and Propagation",
        volume="43",
        number="12",
        pages="1460--1463",
        year="Dec. 1995"
}
 */

void
set_const3rd(const double order)
{
	int	i, j, subdomain;
	struct node_t *np;
	nodno_t	*nnp;

	double	cosine;
	double	xcoor, ycoor, delta;
	double	sq, howmany;
	struct duplex_t ztmp0, ztmp1;
	struct duplex_t *muxp, *muyp, *epsp;

	printf("start of %s\n", __F);

	mux = vector(meshes->m_nodesno, "mux", SP_DUPLEX);
	muy = vector(meshes->m_nodesno, "muy", SP_DUPLEX);
	epsilon = vector(meshes->m_nodesno, "epsilon", SP_DUPLEX);

	muxp = zvec(mux);
	muyp = zvec(muy);
	epsp = zvec(epsilon);

	for (i=0; i < meshes->m_nodesno;i++) {
		np = meshes->m_nodes[i];
		xcoor = np->n_coor[0];
		ycoor = np->n_coor[1];
		nnp = np->n_ements;

		if (0)
			cosine = get_cosine(lighteno_1st, i);
		else
			cosine = get_cosine(lighteno_1st, i);
		for (j=0; j < np->n_ementsno; j++) {
			subdomain = meshes->m_ements[nnp[j]]->e_sdomain;
			switch (subdomain) {
				case LEFT:
				case RIGHT:
					if (subdomain == LEFT)
						delta = fabs(xcoor-XL_MARGIN);
					else
						delta = fabs(xcoor-XR_MARGIN);

					sq = get_beta(order, delta, XL_MARGIN);
					sq /= cosine;

					ztmp0 = duplexv(1.0, sq);
					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp1);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				case NEAR:
				case FAR:
					if (subdomain == NEAR)
						delta = fabs(ycoor-YL_MARGIN);
					else
						delta = fabs(ycoor-YR_MARGIN);

					sq = get_beta(order, delta, YL_MARGIN);
					sq /= cosine;


					ztmp0 = duplexv(1.0, sq);
					zp_inverse(&ztmp1, &ztmp0);

					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp1);
					zp_pplus(&epsp[i], &ztmp0);
					break;
				default:
					ztmp0 = duplexv(1.0, 0.0);
					zp_pplus(&muxp[i], &ztmp0);
					zp_pplus(&muyp[i], &ztmp0);
					zp_pplus(&epsp[i], &ztmp0);
					break;
			}
		}
		howmany = 1.0/(double) j;

		zp_crossdme(&muxp[i], howmany);
		zp_crossdme(&muyp[i], howmany);
		zp_crossdme(&epsp[i], howmany);
	}

	printf("end of %s\n", __F);
	return;
}

int
allgood(const struct node_t *np)
{
	struct ement_t *ep;
	nodno_t *nnp;
	int	i;

	nnp = np->n_ements;
	for (i=0; i < np->n_ementsno; i++) {
		ep = meshes->m_ements[nnp[i]];
		if (ep->e_sdomain)
			return 1;
	}
	return 0;	/* GOOD */
}

double
getradius(const struct node_t *np)
{
	double	x, y;

	x = np->n_coor[0];
	y = np->n_coor[1];

	x -= lightx;
	y -= lighty;
	return sqrt(x*x+y*y);
}

void
radial(const struct vector_t *vp, const char *fname)
{
	FILE	*fp[3];
	char	names[3][132];
	char	*tailes[3]={".real.rad", ".imag.rad", ".inten.rad"};
	int	i;
	struct duplex_t *zp;
	struct node_t *np;
	double	radius;

	for (i=0; i < 3; i++) {
		strcpy(names[i], fname);
		strcat(names[i], tailes[i]);
		fp[i] = fopen(names[i], "w");
		if (fp[i] == NULL)
			fatal("%s():%s", __F, names[i]);
	}

	zp = zvec(vp);

	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		if (allgood(np))
			continue;

		radius = getradius(np);
		fprintf(fp[0], "%7.5f %12.5e\n", radius, zp[i].real);
		fprintf(fp[1], "%7.5f %12.5e\n", radius, zp[i].imag);
		fprintf(fp[2], "%7.5f %12.5e\n", radius, zp_absolute(&zp[i]));
	}

	for (i=0; i<3; i++)
		fclose(fp[i]);
	return;
}

double
diff_awave(const struct vector_t *vp)
{
	int	i, nodecnt;
	struct duplex_t *zp;
	struct node_t *np;
	double	radius, star, endr, ret;

	star = 11.7915;
	endr = star + M_PI;

	star /= wave_k;
	endr /= wave_k;

	nodecnt = 0;
	ret = 0.0;
	zp = zvec(vp);

	for (i=0; i < meshes->m_nodesno; i++) {
		np = meshes->m_nodes[i];
		if (allgood(np))
			continue;

		radius = getradius(np);
		if (radius < star || radius > endr)
			continue;

		ret += zp_absolute(&zp[i]);
		nodecnt++;
	}

	ret /= (double) nodecnt;

	return ret;
}
