#ifndef H_SPARSE_H
#define H_SPARSE_H

#include <sys/types.h>

#include "common.h"
#include "complex.h"
#include "vector.h"

struct sp_roworder {
	int	*spr_rowindex;	/* counterpart of sp_colx */
	int	*spr_colindex;	/* counterpart of sp_rowx */
	int	*spr_dataindex;	/* indicator of location of the data entries */
};

struct sparse_t;

struct forprecond {
	struct sparse_t *pcd_diag;
	struct sparse_t *pcd_diaginv;
	struct sparse_t *pcd_upper;
	struct sparse_t *pcd_upperinv;
	struct sparse_t *pcd_lower;
	struct sparse_t *pcd_lowerinv;
};

/*
 * sparse matrix structure that will be used extensively in this package.
 */

struct sparse_t {
	int	 sp_rown;	/* no of rows */
	int	 sp_coln;	/* no of columns */
	int	 sp_nonzero;	/* no of nonzeros */
	int	*sp_rowx;	/* the index of rows */
	int	*sp_colx;	/* the index of cols */
	int	 sp_type;	/* the sparse_t matrix data type */
	int	*sp_data;	/* the real location where datas are stored */
	char	*sp_name;	/* the name of this sparse_t matrix */
	int	 sp_flag;	/* an indicator for specifying the sparse_t */

	int	 sp_symflag;	/* indicates only the storage is symmetric */
	int	 sp_issquare;	/* indicator whether this is n by n matrix */
	struct sp_roworder *sp_roworder;
	struct forprecond  *sp_precond;
};

#define	sp_rowindex	sp_roworder->spr_rowindex
#define	sp_colindex	sp_roworder->spr_colindex
#define	sp_dataindex	sp_roworder->spr_dataindex

#define sp_diag		sp_precond->pcd_diag
#define sp_diaginv	sp_precond->pcd_diaginv
#define sp_upper	sp_precond->pcd_upper
#define sp_upperinv	sp_precond->pcd_upperinv
#define sp_lower	sp_precond->pcd_lower
#define sp_lowerinv	sp_precond->pcd_lowerinv

#define ONE_OF_SP(c) (c == SP_SINGLE || c == SP_DOUBLE || c == SP_SIMPLEX \
		|| c == SP_DUPLEX)

#define SP_DIAGONAL	0x01000000
#define	SP_ISSPARSE	0x80000000

#define	IS_SPARSE(sp)	(sp->sp_flag & SP_ISSPARSE)
#define	SET_SPARSE(sp)	(sp->sp_flag |= SP_ISSPARSE)

#define	IS_DIAGONAL(sp)	(sp->sp_flag & SP_DIAGONAL)
#define SET_DIAGONAL(sp)(sp->sp_flag |= SP_DIAGONAL)
#define CLR_DIAGONAL(sp)(sp->sp_flag &= ~SP_DIAGONAL)

/*
 * flag of sp_symflag
 */
#define SP_NOSYM	0x00000000
#define SP_SYMPATTERN	0x00000001
#define SP_SYMMETRIC	0x00000002
#define SP_HERMITIAN	0x00000004
#define SP_COMPLEXSYM	0x00000008
#define SP_SYMCHECKED	0x00000010

/*
 * The routines
 */

struct sparse_t *sparse(int, int, int, const char *, int);
void		 sp_check(const char *function, const int arg,
			const struct sparse_t *sp);
void		 sp_pattern(const struct sparse_t *);
void		 sp_pattern2nd(struct sparse_t *sp);

int		 sp_printatfile(const struct sparse_t *, const char *);
int		 sp_print(const struct sparse_t *);
struct sparse_t *sp_read(const char *filename);
void		 sp_info(const struct sparse_t *);
void		 sp_briefinfo(const struct sparse_t *sp);

void		 sp_free(struct sparse_t *);
void	 	 sp_mfree(const int, ...);
void	 	 sp_null(struct sparse_t *);
void	 	 sp_freerowindex(struct sparse_t *);
void		 sp_rowindexstorage(struct sparse_t *sp);
void	 	 sp_mkroworder(struct sparse_t *);
void	 	 sp_mkroworder2nd(struct sparse_t *sp);
void		 sp_putusingrow(struct sparse_t *, const char *, const char *);
struct sparse_t *sp_getupper(const struct sparse_t *sp);
struct sparse_t *sp_getlower(const struct sparse_t *sp);

struct sparse_t *sp_copy(const struct sparse_t *);

float		 *ssp(const struct sparse_t *sp);
double		 *dsp(const struct sparse_t *sp);
struct simplex_t *csp(const struct sparse_t *sp);
struct duplex_t  *zsp(const struct sparse_t *sp);

#define CONJ_NO		0x00000000
#define CONJ_YES	0x00000001
#define CONJ_ALL	(CONJ_NO | CONJ_YES)

struct vector_t *sp_dotvec(struct vector_t *xp, const int indicator,
			struct sparse_t *sp, const struct vector_t *vp);
struct vector_t *sp_transdotvec(struct vector_t *xp, const int indicator,
			struct sparse_t *sp, const struct vector_t *vp);

void	 	 sp_symcheck(struct sparse_t *sp);
int	 	 sp_getwhereis(const struct sparse_t *, const int, const int);
struct vector_t *sp_getdiagvec(struct vector_t *xp, const struct sparse_t *sp);
struct sparse_t *sp_getdiagsp(const struct sparse_t *sp);
struct sparse_t *sp_getinvdiag(const struct sparse_t *sp);

/*
 * to get information of a specified row, or col; here taxis means 'order'.
 * taxis is Greek oriented word as I know. struct pairs will be used only at
 * sp_getpairs(). I think I have no sufficient imagination to invent a good
 * function name. The return value is the total length of pairs[]; Remember
 * that you should provide enough space of struct pairs before calling this
 * sp_getpairs()
 */
#define SP_FINDCOLS 10
#define SP_FINDROWS 20

struct pairs {
	int	pr_taxis;	/* row no when SP_FINDCOL; */
	int	pr_locus;	/* the data location of pr_taxis */
};

int sp_getpairs(struct pairs provided[], const int indicator, const int alloced,
	struct sparse_t *sp, const int col);

/*
 * sparse_t and sparse_t multiplication
 */
struct sparse_t *sp_crosssp(struct sparse_t *spa, struct sparse_t *spb);
void	sp_crossspable(const char *f, const struct sparse_t *a,
		const struct sparse_t *b);
void	sp_tcrossspable(const char *f, const struct sparse_t *a,
		const struct sparse_t *b);

struct matrix_t *sp_tomat(const struct sparse_t *sp);

/*
 * The private functions of the sparse_t matrix
 */

struct sparse_t *sp_sread(const char *filename);
struct sparse_t *sp_dread(const char *filename);
struct sparse_t *sp_cread(const char *filename);
struct sparse_t *sp_zread(const char *filename);

struct sparse_t *sp_sreadraw(const char *filename);
struct sparse_t *sp_dreadraw(const char *filename);
struct sparse_t *sp_creadraw(const char *filename);
struct sparse_t *sp_zreadraw(const char *filename);

struct sparse_t *sp_sadd(const struct sparse_t *, const struct sparse_t *);
struct sparse_t *sp_dadd(const struct sparse_t *, const struct sparse_t *);
struct sparse_t *sp_cadd(const struct sparse_t *, const struct sparse_t *);
struct sparse_t *sp_zadd(const struct sparse_t *, const struct sparse_t *);

/*
 * For sparse matrix addtion
 */
struct detail {
	int     d_whatrow;
	int     d_whichmat;
};

struct addcols {
	int      col_colno;
	int      col_norows;
	struct detail *col_datas;       /* For sorting */
	int      col_alloced;
};

int sp_rowcmp(const struct detail *one, const struct detail *another);

void	sp_dputusingrow(struct sparse_t *sp);
void	sp_zputusingrow(struct sparse_t *sp);

struct sparse_t *sp_scrosssp(struct sparse_t *, struct sparse_t *);
struct sparse_t *sp_dcrosssp(struct sparse_t *,struct sparse_t *);
struct sparse_t *sp_ccrosssp(struct sparse_t *, struct sparse_t *);
struct sparse_t *sp_zcrosssp(struct sparse_t *,struct sparse_t *);

int s_getproduct(struct sparse_t *, int *, int *, int,
		struct sparse_t *, int *, int, int, float *);
int d_getproduct(struct sparse_t *, int *, int *, int,
		struct sparse_t *, int *, int, int, double *);
int c_getproduct(struct sparse_t *, int *, int *, int,
		struct sparse_t *, int *, int, int, struct simplex_t *);
int z_getproduct(struct sparse_t *, int *, int *, int,
		struct sparse_t *, int *, int, int, struct duplex_t *);

/*
 * VECTOR with sparse format
 */

struct spvector_t
{
	int	 spv_size;
	int	 spv_alloced;
	int	*spv_index;
	int	 spv_type;
	int	 spv_flag;
	int	*spv_data;
	char	*spv_name;
};

#define SPVECTOR_TYPE 0x12322121

struct spvector_t	*spvector(int size, const char *name, int type);
inline float		*sspvec(const struct spvector_t *spv);
inline double		*dspvec(const struct spvector_t *spv);
inline struct simplex_t *cspvec(const struct spvector_t *spv);
inline struct duplex_t  *zspvec(const struct spvector_t *spv);

void	spv_null(struct spvector_t *spv);
void	spv_mnull(int  num,...);
void	zspv_info(const struct spvector_t *spv);
void	spv_print(const struct spvector_t *spv);
void	spv_mprint(int  num,...);


#endif
