#include "common.h"
#include "sparse.h"
#include "finite.h"

/*
 * set the nodes whose doundary indicator is int indicator to the value 
 * specified at the arguement.
 * DOUBLE version is ddirichlet() and complex version is zdirichlet();
 */
void
ddirichlet(const int indicator, const double value)
{
	int	i, flag;

	if (meshes->m_datatype != SP_SINGLE && meshes->m_datatype != SP_DOUBLE)
		fatal("%s():dont call ce func at SIMPLEX or DUPLEX world", __F);

	for (i=0; i < meshes->m_nodesno;i++) {
		flag = meshes->m_nodes[i]->n_bdflag;
		if (flag == 0)
			continue;
		if (isset(flag, indicator))
#if 0
			if ((flag & (flag-1)) == 0)
#endif
				di_dimpose(i, value);
	}
	return;
}

void
zdirichlet(const int indicator, const struct duplex_t value)
{
	int	i, flag;

	if (meshes->m_datatype != SP_SIMPLEX && meshes->m_datatype != SP_DUPLEX)
		fatal("%s():dont call ce func at SINGLE or DOUBLE world", __F);

	for (i=0; i < meshes->m_nodesno;i++) {
		flag = meshes->m_nodes[i]->n_bdflag;
		if (flag == 0)
			continue;
		if (isset(flag, indicator))
			if ((flag & (flag-1)) == 0)
				di_zimpose(i, value);
	}
	return;
}
	
/*
 * Possiblely a private function of zdirichlet()
 * 
 * set the node whose node number is const int nodeno, to the value specified
 * at the arguement. So that even though these function can be thought of a 
 * private functions of zdirichlet or ddirichlet, you can call this function
 * to set the value of a certain node.
 * DOUBLE version is di_dimpose() and complex version is di_zimpose();
 */
void
di_zimpose(const int nodeno, const struct duplex_t value)
{
	int	i;
	struct simplex_t *ckmat, *cbmat, ctmp;
	struct duplex_t *zkmat, *zbmat, ztmp;

	int	locus, row;
	struct sparse_t *sp;
	int	norows;
	struct pairs rowarr[50];	/* is this suffic?? */

	sp = meshes->m_kmatrix;

	switch (meshes->m_datatype) {
		case SP_SIMPLEX:
			ckmat = (struct simplex_t *) sp->sp_data;
			cbmat = cvec(meshes->m_bvector);

			norows = sp_getpairs(rowarr, SP_FINDCOLS, 50,
				sp, nodeno);

			ctmp.real = value.real;
			ctmp.imag = value.imag;

			cbmat[nodeno] = ctmp;
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;

				if (row == nodeno) {
					ckmat[locus] = simplex_unit;
					continue;
				}

				cp_mminus(&cbmat[row], cp_cross(&ctmp,
					&ckmat[locus], &cbmat[nodeno]));
				ckmat[locus] = simplex_null;
			}
			norows = sp_getpairs(rowarr, SP_FINDROWS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (row == nodeno)
					continue;
				ckmat[locus] = simplex_null;
			}
			break;
		
		case SP_DUPLEX:
			zkmat = (struct duplex_t *) sp->sp_data;
			zbmat = zvec(meshes->m_bvector);

			norows = sp_getpairs(rowarr, SP_FINDCOLS, 50,
				sp, nodeno);

			zbmat[nodeno] = value;
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;

				if (row == nodeno) {
					zkmat[locus] = duplex_unit;
					continue;
				}

				zp_mminus(&zbmat[row], zp_cross(&ztmp,
					&zkmat[locus], &zbmat[nodeno]));
				zkmat[locus] = duplex_null;
			}
			norows = sp_getpairs(rowarr, SP_FINDROWS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (row == nodeno)
					continue;
				zkmat[locus] = duplex_null;
			}
			break;
	}
	return;
}

/*
 * The private function of ddirichlet()
 */
void
di_dimpose(const int nodeno, const double value)
{
	int	i;
	float	*skmat, *sbmat;
	double	*dkmat, *dbmat;

	int	locus, row;
	struct sparse_t *sp;
	int	norows;
	struct pairs rowarr[50];	/* is this suffic?? */

	if (meshes->m_datatype != SP_SINGLE && meshes->m_datatype != SP_DOUBLE)
		fatal("%s():dont call ce func at SINGLE or DOUBLE world", __F);

	sp = meshes->m_kmatrix;

	switch (meshes->m_datatype) {
		case SP_SINGLE:
			skmat = (float *) sp->sp_data;
			sbmat = svec(meshes->m_bvector);
			sbmat[nodeno] = value;

			norows = sp_getpairs(rowarr, SP_FINDCOLS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (locus == -1)
					fatal("%s(): ", __F);

				if (row == nodeno) {
					skmat[locus] = 1.0;
					continue;
				}

				sbmat[row] -= skmat[locus] * sbmat[nodeno];
				skmat[locus] = 0.0;
			}
			norows = sp_getpairs(rowarr, SP_FINDROWS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (row == nodeno)
					continue;
				skmat[locus] = 0.0;
			}
			break;
		
		case SP_DOUBLE:
			dkmat = (double *) sp->sp_data;
			dbmat = dvec(meshes->m_bvector);
			dbmat[nodeno] = value;
			norows = sp_getpairs(rowarr, SP_FINDCOLS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (locus == -1)
					fatal("%s(): ", __F);

				if (row == nodeno) {
					dkmat[locus] = 1.0;
					continue;
				}

				dbmat[row] -= dkmat[locus] * dbmat[nodeno];
				dkmat[locus] = 0.0;
			}
			norows = sp_getpairs(rowarr, SP_FINDROWS, 50,
				sp, nodeno);
			for (i = 0; i < norows; i++) {
				row   = rowarr[i].pr_taxis;
				locus = rowarr[i].pr_locus;
				if (row == nodeno)
					continue;
				dkmat[locus] = 0.0;
			}
			break;
	}
	return;
}
