#include <stdio.h>
#include <math.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"

#define ON_CORNER	4	/* at corner */
#define ON_LEFTRIGHT	5	/* left or right side */
#define ON_UPDOWN 	6	/* upper or lower side */

static void     sq8_setunit(void);
static void	sq8_integpoints(void);
static void	sq8_mkhalfjacobm(void);
static void	sq8_mktrans(void);

struct integral_t *sq8_intg;
static int sq8_unitflag[8];

int
sq8_init(struct proto_t *p)
{
	int	*lseg;
	static int sq8_inited = 0;

	if (sq8_inited)
		return 0;

	sq8_inited++;
	sq8_intg = p->p_intg;

	/*
	 * ORDER is important
	 */
	sq8_setunit();
	sq8_integpoints();
	sq8_mkhalfjacobm();
	sq8_mktrans();

	lseg = p->p_lineseg;
	lseg[0] = 0; lseg[1] = 2; lseg[2] = 7; lseg[3] = 5; lseg[4] = 0;

	printf("%s(): YOU SHOULD NOT USE THE ELEMENT OF THIS TYPE %s\n",
		__F, p->p_name);

	return 0;
}
/*
 * Set eight surrounding nodes coordinates in the base frame
 *
 *        5----6----7 
 *        |    |    |
 *        |    |    | 
 *    ----3----+----4------>
 *        |    |    |
 *        |    |    | 
 *        0----1----2
 *
 */

static void
sq8_setunit(void)
{
	double	**mat;

	mat  = dmat(sq8_intg->intg_unit);

	mat[0][0] = -1.0; mat[0][1] = -1.0; sq8_unitflag[0] = ON_CORNER;
	mat[1][0] =  0.0; mat[1][1] = -1.0; sq8_unitflag[1] = ON_UPDOWN;
	mat[2][0] =  1.0; mat[2][1] = -1.0; sq8_unitflag[2] = ON_CORNER;

	mat[3][0] = -1.0; mat[3][1] =  0.0; sq8_unitflag[3] = ON_LEFTRIGHT;
	mat[4][0] =  1.0; mat[4][1] =  0.0; sq8_unitflag[4] = ON_LEFTRIGHT;

	mat[5][0] = -1.0; mat[5][1] =  1.0; sq8_unitflag[5] = ON_CORNER;
	mat[6][0] =  0.0; mat[6][1] =  1.0; sq8_unitflag[6] = ON_UPDOWN;
	mat[7][0] =  1.0; mat[7][1] =  1.0; sq8_unitflag[7] = ON_CORNER;

	return;
}

#if COMMENTED
void
sq8_integpoints2nd(void)
{
	int i, j, ip;
	double	scale, x, y;
	double	**mat;
	double	*weights, weightx, weighty;
	double	coor[4], wei[4], tmp0, tmp1;

	tmp0 = 0.3399810435;
	tmp1 = 0.8611363115;
	coor[0] = -1.0* tmp1; coor[1] = -1.0* tmp0;
	coor[2] = tmp0; coor[3] = tmp1;

	tmp0 = 0.6521451548;
	tmp1 = 0.3478548451;

	wei[0] = tmp1; wei[1] = tmp0;
	wei[2] = tmp0; wei[3] = tmp1;

	weights = sq8_intg->intg_weights;

	mat = dmat(sq8_intg->intg_intgcoors);
	for (i=0; i < 4; i++) {
		for (j=0;j < 4;j++) {
			ip = 4 * i + j;
			mat[ip][0] = coor[j]; mat[ip][1] = coor[i];
			weights[ip] = wei[i] * wei[j];
			printf("%8.5f ", weights[ip]);
		}
		printf("\n");
	}
	return;
}
#endif

void
sq8_integpoints(void)
{
	int i, j, ip;
	double	scale, x, y;
	double	**mat;
	double	*weights, weightx, weighty;

	scale = sqrt(3.0/5.0);
	weights = sq8_intg->intg_weights;

	mat = dmat(sq8_intg->intg_intgcoors);
	for (i=0; i < 3; i++) {
		y = (double) (i-1);

		weighty = 5.0/9.0;
		if (i == 1)
			weighty = 8.0/9.0;
		for (j=0;j < 3;j++) {
			ip = i * 3 + j;
			x = (double) (j-1);

			weightx = 5.0/9.0;
			if (j == 1)
				weightx = 8.0/9.0;

			mat[ip][0] = scale * x; mat[ip][1] = scale * y;
			weights[ip] = weightx * weighty;
		}
	}
	if (0)
		for (i=0; i< sq8_intg->intg_proto->p_intgno; i++)
			printf("[%d]: %7.5f\n", i, weights[i]);

	return;
}


static void
sq8_mkhalfjacobm(void)
{
	double	**mat;
	int	i;
	double **points;
	double	ksi, eta;

	points = dmat(sq8_intg->intg_intgcoors);
	for (i=0; i< sq8_intg->intg_proto->p_intgno; i++) {
		mat = dmat(sq8_intg->intg_halfjacobm[i]);
		ksi = points[i][0];	eta = points[i][1];

		mat[0][0]= -0.25*((1.0-eta)*(-1.0-ksi-eta)+(1.0-ksi)*(1.0-eta));
		mat[1][0]= -0.25*((1.0-ksi)*(-1.0-ksi-eta)+(1.0-ksi)*(1.0-eta));

		mat[0][1]= -ksi*(1.0-eta);
		mat[1][1]= -0.50*(1.0-ksi*ksi);

		mat[0][2]= +0.25*((1.0-eta)*(-1.0+ksi-eta)+(1.0+ksi)*(1.0-eta));
		mat[1][2]= -0.25*((1.0+ksi)*(-1.0+ksi-eta)+(1.0+ksi)*(1.0-eta));

		mat[0][3]= -0.50*(1.0-eta*eta);
		mat[1][3]= -eta*(1.0-ksi);

		mat[0][4]=  0.50*(1.0-eta*eta);
		mat[1][4]= -eta*(1.0+ksi);

		mat[0][5]= -0.25*((1.0+eta)*(-1.0-ksi+eta)+(1.0-ksi)*(1.0+eta));
		mat[1][5]=  0.25*((1.0-ksi)*(-1.0-ksi+eta)+(1.0-ksi)*(1.0+eta));

		mat[0][6]= -ksi*(1.0+eta);
		mat[1][6]= 0.50*(1.0-ksi*ksi);

		mat[0][7]=  0.25*((1.0+eta)*(-1.0+ksi+eta)+(1.0+ksi)*(1.0+eta));
		mat[1][7]=  0.25*((1.0+ksi)*(-1.0+ksi+eta)+(1.0+ksi)*(1.0+eta));
	}
	return;
}
		
static void
sq8_mktrans(void)
{
	double	**mat, **points;
	int	i;
	double	ksi, dksi, eta, deta;

	mat    = dmat(sq8_intg->intg_trans);
	points = dmat(sq8_intg->intg_intgcoors);

	for (i=0; i <  sq8_intg->intg_proto->p_intgno; i++) {
		ksi = points[i][0]; eta = points[i][1];
		dksi = ksi * ksi;  deta = eta * eta;

		mat[i][0] = 0.25*(1.0-ksi)*(1.0-eta)*(-1.0-ksi-eta);
		mat[i][2] = 0.25*(1.0+ksi)*(1.0-eta)*(-1.0+ksi-eta);
		mat[i][5] = 0.25*(1.0-ksi)*(1.0+eta)*(-1.0-ksi+eta);
		mat[i][7] = 0.25*(1.0+ksi)*(1.0+eta)*(-1.0+ksi+eta);

		mat[i][1] = 0.50*(1.0-dksi)*(1-eta);
		mat[i][6] = 0.50*(1.0-dksi)*(1+eta);

		mat[i][3] = 0.50*(1-ksi)*(1.0-deta);
		mat[i][4] = 0.50*(1+ksi)*(1.0-deta);

	}
	return;
}
