#include <stdio.h>
#include <sys/types.h>
#include <math.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"

static void     sq9_setunit(void);
static void	sq9_integpoints(void);
static void	sq9_mkhalfjacobm(void);
static void	sq9_mktrans(void);

struct integral_t *sq9_intg;

int
sq9_init(struct proto_t *p)
{
	int	*lseg;
	static int sq9_inited = 0;

	if (sq9_inited)
		return 0;

	sq9_inited++;
	sq9_intg = p->p_intg;

	/*
	 * ORDER is important
	 */
	sq9_setunit();
	sq9_integpoints();
	sq9_mkhalfjacobm();
	sq9_mktrans();

	lseg = p->p_lineseg;
	lseg[0] = 0; lseg[1] = 2; lseg[2] = 8; lseg[3] = 6; lseg[4] = 0;

	return 0;
}

static void
sq9_setunit(void)
{
	struct matrix_t *matf;
	double	**mat;

	matf = sq9_intg->intg_unit;
	mat  = dmat(matf);

	mat[0][0] = -1.0; mat[0][1] = -1.0;
	mat[1][0] =  0.0; mat[1][1] = -1.0;
	mat[2][0] =  1.0; mat[2][1] = -1.0;
	mat[3][0] = -1.0; mat[3][1] =  0.0;
	mat[4][0] =  0.0; mat[4][1] =  0.0;
	mat[5][0] =  1.0; mat[5][1] =  0.0;
	mat[6][0] = -1.0; mat[6][1] =  1.0;
	mat[7][0] =  0.0; mat[7][1] =  1.0;
	mat[8][0] =  1.0; mat[8][1] =  1.0;

	return;
}


void
sq9_integpoints(void)
{
	int i, j, ip;
	double	scale, x, y;
	double	**mat;
	double	*weights, weightx, weighty;

	scale = sqrt(3.0/5.0);
	weights = sq9_intg->intg_weights;

	mat = dmat(sq9_intg->intg_intgcoors);
	for (i=0; i < 3; i++) {
		y = (double) (i-1);

		weighty = 5.0/9.0;
		if (i == 1)
			weighty = 8.0/9.0;
		for (j=0;j < 3;j++) {
			ip = i * 3 + j;
			x = (double) (j-1);

			weightx = 5.0/9.0;
			if (j == 1)
				weightx = 8.0/9.0;

			mat[ip][0] = scale * x; mat[ip][1] = scale * y;
			weights[ip] = weightx * weighty;
		}
	}
	return;
}

static void
sq9_mktrans(void)
{
	double	**mat;
	double	*at_intp;
	double	ksi, dksi, ksi2, eta, deta, eta2;
	int	i;

	mat = dmat(sq9_intg->intg_trans);

	for (i=0; i < 9; i++) {	/* for four points */
		at_intp = dm_row(sq9_intg->intg_intgcoors, i);

		ksi  = at_intp[0]; dksi = 2.0 * ksi; ksi2 = ksi*ksi;
		eta  = at_intp[1]; deta = 2.0 * eta; eta2 = eta*eta;

		mat[i][0] = 0.25 * (ksi2-ksi)  * (eta2-eta);
		mat[i][1] = 0.50 * (1.0 -ksi2) * (eta2-eta);
		mat[i][2] = 0.25 * (ksi2+ksi)  * (eta2-eta);
		mat[i][3] = 0.50 * (ksi2-ksi)  * (1.0 -eta2);
		mat[i][4] =        (1.0 -ksi2) * (1.0 -eta2);
		mat[i][5] = 0.50 * (ksi2+ksi)  * (1.0 -eta2);
		mat[i][6] = 0.25 * (ksi2-ksi)  * (eta2+eta);
		mat[i][7] = 0.50 * (1.0 -ksi2) * (eta2+eta);
		mat[i][8] = 0.25 * (ksi2+ksi)  * (eta2+eta);

	}
	return;
}
		

static void
sq9_mkhalfjacobm(void)
{
	double	ksi, dksi,ksi2, eta, deta, eta2;
	double	*atframe;
	double	**matpp;
	int	i;

	for (i=0; i<9;i++) {
		matpp = dmat(sq9_intg->intg_halfjacobm[i]);

		atframe = dm_row(sq9_intg->intg_intgcoors, i);
		ksi  = atframe[0]; dksi = 2.0 * ksi; ksi2 = ksi*ksi;
		eta  = atframe[1]; deta = 2.0 * eta; eta2 = eta*eta;

		matpp[0][0] = 0.25 * (dksi-1.0) * (eta2-eta);
		matpp[1][0] = 0.25 * (ksi2-ksi) * (deta-1.0);

		matpp[0][1] = 0.50 * (-dksi)    * (eta2-eta);
		matpp[1][1] = 0.50 * (1.0-ksi2) * (deta-1.0);

		matpp[0][2] = 0.25 * (dksi+1.0) * (eta2-eta);
		matpp[1][2] = 0.25 * (ksi2+ksi) * (deta-1.0);


		matpp[0][3] = 0.50 * (dksi-1.0) * (1.0-eta2);
		matpp[1][3] = 0.50 * (ksi2-ksi) * (-deta);

		matpp[0][4] =             -dksi * (1.0-eta2);
		matpp[1][4] =        (1.0-ksi2) * (-deta);

		matpp[0][5] = 0.50 * (dksi+1.0) * (1.0-eta2);
		matpp[1][5] = 0.50 * (ksi2+ksi) * (-deta);


		matpp[0][6] = 0.25 * (dksi-1.0) * (eta2+eta);
		matpp[1][6] = 0.25 * (ksi2-ksi) * (deta+1.0);

		matpp[0][7] = 0.50 * (-dksi)    * (eta2+eta);
		matpp[1][7] = 0.50 * (1.0-ksi2) * (deta+1.0);

		matpp[0][8] = 0.25 * (dksi+1.0) * (eta2+eta);
		matpp[1][8] = 0.25 * (ksi2+ksi) * (deta+1.0);

	}
	return;
}
