#include <stdio.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"

#define UPPER 10
#define LOWER 20
/*
 * returns a sparse_t matrix which contains the diagonal of  sparse_t *sp
 */
struct sparse_t *
sp_getdiagsp(const struct sparse_t *sp)
{
	char	name[1024];
	int	i, j, found;
	float	*sdata, *sofsp;
	double	*ddata, *dofsp;
	struct simplex_t *cdata, *cofsp;
	struct duplex_t *zdata, *zofsp;
	struct sparse_t *retsp;

	if (sp == NULL || (IS_SPARSE(sp) == 0))
		fatal("%s(sp == %p): NULL pointer or not sparse_t", __F, sp);
	if (sp->sp_issquare == 0)
		fatal("%s(retsp, sp): sparse_t sp is not square types", __F);

	sprintf(name, "diag of \"%s\"", sp->sp_name);
	retsp = sparse(sp->sp_rown, sp->sp_coln, sp->sp_coln,
		name, sp->sp_type);

	for (i=0;i < sp->sp_coln; i++) {
		retsp->sp_colx[i] = i;
		retsp->sp_rowx[i] = i;
	}
	retsp->sp_colx[i] = i;
	SET_DIAGONAL(retsp);
	
	switch (sp->sp_type) {
		case SP_SINGLE:
			sdata = ssp(retsp);
			sofsp = ssp(sp);
			for (i=0;i < retsp->sp_coln; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					sdata[i] = sofsp[j];
				else
					sdata[i] = 0;
			}
			break;

		case SP_DOUBLE:
			ddata = dsp(retsp);
			dofsp = dsp(sp);
			for (i=0;i < retsp->sp_coln; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					ddata[i] = dofsp[j];
				else
					ddata[i] = 0.0;	/* XXX */
			}
			break;
		case SP_SIMPLEX:
			cdata = csp(retsp);
			cofsp = csp(sp);
			for (i=0;i < retsp->sp_coln; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					cdata[i] = cofsp[j];
				else
					cdata[i] = simplex_null;
			}
			break;
		case SP_DUPLEX:
			zdata = zsp(retsp);
			zofsp = zsp(sp);
			for (i=0;i < retsp->sp_coln; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					zdata[i] = zofsp[j];
				else
					zdata[i] = duplex_null;
			}
			break;
		default:
			fatal("%s(\"%s\"): IMPOSSIBLE sp->sp_type", __F,
				sp->sp_name);
	}
	return retsp;
}

struct sparse_t *
sp_getinvdiag(const struct sparse_t *sp)
{
	char	name[1024];
	int	i;
	float	*sdata, *sofsp;
	double	*ddata, *dofsp;
	struct simplex_t *cdata, *cofsp;
	struct duplex_t *zdata, *zofsp;
	struct sparse_t *retsp;

	if (sp == NULL || (IS_SPARSE(sp) == 0))
		fatal("%s(sp == %p): NULL pointer or not sparse_t", __F, sp);
	if (IS_DIAGONAL(sp) == 0)
		fatal("%s(sp): sparse_t sp is not diagonal form", __F);

	sprintf(name, "invdiag of <%s>", sp->sp_name);
	retsp = sparse(sp->sp_rown, sp->sp_coln, sp->sp_coln,
		name, sp->sp_type);

	for (i=0;i < sp->sp_coln; i++) {
		retsp->sp_colx[i] = i;
		retsp->sp_rowx[i] = i;
	}
	retsp->sp_colx[i] = i;
	SET_DIAGONAL(retsp);
	
	switch (sp->sp_type) {
		case SP_SINGLE:
			sdata = ssp(retsp);
			sofsp = ssp(sp);
			for (i=0;i < retsp->sp_coln; i++)
				sdata[i] = 1.0/sofsp[i];
			break;

		case SP_DOUBLE:
			ddata = dsp(retsp);
			dofsp = dsp(sp);
			for (i=0;i < retsp->sp_coln; i++)
				ddata[i] = 1.0/dofsp[i];
			break;
		case SP_SIMPLEX:
			cdata = csp(retsp);
			cofsp = csp(sp);
			for (i=0;i < retsp->sp_coln; i++)
				cp_inverse(&cdata[i], &cofsp[i]);
			break;
		case SP_DUPLEX:
			zdata = zsp(retsp);
			zofsp = zsp(sp);
			for (i=0;i < retsp->sp_coln; i++)
				zp_inverse(&zdata[i], &zofsp[i]);
			break;
		default:
			fatal("%s(\"%s\"): IMPOSSIBLE sp->sp_type", __F,
				sp->sp_name);
	}
	return retsp;
}

int
sp_getupperlowerno(const struct sparse_t *sp, int indicator)
{
	int	*colx, *rowx;
	int	cnt, i, j;

	if (indicator != UPPER && indicator != LOWER)
		fatal("%s(\"%s\"): 2nd argument should be 1 or 2",
			__F, sp->sp_name);
	cnt = 0;
	colx = sp->sp_colx;
	rowx = sp->sp_rowx;

	switch (indicator) {
		case UPPER:
			for (i=0; i < sp->sp_coln;i++)
				for (j=colx[i]; j < colx[i+1]; j++)
					if (rowx[j] < i)
						cnt++;
			break;
		case LOWER:
			for (i=0; i < sp->sp_coln;i++)
				for (j=colx[i]; j < colx[i+1]; j++)
					if (rowx[j] > i)
						cnt++;
			break;
	}
	return cnt;
}

struct sparse_t *
sp_getupper(const struct sparse_t *sp)
{
	char	name[1024];
	int	i, j;
	float	*sdata, *sofsp;
	double	*ddata, *dofsp;
	struct simplex_t *cdata, *cofsp;
	struct duplex_t *zdata, *zofsp;
	struct sparse_t *retsp;
	int	nonzero, cnt;
	int	*rowx, *colx;

	if (sp == NULL)
		fatal("%s(sp == %p): NULL pointer", __F, sp);
	if (sp->sp_issquare == 0)
		fatal("%s(retsp, sp): sparse_t sp is not square types", __F);

	nonzero = sp_getupperlowerno(sp, UPPER);
	sprintf(name, "upper of <%s>", sp->sp_name);
	retsp = sparse(sp->sp_rown, sp->sp_coln, nonzero,
		name, sp->sp_type);

	rowx = retsp->sp_rowx;
	colx = retsp->sp_colx;
	cnt  = 0;
	switch (sp->sp_type) {
		case SP_SINGLE:
			sdata = ssp(retsp);
			sofsp = ssp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] < i) {
						rowx[cnt]  = sp->sp_rowx[j];
						sdata[cnt] = sofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_DOUBLE:
			ddata = dsp(retsp);
			dofsp = dsp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] < i) {
						rowx[cnt]  = sp->sp_rowx[j];
						ddata[cnt] = dofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_SIMPLEX:
			cdata = csp(retsp);
			cofsp = csp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] < i) {
						rowx[cnt]  = sp->sp_rowx[j];
						cdata[cnt] = cofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_DUPLEX:
			zdata = zsp(retsp);
			zofsp = zsp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] < i) {
						rowx[cnt]  = sp->sp_rowx[j];
						zdata[cnt] = zofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
	}
	return retsp;
}

struct sparse_t *
sp_getlower(const struct sparse_t *sp)
{
	char	name[1024];
	int	i, j;
	float	*sdata, *sofsp;
	double	*ddata, *dofsp;
	struct simplex_t *cdata, *cofsp;
	struct duplex_t *zdata, *zofsp;
	struct sparse_t *retsp;
	int	nonzero, cnt;
	int	*rowx, *colx;

	if (sp == NULL)
		fatal("%s(sp == %p): NULL pointer", __F, sp);
	if (sp->sp_issquare == 0)
		fatal("%s(retsp, sp): sparse_t sp is not square types", __F);

	nonzero = sp_getupperlowerno(sp, LOWER);
	sprintf(name, "lower of <%s>", sp->sp_name);
	retsp = sparse(sp->sp_rown, sp->sp_coln, nonzero,
		name, sp->sp_type);

	rowx = retsp->sp_rowx;
	colx = retsp->sp_colx;
	cnt  = 0;
	switch (sp->sp_type) {
		case SP_SINGLE:
			sdata = ssp(retsp);
			sofsp = ssp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] > i) {
						rowx[cnt]  = sp->sp_rowx[j];
						sdata[cnt] = sofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_DOUBLE:
			ddata = dsp(retsp);
			dofsp = dsp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] > i) {
						rowx[cnt]  = sp->sp_rowx[j];
						ddata[cnt] = dofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_SIMPLEX:
			cdata = csp(retsp);
			cofsp = csp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] > i) {
						rowx[cnt]  = sp->sp_rowx[j];
						cdata[cnt] = cofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
		case SP_DUPLEX:
			zdata = zsp(retsp);
			zofsp = zsp(sp);
			for (i=0;i < sp->sp_coln; i++) {
				colx[i] = cnt;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] > i) {
						rowx[cnt]  = sp->sp_rowx[j];
						zdata[cnt] = zofsp[j];
						cnt++;
					}
				}
			}
			colx[i] = cnt;
			break;
	}
	return retsp;
}
