#include <stdio.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"

/*
 * Multiply a sparse_t with a vector
 * They should be same type.
 * These functions are private functions so that you should not call these
 * functions directly.
 */
void
spdotvec_bontriplets(const char *func, const struct vector_t *ap,
	const struct sparse_t *sp, const struct vector_t *bp)
{
	if (sp == NULL || ap == NULL || bp == NULL)
		fatal("%s()/%s(): one of the triplets..is NULL",
			func, __F);

	if (sp->sp_type != ap->vt_type || ap->vt_type != bp->vt_type)
		fatal("%s()/%s(): the types of the triplets..are not same",
			func, __F);
	if (sp->sp_coln != bp->vt_size)
		fatal("%s()/%s():sp->sp_coln != bp->vt_size", func, __F);

	if (ap->vt_size != sp->sp_rown)
		fatal("%s()/%s(): ap->vt_size != sp->sp_rown", func, __F);

	return;
}
void
sptransdotvec_bontriplets(const char *func, const struct vector_t *ap,
	const struct sparse_t *sp, const struct vector_t *bp)
{
	if (sp == NULL || ap == NULL || bp == NULL)
		fatal("%s()/%s(): one of the triplets..is NULL",
			func, __F);

	if (sp->sp_type != ap->vt_type || ap->vt_type != bp->vt_type)
		fatal("%s()/%s(): the types of the triplets..are not same",
			func, __F);
	if (sp->sp_rown != bp->vt_size)
		fatal("%s()/%s():sp->sp_coln != bp->vt_size", func, __F);

	if (ap->vt_size != sp->sp_coln)
		fatal("%s()/%s(): ap->vt_size != sp->sp_rown", func, __F);

	return;
}

struct vector_t *
sp_dotvec(struct vector_t *xp, const int indicator, struct sparse_t *sp,
		const struct vector_t *vp)
{
	int	i, j;
	int	*rowp, *colp, *datp;

	float	*sxp, *svp, *spp;
	double	*dxp, *dvp, *dpp;
	struct simplex_t *cxp, *cvp, *cpp, ctmp;
	struct duplex_t *zxp, *zvp, *zpp, ztmp;

	struct simplex_t *(*cp_dotfunc)(struct simplex_t *ret,
		const struct simplex_t *a, const struct simplex_t *b);
	struct duplex_t *(*zp_dotfunc)(struct duplex_t *ret,
		const struct duplex_t *a, const struct duplex_t *b);

	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	rowp = sp->sp_rowindex;
	colp = sp->sp_colindex;
	datp = sp->sp_dataindex;

	spdotvec_bontriplets(__F, xp, sp, vp);

	vec_null(xp);
	switch (xp->vt_type) {
		case SP_SINGLE:
			sxp = svec(xp);
			svp = svec(vp);
			spp = ssp(sp);

			for (i=0; i < vp->vt_size; i++)
				for (j = rowp[i]; j < rowp[i+1]; j++)
					sxp[i] += (spp[datp[j]] * svp[colp[j]]);
			break;
		case SP_DOUBLE:
			dxp = dvec(xp);
			dvp = dvec(vp);
			dpp = dsp(sp);

			for (i=0; i < vp->vt_size; i++)
				for (j = rowp[i]; j < rowp[i+1]; j++)
					dxp[i] += (dpp[datp[j]] * dvp[colp[j]]);
			break;
		case SP_SIMPLEX:
			cxp = cvec(xp);
			cvp = cvec(vp);
			cpp = csp(sp);

			if (indicator)
				cp_dotfunc = cp_conjcross;
			else
				cp_dotfunc = cp_cross;

			for (i=0; i < vp->vt_size; i++)
				for (j = rowp[i]; j < rowp[i+1]; j++)
					cp_pplus(&cxp[i], cp_dotfunc(&ctmp,
						&cpp[datp[j]],&cvp[colp[j]]));
			break;
		case SP_DUPLEX:
			zxp = zvec(xp);
			zvp = zvec(vp);
			zpp = zsp(sp);

			if (indicator)
				zp_dotfunc = zp_conjcross;
			else
				zp_dotfunc = zp_cross;
			for (i=0; i < vp->vt_size; i++)
				for (j = rowp[i]; j < rowp[i+1]; j++)
					zp_pplus(&zxp[i], zp_dotfunc(&ztmp,
						&zpp[datp[j]],&zvp[colp[j]]));
			break;
	}
	return  xp;
}

/*
 * indicator!!! XXX
 */
struct vector_t *
sp_transdotvec(struct vector_t *xp, const int indicator,
		struct sparse_t *sp, const struct vector_t *vp)
{
	int	i, j;
	int	*rowp, *colp;

	float	*sxp, *svp, *spp;
	double	*dxp, *dvp, *dpp;
	struct simplex_t *cxp, *cvp, *cpp, ctmp;
	struct duplex_t *zxp, *zvp, *zpp, ztmp;

	struct simplex_t *(*cp_dotfunc)(struct simplex_t *ret,
		const struct simplex_t *a, const struct simplex_t *b);
	struct duplex_t *(*zp_dotfunc)(struct duplex_t *ret,
		const struct duplex_t *a, const struct duplex_t *b);

	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	rowp = sp->sp_rowx;
	colp = sp->sp_colx;

	sptransdotvec_bontriplets(__F, xp, sp, vp);

	vec_null(xp);
	switch (sp->sp_type) {
		case SP_SINGLE:
			sxp = svec(xp);
			svp = svec(vp);
			spp = ssp(sp);

			for (i=0; i < sp->sp_coln; i++)
				for (j = colp[i]; j < colp[i+1]; j++)
					sxp[i] += (spp[j] * svp[rowp[j]]);
			break;
		case SP_DOUBLE:
			dxp = dvec(xp);
			dvp = dvec(vp);
			dpp = dsp(sp);

			for (i=0; i < sp->sp_coln; i++)
				for (j = colp[i]; j < colp[i+1]; j++)
					dxp[i] += (dpp[j] * dvp[rowp[j]]);
			break;
		case SP_SIMPLEX:
			cxp = cvec(xp);
			cvp = cvec(vp);
			cpp = csp(sp);

			if (indicator)
				cp_dotfunc = cp_conjcross;
			else
				cp_dotfunc = cp_cross;
			for (i=0; i < sp->sp_coln; i++)
				for (j = colp[i]; j < colp[i+1]; j++)
					cp_pplus(&cxp[i], cp_dotfunc(&ctmp,
						&cpp[j], &cvp[rowp[j]]));
			break;
		case SP_DUPLEX:
			zxp = zvec(xp);
			zvp = zvec(vp);
			zpp = zsp(sp);

			if (indicator)
				zp_dotfunc = zp_conjcross;
			else
				zp_dotfunc = zp_cross;
			for (i=0; i < sp->sp_coln; i++)
				for (j = colp[i]; j < colp[i+1]; j++)
					zp_pplus(&zxp[i], zp_dotfunc(&ztmp,
						&zpp[j], &zvp[rowp[j]]));
			break;
		default:
			fatal("sp_transdotvec(): IMPOSSIBLE");
	}
	return xp;
}
