#include <stdio.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"

#define sp_ops

/*
 * returns the diagonal vector of sparse_t *sp
 * You _MUST_ provide the space of xp first.
 */

struct vector_t *
sp_getdiagvec(struct vector_t *xp, const struct sparse_t *sp)
{
	int	i, j, found;
	float	*sdata, *sofsp;
	double	*ddata, *dofsp;
	struct simplex_t *cdata, *cofsp;
	struct duplex_t *zdata, *zofsp;

	if (sp == NULL || xp == NULL)
		fatal("%s(sp == %p, xp == %p): NULL pointer", __F, sp, xp);
	if (sp->sp_issquare == 0)
		fatal("%s(xp, sp): sparse_t sp is not square types", __F);
	if (!IS_VECTOR(xp))
		fatal("%s(xp, sp): xp is not vector_t", __F);
	if (sp->sp_type != xp->vt_type)
		fatal("%s(xp, sp): different types", __F);
	if (sp->sp_coln != xp->vt_size)
		fatal("%s(xp, sp): different sizes", __F);

	vec_null(xp);

	switch (sp->sp_type) {
		case SP_SINGLE:
			sdata = svec(xp);
			sofsp = ssp(sp);
			for (i=0;i<xp->vt_size; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					sdata[i] = sofsp[j];
				else
					sdata[i] = 0;
			}
			break;

		case SP_DOUBLE:
			ddata = dvec(xp);
			dofsp = dsp(sp);
			for (i=0;i<xp->vt_size; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					ddata[i] = dofsp[j];
				else
					ddata[i] = 0.0;	/* XXX */
			}
			break;
		case SP_SIMPLEX:
			cdata = cvec(xp);
			cofsp = csp(sp);
			for (i=0;i<xp->vt_size; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					cdata[i] = cofsp[j];
				else
					cdata[i] = simplex_null;
			}
			break;
		case SP_DUPLEX:
			zdata = zvec(xp);
			zofsp = zsp(sp);
			for (i=0;i<xp->vt_size; i++) {
				found = 0;
				for (j=sp->sp_colx[i];j<sp->sp_colx[i+1]; j++) {
					if (sp->sp_rowx[j] == i) {
						found++;
						break;
					}
				}
				if (found)
					zdata[i] = zofsp[j];
				else
					zdata[i] = duplex_null;
			}
			break;
		default:
			fatal("%s(\"%s\"): IMPOSSIBLE sp->sp_type", __F,
				sp->sp_name);
	}
	return xp;
}

void	sp_symmetric(struct sparse_t *sp);
int	sp_sympattern(struct sparse_t *sp);
int	sp_hermitian(struct sparse_t *sp);
void	sp_complexsym(struct sparse_t *sp);
void	sp_sym(struct sparse_t *sp);

void
sp_symcheck(struct sparse_t *sp)
{
	if (sp == NULL)
		fatal("%s(sp): sp == NULL", __F);
	if (!IS_SPARSE(sp))
		fatal("%s(sp): sp is not sparse_t pointer", __F);
	if (sp->sp_issquare == 0)
		return;
	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	if (sp_sympattern(sp))
		return;

	switch (sp->sp_type) {
		case SP_SINGLE:
		case SP_DOUBLE:
			sp_symmetric(sp);
			break;
		case SP_SIMPLEX:
		case SP_DUPLEX:
			if (sp_hermitian(sp)) /* if failed */
				sp_complexsym(sp);
			break;
	}
	sp_sym(sp);
	return;
}


/*
 * A private function of sp_symcheck()
 * You dont need to call this fucntion directly
 */
int
sp_sympattern(struct sparse_t *sp)
{
	int	size;
	int	ret;

	size = sizeof(int) * (sp->sp_coln+1);
	ret = memcmp(sp->sp_rowindex, sp->sp_colx, size);
	if (ret)
		return 1;

	size = sizeof(int) * sp->sp_nonzero;
	ret = memcmp(sp->sp_colindex, sp->sp_rowx, size);
	if (ret)
		return 1;

	sp->sp_symflag |= SP_SYMPATTERN;

	return 0;
}

/*
 * A private function of sp_symcheck()
 * I can assume that sp->sp_symflag & SP_SYMPATTERN != 0
 * You dont need to call this fucntion directly
 */
void
sp_symmetric(struct sparse_t *sp)
{
	int *colx, *rowx, *rowix, *colix, *datix;
	int	i, j, k, rowno;
	float	*sdata, *s1, *s2;
	double	*ddata, *d1, *d2;

	colx = sp->sp_colx;
	rowx = sp->sp_rowx;
	rowix = sp->sp_rowindex;
	colix = sp->sp_colindex;
	datix = sp->sp_dataindex;

	switch (sp->sp_type) {
	case SP_SINGLE:
		sdata = ssp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				s1 = &sdata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible",
						__F, sp->sp_name);
				s2 = &sdata[datix[k]];

				if (fabs(*s1 - *s2) > 1.0e-10)
					return;
			}
		}
		break;
	case SP_DOUBLE:
		ddata = dsp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				d1 = &ddata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible (%d,%d)",
						__F, sp->sp_name, rowno, i);
				d2 = &ddata[datix[k]];

				if (fabs(*d1 - *d2) > 1.0e-10)
					return;
			}
		}
		break;
	}
	sp->sp_symflag |= SP_SYMMETRIC;
	return;
}

/*
 * A private function of sp_symcheck()
 * You dont need to call this fucntion directly
 */
int
sp_hermitian(struct sparse_t *sp)
{
	int	i, j, k, rowno;
	int *colx, *rowx, *rowix, *colix, *datix;
	struct simplex_t *sdata, *s1, *s2;
	struct  duplex_t *ddata, *d1, *d2;

	colx = sp->sp_colx;
	rowx = sp->sp_rowx;
	rowix = sp->sp_rowindex;
	colix = sp->sp_colindex;
	datix = sp->sp_dataindex;

	switch (sp->sp_type) {
	case SP_SIMPLEX:
		sdata = csp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				s1 = &sdata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible",
						__F, sp->sp_name);
				s2 = &sdata[datix[k]];

				/* Take care of this relations */
				if (s1->real != s2->real
				    || s1->imag != -1.0 * s2->imag)
					return 1;
			}
		}
		break;
	case SP_DUPLEX:
		ddata = zsp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				d1 = &ddata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible",
						__F, sp->sp_name);
				d2 = &ddata[datix[k]];

				/* Take care of this relations */
				if (d1->real != d2->real
				    || d1->imag != -1.0 * d2->imag)
					return 1;
			}
		}
		break;
	}
	sp->sp_symflag |= SP_HERMITIAN;
	return 0;
}

/*
 * A private function of sp_symcheck()
 * You dont need to call this fucntion directly
 * In the space of complex, this function checks the symmetry property of
 * the sparse_t matrix.
 */
void
sp_complexsym(struct sparse_t *sp)
{
	int	i, j, k, rowno;
	int *colx, *rowx, *rowix, *colix, *datix;
	struct simplex_t *sdata, *s1, *s2;
	struct  duplex_t *ddata, *d1, *d2;

	colx = sp->sp_colx;
	rowx = sp->sp_rowx;
	rowix = sp->sp_rowindex;
	colix = sp->sp_colindex;
	datix = sp->sp_dataindex;

	switch (sp->sp_type) {
	case SP_SIMPLEX:
		sdata = csp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				if (rowno == i) /* dont need to check */
					continue;
				s1 = &sdata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible",
						__F, sp->sp_name);
				s2 = &sdata[datix[k]];

				/* Take care of this relations */
				if (s1->real != s2->real
				    || s1->imag != s2->imag)
					return;
			}
		}
		break;
	case SP_DUPLEX:
		ddata = zsp(sp);
		for (i=0;i<sp->sp_coln; i++) {
			for (j=colx[i]; j<colx[i+1]; j++) {
				rowno = rowx[j]; /* for this node */
				if (i < rowno)
					break;
				if (rowno == i) /* dont need to check */
					continue;
				d1 = &ddata[j];

				for (k=rowix[i]; k < rowix[i+1]; k++) {
					if (colix[k] == rowno)
						break;
				}
				if (k == rowix[i+1])
					fatal("%s(\"%s\"): impossible",
						__F, sp->sp_name);
				d2 = &ddata[datix[k]];

				/* Take care of this relations */
				if (d1->real != d2->real
				    || d1->imag != d2->imag)
					return;
			}
		}
		break;
	}
	sp->sp_symflag |= SP_COMPLEXSYM;
	return;
}

void
sp_sym(struct sparse_t *sp)
{
	if (sp->sp_symflag == 0) {
		printf("%s has NO SYMMETRY\n", sp->sp_name);
		return;
	}
	if (sp->sp_symflag & SP_SYMPATTERN)
		printf("%s has SP_SYMPATTERN property\n", sp->sp_name);
	if (sp->sp_symflag & SP_SYMMETRIC)
		printf("%s has SP_SYMMETRIC property\n", sp->sp_name);
	if (sp->sp_symflag & SP_COMPLEXSYM)
		printf("%s has SP_COMPLEXSYM property\n", sp->sp_name);
	if (sp->sp_symflag & SP_HERMITIAN)
		printf("%s has SP_HERMITIAN property\n", sp->sp_name);
	return;
}

void
sp_trancrossspable(const char *f,const struct sparse_t *a,
		const struct sparse_t *b)
{
	if (a == NULL || b == NULL)
		fatal("%s(a, b): with NULL a or NULL b", f);
	if (!IS_SPARSE(a))
		fatal("%s(\"%s\"), b):not arse_t type", f, a->sp_name);
	if (!IS_SPARSE(b))
		fatal("%s(a, \"%s\"):not arse_t type", f, b->sp_name);

	if (a->sp_type != b->sp_type)
		fatal("%s(\"%s\", \"%s\"): are different type", f,
			a->sp_name, b->sp_name);

	if (a->sp_rown != b->sp_rown) {
		fprintf(stderr, "%s(\"%s\", \"%s\"): are impossible", f, 
			a->sp_name, b->sp_name);
		fprintf(stderr, "as A->sp_coln != B->sp_rown");
		exit(1);
	}
	return;
}
