#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "common.h"
#include "complex.h"
#include "vector.h"
#include "sparse.h"

void spv_check(const char *function, const struct spvector_t *spv, int type);

/*
 * sparse VECTOR CREATION
 */
struct spvector_t *
spvector(int size, const char *name, int type)
{
	struct spvector_t *spv;
	int	allocing;

	if (size < 1)
		fatal("int vector size less than 1?");

	spv = (struct spvector_t *)
		malloc((size_t) (sizeof(struct spvector_t)));
	if (spv == NULL)
		memerr("vector(\"%s\"): mallocing spv");

	switch (type) {
		case SP_INTEGER:
			allocing = sizeof(int)*size;
			spv->spv_data = malloc(allocing);
			spv->spv_type = SP_INTEGER;
			break;
		case SP_SINGLE:
			allocing = sizeof(float)*size;
			spv->spv_data = malloc(allocing);
			spv->spv_type = SP_SINGLE;
			break;
		case SP_DOUBLE:
			allocing = sizeof(double)*size;
			spv->spv_data = malloc(allocing);
			spv->spv_type = SP_DOUBLE;
			break;
		case SP_SIMPLEX:
			allocing = sizeof(struct simplex_t)*size;
			spv->spv_data = malloc(allocing);
			spv->spv_type = SP_SIMPLEX;
			break;
		case SP_DUPLEX:
			allocing = sizeof(struct duplex_t)*size;
			spv->spv_data = malloc(allocing);
			spv->spv_type = SP_DUPLEX;
			break;
		default:
			fatal("%s(\"%s\"): wrong type %d",__F,  name, type);
	}
	spv->spv_flag = SPVECTOR_TYPE;
	spv->spv_size = size;
	spv->spv_index = (int *) malloc(size * sizeof(int));
	if (spv->spv_index == NULL)
		memerr("%s(): spv->spv_index == NULL", __F);

	/*
	 * Initialied as zero's
	 */

        if (spv->spv_data == NULL)
                memerr("%s(\"%s\"): [2]", __F, name ? name : UNKNOWN);

	memset((void *) spv->spv_data, 0x00, allocing); /* make all zeros */
	if (name && *name)
		spv->spv_name = strdup(name);
	else
		spv->spv_name = noname;

	if (spv->spv_name == NULL)
		memerr("%s():please spv->spv_name == NUL", __F);
	return spv;
}

void
spv_check(const char *function, const struct spvector_t *spv, const int type)
{
	if (spv == NULL)
		fatal("%s(): I received NULL spvector_t pointer", function);
	if (spv->spv_flag != SPVECTOR_TYPE)
		fatal("%s(): I received a pointer but this is not spvector_t",
			function);
	if (!ATYPE(type))
		return;	/* only checked whether this pointer is spvector_t */

	if (type == spv->spv_type)
		return;	/* type is matched */

	/*
	 * Now mismatched case
	 */
	switch (type) {
		case SP_INTEGER:
			fatal("%s(): spvector %s is not SP_INTEGER type",
				function, spv->spv_name);
			break;
		case SP_SINGLE:
			fatal("%s(): spvector %s is not SP_SINGLE type",
				function, spv->spv_name);
			break;
		case SP_DOUBLE:
			fatal("%s(): spvector %s is not SP_DOUBLE type",
				function, spv->spv_name);
			break;
		case SP_SIMPLEX:
			fatal("%s(): spvector %s is not SP_SIMPLEX type",
				function, spv->spv_name);
			break;
		case SP_DUPLEX:
			fatal("%s(): spvector %s is not SP_DUPLEX type",
				function, spv->spv_name);
			break;
	}
	return;
}

/*
 * DATA ACCESS ROUTINES
 */

inline float *
sspvec(const struct spvector_t *spv)
{
	spv_check(__F, spv, SP_SINGLE);
	return (float *) spv->spv_data;
}


inline double *
dspvec(const struct spvector_t *spv)
{
	spv_check(__F, spv, SP_DOUBLE);
	return (double *) spv->spv_data;
}

inline struct simplex_t *
cspvec(const struct spvector_t *spv)
{
	spv_check(__F, spv, SP_SIMPLEX);
	return (struct simplex_t *) spv->spv_data;
}

inline struct duplex_t *
zspvec(const struct spvector_t *spv)
{
	spv_check(__F, spv, SP_DUPLEX);
	return (struct duplex_t *) spv->spv_data;
}

void
spv_free(struct spvector_t *spv)
{
	free((void *) spv->spv_data);
	free((void *) spv->spv_index);
	free((void *) spv);
	return;
}

void
spv_mfree(int  num,...)
{
	int     i;
	va_list ap;
	struct spvector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct spvector_t *);
		spv_free(p);
	}
	va_end(ap);
	return;
}

#if NOT
void
spv_mprint(int  num,...)
{
	int     i;
	va_list ap;
	struct spvector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct spvector_t *);
		spv_print(p);
	}
	va_end(ap);
	return;
}

void
spv_print(const struct spvector_t *spv)
{
	int i;

	int	*ip;
	float	*sp;
	double	*dp;
	struct simplex_t *cp;
	struct duplex_t *zp;

	spv_check(__F, spv, -1);
	printf("--spv_print(%s)------\n", spv->spv_name);
	switch (spv->spv_type) {
		case SP_INTEGER:
			ip = ivec(spv);
			for (i=0;i < spv->spv_size; i++)
				(void) printf("%10d\n", ip[i]);
			break;
		case SP_SINGLE:
			sp = svec(spv);
			for (i=0;i < spv->spv_size; i++)
				(void) printf("%13.5e\n", sp[i]);
			break;
		case SP_DOUBLE:
			dp = dvec(spv);
			for (i=0;i < spv->spv_size; i++)
				(void) printf("%13.5e\n", dp[i]);
			break;
		case SP_SIMPLEX:
			cp = cvec(spv);
			for (i=0;i < spv->spv_size; i++)
				(void) printf("%13.5e %13.5e\n",
					cp[i].real, cp[i].imag);
			break;
		case SP_DUPLEX:
			zp = zvec(spv);
			for (i=0;i < spv->spv_size; i++)
				(void) printf("%13.5e %13.5e\n",
					zp[i].real, zp[i].imag);
			break;
		default:
			fatal("spv_print(): I dont understand your type");
			break;
	}
	return;
}

void
zspv_info(const struct spvector_t *spv)
{
	int	i;
	struct duplex_t *dux;

	spv_check(__F, spv, SP_DUPLEX);
	(void) printf("--zspv_info() duplex_t vector %s--\n", spv->spv_name);
	dux = zvec(spv);
	for (i=0;i < spv->spv_size; i++, dux++) {
			(void) printf("%7.3f %7.3f\n", dux->real, dux->imag);
	}
	(void) printf("+++\n");
	return;
}
#endif

void
spv_mnull(int  num,...)
{
	int     i;
	va_list ap;
	struct spvector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct spvector_t *);
		spv_null(p);
	}
	va_end(ap);
	return;
}

void
spv_null(struct spvector_t *spv)
{
	int	size;

	spv_check(__F, spv, -1);
	switch (spv->spv_type) {
		case SP_INTEGER:
			size = sizeof(int) * spv->spv_size; break;
		case SP_SINGLE:
			size = sizeof(float) * spv->spv_size; break;
		case SP_DOUBLE:
			size = sizeof(double) * spv->spv_size; break;
		case SP_SIMPLEX:
			size = sizeof(struct simplex_t) * spv->spv_size; break;
		case SP_DUPLEX:
			size = sizeof(struct duplex_t) * spv->spv_size; break;
	}
	memset((void *) spv->spv_data, 0x00, (size_t) size);
	return;
}
