#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "common.h"
#include "complex.h"
#include "sparse.h"

/*
 * creator of a sparse_t matrix. So if you want a sparse_t matrix, this function
 * is the most important one. const char *name is the name of the sparse matrix.
 * If the name == NULL, then noname was given to the sparse.
 */
 
struct sparse_t *
sparse(int row, int col, int nonzero, const char *name, int type)
{
	struct sparse_t *sp;
	size_t	allocing;

	if (!ATYPE(type))
		fatal("%s(): wrong sp_type for \"%s\"", __F, name);
	sp = (struct sparse_t *) malloc((size_t) sizeof(struct sparse_t));
	if (sp == NULL)
		memerr("%s(): sparse matrix %s creation when mallocing", __F,
			name ? name : UNKNOWN);

	if (name && *name)
		sp->sp_name = strdup(name);
	else
		sp->sp_name = noname;

	if (sp->sp_name == NULL)
		memerr("%s(): What's wrong!!", __F);


	switch (type) {
		case SP_SINGLE:
			allocing = sizeof(float)*nonzero; 
			sp->sp_data = malloc(allocing);
			sp->sp_type = SP_SINGLE;
			break;
		case SP_DOUBLE:
			allocing = sizeof(double)*nonzero; 
			sp->sp_data = malloc(allocing);
			sp->sp_type = SP_DOUBLE;
			break;
		case SP_SIMPLEX:
			allocing = sizeof(struct simplex_t)*nonzero; 
			sp->sp_data = malloc(allocing);
			sp->sp_type = SP_SIMPLEX;
			break;
		case SP_DUPLEX:
			allocing = sizeof(struct duplex_t)*nonzero; 
			sp->sp_data = malloc(allocing);
			sp->sp_type = SP_DUPLEX;
			break;
		default:
			fatal("%s(%s): wrong type %d", __F, sp->sp_name, type);
	}

	if (sp->sp_data == NULL)
		memerr("%s(): [2] %s", __F, sp->sp_name);

	memset((void *) sp->sp_data, 0x00, allocing);

	sp->sp_colx = (int *)malloc((size_t)(sizeof(int)* (col+1)));
	if (sp->sp_colx == NULL)
		memerr("%s(): [4] %s", __F, sp->sp_name);

	sp->sp_rowx = (int *)malloc((size_t)(sizeof(int)*(nonzero)));
	if (sp->sp_rowx == NULL)
		memerr("%s(): [3] %s", __F, sp->sp_name);

	sp->sp_coln = col;
	sp->sp_nonzero = nonzero;
	sp->sp_rown = row;
	SET_SPARSE(sp);	/* indicates that this sp pointer is sparse_t */
	sp->sp_roworder = NULL; /* will be alloced when necessary */
	sp->sp_symflag = SP_NOSYM;
	if (sp->sp_coln == sp->sp_rown)
		sp->sp_issquare = 1;
	else
		sp->sp_issquare = 0;

	allocing= sizeof(struct forprecond);
	sp->sp_precond = (struct forprecond *) malloc(allocing);
	if (sp->sp_precond == NULL)
		memerr("%s(\"%s\"): at sp->sp_precond", __F, sp->sp_name);

	sp->sp_diag	= NULL;
	sp->sp_diaginv	= NULL;
	sp->sp_upper	= NULL;
	sp->sp_upperinv	= NULL;
	sp->sp_lower	= NULL;
	sp->sp_lowerinv	= NULL;
	return sp;
}

void
sp_check(const char *function, const int arg, const struct sparse_t *sp)
{
	char	*p;

	p = get_ordinal(arg);

	if (sp == NULL)
		fatal("%s(): %th argument was NULL pointer!!\n", function, p);

	if (IS_SPARSE(sp) == 0)
		fatal("%s(): Even though %s argument was not NULL pointer!\n"
			"that argument was not sparse_t type!!\n", function, p);
	if (ATYPE(sp->sp_type) == 0)
		fatal("%s(): Even though %s argument was sparse_t pointer!\n"
			"the data type of that argument was not correct one!\n",
				function, p);
	return;
}

void
sp_free(struct sparse_t *sp)
{
	sp_check(__F, 1, sp);

	if (sp->sp_roworder) {
		free(sp->sp_rowindex);
		free(sp->sp_colindex);
		free(sp->sp_dataindex);
	}

	sp->sp_roworder = NULL;
	free(sp->sp_rowx);
	free(sp->sp_colx);
	free(sp->sp_data);
	if (strcmp(sp->sp_name, noname))
		free(sp->sp_name);

	free(sp);
	return;
}

void
sp_mfree(int num,...)
{
	int     i;
	va_list ap;
	struct sparse_t *sp;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		sp = va_arg(ap, struct sparse_t *);
		sp_free(sp);
	}
	va_end(ap);
	return;
}


void
sp_pattern(const struct sparse_t *sp)
{
	int	i, j, k;
	int	start, end;
	int	exists;

	sp_check(__F, 1, sp);

	fprintf(stderr, "++++sp_pattern of \"%s\"\n", sp->sp_name);
	for (i=0;i < sp->sp_rown; i++) {
		for (j=0;j < sp->sp_coln; j++) {
			start = sp->sp_colx[j];
			end   = sp->sp_colx[j+1];
			exists = 0;
			for (k=start; k< end; k++) {
				if (sp->sp_rowx[k] == i) {
					exists = 1;
					break;
				}
			}
			printf("%c", exists ? '+' : '.');
		}
		printf("\n");
	}
	printf("------\n");
	return;
}

void
sp_pattern2nd(struct sparse_t *sp)
{
	int	i, j;
	int	*rowix, *colix, *cx;

	sp_check(__F, 1, sp);

	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	rowix = sp->sp_rowindex;
	colix = sp->sp_colindex;

	printf("++++sp_pattern of \"%s\"\n", sp->sp_name);
	for (i=0;i < sp->sp_rown; i++) {
		for (j = 0; j < colix[rowix[i]]; j++)
			printf(".");
		cx = &colix[rowix[i]];
		if (rowix[i+1] - rowix[i])
			for (j = colix[rowix[i]];j <= colix[rowix[i+1]-1]; j++)
				if (*cx == j) {
					cx++;
					printf("+");
				} else
					printf(".");
				

		for (; j < sp->sp_coln; j++)
			printf(".");
		printf("\n");
	}
	printf("------\n");
	return;
}

/*
 * for data access routines
 */

float *
ssp(const struct sparse_t *sp)
{
	if (sp->sp_type != SP_SINGLE)
		fatal("%s(%s):isn't SP_SINGLE sparse_t", __F, sp->sp_name);
	return (float *) sp->sp_data;
}

double *
dsp(const struct sparse_t *sp)
{
	if (sp->sp_type != SP_DOUBLE)
		fatal("%s(%s):isn't SP_DOUBLE sparse_t", __F, sp->sp_name);
	return (double *) sp->sp_data;
}
struct simplex_t *
csp(const struct sparse_t *sp)
{
	if (sp->sp_type != SP_SIMPLEX)
		fatal("%s(%s):isn't SP_SIMPLEX sparse_t", __F, sp->sp_name);
	return (struct simplex_t *) sp->sp_data;
}

struct duplex_t  *
zsp(const struct sparse_t *sp)
{
	if (sp->sp_type != SP_DUPLEX)
		fatal("%s(%s):isn't SP_DUPLEX sparse_t", __F, sp->sp_name);
	return (struct duplex_t  *) sp->sp_data;
}


int
sp_rowcmp(const struct detail *first, const struct detail *second)
{
	struct detail one, two;
	one = (struct detail) *first;
	two = (struct detail) *second;

	if (one.d_whatrow >= two.d_whatrow)
		return -1;
	else
		return  1;
}

int
sp_printatfile(const struct sparse_t *sp, const char *fname)
{
	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t *zdata;

	int start, end;
	int	i, j, k;
	int	exists, atleast;
	FILE	*fp;
	int	type;

	sp_check(__F, 1, sp);

	fp = fopen(fname, "w");
	if (fp == NULL)
		fatal("%s(\"%s\", \"%s\"): can not fopen()", __F,
			sp->sp_name, fname);

	sdata = (float *) sp->sp_data;
	ddata = (double *) sp->sp_data;
	cdata = (struct simplex_t *) sp->sp_data;
	zdata = (struct duplex_t *) sp->sp_data;
	type = sp->sp_type;

	for (i=0;i<sp->sp_rown;i++) {
		atleast = 0;
		for (j=0;j<sp->sp_coln; j++) {
			start = sp->sp_colx[j];
			end   = sp->sp_colx[j+1];
			exists = 0;
			for (k=start; k< end;k++) {
				if (sp->sp_rowx[k] == i) {
					exists = 1;
					atleast++;
					break;
				}
			}
			if (exists) {
				switch (type) {
				case SP_SINGLE:
					fprintf(fp, "%6d %6d %10.6e\n", i, j,
						sdata[k]);
					break;
				case SP_DOUBLE:
					fprintf(fp, "%6d %6d %10.6e\n", i, j,
						ddata[k]);
					break;
				case SP_SIMPLEX:
					fprintf(fp, "%6d %6d %10.6e %10.6e\n",
						i, j,
						cdata[k].real, cdata[k].imag);
					break;
				case SP_DUPLEX:
					fprintf(fp, "%6d %6d %10.6e %10.6e\n",
						i, j,
						zdata[k].real, zdata[k].imag);
					break;
				}
			}
		}
		if (atleast > 1)
			fprintf(fp, "\n");
	}
	fclose(fp);
	return 0;
}

int
sp_print(const struct sparse_t *sp)
{
	int start, end;
	int	i, j, k;
	int	exists, atleast;

	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t *zdata;

	sp_check(__F, 1, sp);

	sdata = (float *) sp->sp_data;
	ddata = (double *) sp->sp_data;
	cdata = (struct simplex_t *) sp->sp_data;
	zdata = (struct duplex_t *) sp->sp_data;

	printf("++++sp_print of \"%s\"\n", sp->sp_name);
	for (i=0;i<sp->sp_rown;i++) {
		atleast = 0;
		for (j=0;j<sp->sp_coln; j++) {
			start = sp->sp_colx[j];
			end   = sp->sp_colx[j+1];
			exists = 0;
			for (k=start; k< end;k++) {
				if (sp->sp_rowx[k] == i) {
					exists = 1;
					atleast++;
					break;
				}
			}
			if (exists) {
				printf("%2d %2d ", i, j);
				switch (sp->sp_type) {
				case SP_SINGLE:
					printf("%12.5e\n", sdata[k]);
					break;
				case SP_DOUBLE:
					printf("%12.5e\n", ddata[k]);
					break;
				case SP_SIMPLEX:
					printf("%12.5e %12.5e\n",
						cdata[k].real, cdata[k].imag);
					break;
				case SP_DUPLEX:
					printf("%12.5e %12.5e\n",
						zdata[k].real, zdata[k].imag);
					break;
				}
			}
		}
#if 0
		if (atleast > 1)
			printf("\n");
#endif
	}
	printf("------\n");
	return 0;
}


struct sparse_t *
sp_read(const char *fname)
{
	FILE	*fp;
	char	*p, line[128], typename[32];
	struct sparse_t *sp;
	int	tmp;
	int	row, col, size, nonzero;

	if (fname == NULL || *fname == 0x00)
		fatal("%s(): Wrong fname", __F);

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): Can't fopen(\"%s\")", __F, fname);

	fgets(line, 128, fp);
	fclose(fp);

	tmp = sscanf(line, "%d %d %d %s",
		&row, &col, &nonzero, (char *) typename);
	p = (char *) typename;
	if (tmp == 4) {
		switch (*p) {
			case 's': sp = sp_sread(fname); break;
			case 'd': sp = sp_dread(fname); break;
			case 'c': sp = sp_cread(fname); break;
			case 'z': sp = sp_zread(fname); break;
			default:
			fatal("%s(\"%s\"): type should be one of  s, d, c, z",
					__F, fname);
		}
		return sp;
	}

	tmp = sscanf(line, "%d %s", &size, (char *) typename);
	if (tmp == 2) {
		switch (*p) {
			case 's': sp = sp_sreadraw(fname); break;
			case 'd': sp = sp_dreadraw(fname); break;
			case 'c': sp = sp_creadraw(fname); break;
			case 'z': sp = sp_zreadraw(fname); break;
			default:
			fatal("%s(\"%s\"): type should be one of  s, d, c, z",
					__F, fname);
		}
		return sp;
	}
	fatal("%s(\"%s\"): [3]", __F, fname);
	
	return NULL;
}

/*
 * A debugging routine
 */
void
sp_info(const struct sparse_t *sp)
{
	int	*ip;
	int	i;

	printf("---SP_INFO of \"%s\"--\nCOLX : ", sp->sp_name);
	sp_check(__F, 1, sp);
	
	ip = sp->sp_colx;
	for (i=0;i<=sp->sp_coln;i++) {
		printf("%2d ", ip[i]);
	}
	printf("\nROWX : ");

	ip = sp->sp_rowx;
	for (i=0;i<=sp->sp_nonzero;i++) {
		printf("%2d ", ip[i]);
	}

	printf("\n");
	return;
}
void
sp_briefinfo(const struct sparse_t *sp)
{
	printf("name   =%s\n", sp->sp_name);
	printf("coln   =%5d\n", sp->sp_coln);
	printf("rown   =%5d\n", sp->sp_rown);
	printf("nozero =%5d\n", sp->sp_nonzero);
	return;
}

void
sp_null(struct sparse_t *sp)
{
	int	size;

	sp_check(__F, 1, sp);
	switch (sp->sp_type) {
	case SP_SINGLE: size = sp->sp_nonzero * sizeof(float); break;
	case SP_DOUBLE: size = sp->sp_nonzero * sizeof(double); break;
	case SP_SIMPLEX:size = sp->sp_nonzero * sizeof(struct simplex_t); break;
	case SP_DUPLEX: size = sp->sp_nonzero * sizeof(struct duplex_t); break;
	}
	memset(sp->sp_data, 0x00, size);
	return;
}


void
sp_freerowindex(struct sparse_t *sp)
{
	sp_check(__F, 1, sp);
	if (sp->sp_rowindex == NULL)
		return;
	free(sp->sp_rowindex);
	free(sp->sp_colindex);
	free(sp->sp_dataindex);
	free(sp->sp_roworder);
	sp->sp_roworder = NULL;

	return;
}

/*
 * print using row index information..
 * test routine
 */

void
sp_putusingrow(struct sparse_t *sp, const char *patt, const char *term)
{
	int	start, end, probe, dinx;
	int	i, j, onoff;
	int	*rowp, *colp, *datax;
	int	oneof;
	char	patterns[64], spaces[64];

	float	*sdata;
	double	*ddata;
	struct duplex_t *zdata;
	struct simplex_t *cdata;

	sp_check(__F, 1, sp);

	if (sp->sp_roworder == NULL)
		sp_mkroworder(sp);

	rowp = sp->sp_rowindex;
	colp = sp->sp_colindex;
	datax = sp->sp_dataindex;
	printf("++++sp_putusingrow of <\"%s\">\n", sp->sp_name);

	sdata = (float *) sp->sp_data;
	ddata = (double *) sp->sp_data;
	cdata = (struct simplex_t *) sp->sp_data;
	zdata = (struct duplex_t *) sp->sp_data;

	oneof = 1;
	switch (sp->sp_type) {
	case SP_SINGLE:
		oneof = 0;
	case SP_DOUBLE:
		patt_rgetpatt(patterns, spaces, patt, term);
		for (i=0;i<sp->sp_rown;i++) {
			start = rowp[i];	/* col starts */
			end   = rowp[i+1];	/* col ends */
			if (start == end)
				onoff = 0;
			else
				onoff = 1;
			for (j=0, probe = start;j<sp->sp_coln; j++) {
				if (j == colp[probe] && onoff) {/* found */
					dinx = datax[probe];
					if (++probe == end)
						onoff = 0;
					if (oneof)
						printf(patterns, ddata[dinx]);
					else
						printf(patterns, sdata[dinx]);
					continue;
				}
				printf(spaces);
			}
			printf("\n");
		}
		break;
	case SP_SIMPLEX:
		oneof = 0;
	case SP_DUPLEX:
		patt_cgetpatt(patterns, spaces, patt, term);
		for (i=0;i<sp->sp_rown;i++) {
			start = rowp[i];	/* col starts */
			end   = rowp[i+1];	/* col ends */
			if (start == end)
				onoff = 0;
			else
				onoff = 1;
			for (j=0, probe = start;j<sp->sp_coln; j++) {
				if (j == colp[probe] && onoff) {/* found */
					dinx = datax[probe];
					if (++probe == end)
						onoff = 0;
					if (oneof)
						printf(patterns,
							zdata[dinx].real,
							zdata[dinx].imag);
					else
						printf(patterns,
							cdata[dinx].real,
							cdata[dinx].imag);
					continue;
				}
				printf(spaces);
			}
			printf("\n");
		}
	}
	return;
}


/*
 * mulifly a sparse_t with another sparse_t
 */
struct sparse_t *
sp_crosssp(struct sparse_t *spa, struct sparse_t *spb)
{
	struct sparse_t *retsp;
	sp_check(__F, 1, spa);
	sp_check(__F, 2, spb);
	sp_crossspable(__F, spa, spb);

	switch (spa->sp_type) {
		case SP_SINGLE:  retsp = sp_scrosssp(spa, spb); break;
		case SP_DOUBLE:  retsp = sp_dcrosssp(spa, spb); break;
		case SP_SIMPLEX: retsp = sp_ccrosssp(spa, spb); break;
		case SP_DUPLEX:  retsp = sp_zcrosssp(spa, spb); break;
	}
	return retsp;
}

void
sp_crossspable(const char *f, const struct sparse_t *a,const struct sparse_t *b)
{
	if (a->sp_type != b->sp_type)
		fatal("%s(%s, %s): are different type", f,
			a->sp_name, b->sp_name);

	if (a->sp_coln != b->sp_rown) {
		fatal("%s(%s, %s): cannot multiply", f, a->sp_name, b->sp_name);
	}
	return;
}

/*
 * makes another sparse_t matrix whose entries are same that of the given
 * sparse_t matrix.
 */

struct sparse_t *
sp_copy(const struct sparse_t *sp)
{
	struct sparse_t *retsp;
	char	name[128];
	int	size;

	sp_check(__F, 1, sp);

	sprintf(name, "copied from %s", sp->sp_name);
	retsp = sparse(sp->sp_rown, sp->sp_coln,
		sp->sp_nonzero, name, sp->sp_type);

	switch (sp->sp_type) {
		case SP_SINGLE:
			size = sp->sp_nonzero * sizeof(float); break;
		case SP_DOUBLE:
			size = sp->sp_nonzero * sizeof(double); break;
		case SP_SIMPLEX:
			size = sp->sp_nonzero * sizeof(struct simplex_t); break;
		case SP_DUPLEX:
			size = sp->sp_nonzero * sizeof(struct duplex_t); break;
	}

	memcpy(retsp->sp_data, sp->sp_data, size);
	memcpy(retsp->sp_colx, sp->sp_colx, sizeof(int) * (sp->sp_coln+1));
	memcpy(retsp->sp_rowx, sp->sp_rowx, sizeof(int) * (sp->sp_nonzero));

	if (sp->sp_roworder) {
		sp_rowindexstorage(retsp);
		memcpy(retsp->sp_rowindex, sp->sp_rowindex,
			sizeof(int) * (sp->sp_rown+1));
		memcpy(retsp->sp_colindex, sp->sp_colindex,
			sizeof(int) * (sp->sp_nonzero));
		memcpy(retsp->sp_dataindex, sp->sp_dataindex,
			sizeof(int) * (sp->sp_nonzero));
	}
	return retsp;
}



/*
 * returns the location of the data whose row and col are specified at argument
 * if there is no entry.. just returns -1
 */

int
sp_getwhereis(const struct sparse_t *sp, const int rowno, const int colno)
{
	int	i;

	sp_check(__F, 1, sp);
	if (sp->sp_rown <= rowno || sp->sp_coln <= colno)
		fatal("%s(\"%s\", %d, %d): request of out of range", __F,
			sp->sp_name, rowno, colno);

	for (i = sp->sp_colx[colno]; i < sp->sp_colx[colno+1]; i++)
		if (sp->sp_rowx[i] == rowno)
			break;

	if (i == sp->sp_colx[colno+1])
		return -1;
	return i;
}

/*
 * returns the pairs[] of indexes of rows which is corressponding to a specified
 * column col. You should provid space before calling this function.
 * for an example, see di_dimpose, etc.
 */
int sp_getpairs(struct pairs provided[], const int indicator, const int alloced,
	struct sparse_t *sp, const int col)
{
	int	i, cnt;

	if (indicator != SP_FINDCOLS && indicator != SP_FINDROWS)
		fatal("%s(): 2nd argument shoube be SP_FINDCOLS or etc", __F);
	cnt = 0;
	switch (indicator) {
	case SP_FINDCOLS:
		if (sp->sp_colx[col+1] - sp->sp_colx[col] > alloced)
			fatal("%s():You should have provided enough space",__F);
		for (i = sp->sp_colx[col]; i < sp->sp_colx[col+1]; i++, cnt++) {
			provided[cnt].pr_taxis = sp->sp_rowx[i];
			provided[cnt].pr_locus = i;
		}
		break;
	case SP_FINDROWS:
		if (sp->sp_roworder == NULL)
			sp_mkroworder(sp);

		if (sp->sp_rowindex[col+1] - sp->sp_rowindex[col] > alloced)
			fatal("%s():You should have provided enough space",__F);
		for (i=sp->sp_rowindex[col];i<sp->sp_rowindex[col+1];i++,cnt++){
			provided[cnt].pr_taxis = sp->sp_colindex[i];
			provided[cnt].pr_locus = sp->sp_dataindex[i];
		}
		break;
	}
	return cnt;
}

/*
 * A routine to used at qsort in sp_mkroworder():
 * Therefore you will not given any chance to use this function.
 */

struct datas {
	int	da_rowno;
	int	da_colno;
	int	da_posit;
};

int
data_compare(struct datas **onep, struct datas **twop)
{
	struct datas *one, *two;

	one = *onep;
	two = *twop;
	if (one->da_rowno != two->da_rowno)
		fatal("%s() of qsort of sp_mkroworder()", __F);
	if (one->da_colno > two->da_colno)
		return 1;
	else if (one->da_colno < two->da_colno)
		return -1;

	fatal("%s()of qsort of sp_mkroworder():IMPOSSIBLE!! SAME ENTRY",__F);

	/* NOTREACHED */	
	return 100;
}

void
sp_rowindexstorage(struct sparse_t *sp)
{
	sp_check(__F, 1, sp);
	if (sp->sp_roworder != NULL) {
		return;
	}

	sp->sp_roworder = (struct sp_roworder *)
		malloc((size_t) (sizeof(struct sp_roworder)));
	if (sp->sp_roworder == NULL)
		memerr("%s(\"%s\"): allocing sp->sp_roworder",__F, sp->sp_name);

	sp->sp_rowindex = (int *) malloc(((sp->sp_rown + 1)* sizeof(int)));
	sp->sp_colindex = (int *) malloc(((sp->sp_nonzero)* sizeof(int)));
	sp->sp_dataindex = (int *) malloc(((sp->sp_nonzero)* sizeof(int)));

	if (sp->sp_rowindex == NULL || sp->sp_colindex == NULL
	    || sp->sp_dataindex == NULL)
		memerr("%s(\"%s\"): allocing spr_..'s", __F, sp->sp_name);
	return;
}

/*
 * makes indexes of row-oriented.  This information is important because this
 * information is used at the time of multiplying a sparse_t matrix with a
 * a vector, or multiplying a sparse_t matrix with another sparse_t matrix.
 */
void
sp_mkroworder(struct sparse_t *sp)
{
	struct datas **datroot, *datp;

	struct mkrow {
		int	mkr_rowno;
		int	mkr_cnt;
		int	mkr_malloced;
		struct datas  **mkr_datas;
	};
	struct mkrow **mkroot, *mkp;	/* XXX */

	int	i, j;
	int	*rowp, *colp, *dataindexp;
	int	*oldrp, *oldcp;
	int	cnt;

	int (*compar)();

	compar = data_compare;

	sp_check(__F, 1, sp);
	if (sp->sp_roworder != NULL) {
		return;
	}

	sp_rowindexstorage(sp);

	oldcp = sp->sp_colx;
	oldrp = sp->sp_rowx;

	/*
	 * NOW malloc temporay allocation to make...sorting easier..
	 */
	datroot = (struct datas **)
		malloc(sizeof(struct datas *) * sp->sp_nonzero);
	if (datroot == NULL)
		memerr("%s(\"%s\"): malloc datroot", __F, sp->sp_name);

	datp = (struct datas *) malloc(sizeof(struct datas) * sp->sp_nonzero);
	if (datp == NULL)
		memerr("%s(\"%s\"): malloc datp", __F, sp->sp_name);

	for (i=0; i<sp->sp_nonzero; i++)	/* for each data */
		datroot[i] = datp++;

	/*
	 * for each row, there is one struct mkrow * structure
	 * thease structures are used to sort the colno..
	 */
	mkroot = (struct mkrow **) malloc(sizeof(struct mkrow *)*sp->sp_rown);
	if (mkroot == NULL)
		memerr("%s(\"%s\"): malloc mkroot", __F, sp->sp_name);

	mkp = (struct mkrow *) malloc(sizeof(struct mkrow) * sp->sp_rown);
	if (mkp == NULL)
		memerr("%s(\"%s\"): malloc mkp", __F, sp->sp_name);

	for (i=0; i<sp->sp_rown; i++, mkp++) { /* for each row */
		mkroot[i] = mkp;
		mkp->mkr_rowno = i;
		mkp->mkr_cnt = 0;	/* there is no datas */
		mkp->mkr_malloced = 30;	/* for FEM, 30 is..meadium */
		mkp->mkr_datas = (struct datas **)
			malloc(sizeof(struct datas *) * mkp->mkr_malloced);
	}

	/*
	 * end of the temporary allocations
	 * Now, let's scan the all datas;
	 */

	for (i=0; i < sp->sp_coln; i++) {
		for (j = oldcp[i]; j < oldcp[i+1]; j++) {
			datp = datroot[j];
			datp->da_colno = i;
			datp->da_rowno = oldrp[j];	/* row no */
			datp->da_posit = j;

			/* Now let's make.. row.. data */

			mkp = mkroot[oldrp[j]];
			if (mkp->mkr_cnt == mkp->mkr_malloced) {
				mkp->mkr_malloced += 30;
				mkp->mkr_datas = realloc(mkp->mkr_datas,
					sizeof(struct datas *)
					* mkp->mkr_malloced);
			}
			mkp->mkr_datas[mkp->mkr_cnt] = datp;
			mkp->mkr_cnt++;
		}
	}

	/*
	 * Let's sort the mkroot[i]->mkr_datas;
	 */

	cnt = 0;
	rowp = sp->sp_rowindex;
	colp = sp->sp_colindex;
	dataindexp = sp->sp_dataindex;

	for (i=0; i < sp->sp_rown; i++) { /* for each row */
		*rowp++ = cnt;
		mkp = mkroot[i];
		qsort(mkp->mkr_datas, mkp->mkr_cnt,
			(size_t)sizeof(struct datas *), compar);

		for (j=0; j < mkp->mkr_cnt; j++) {
			datp = mkp->mkr_datas[j];
			*colp++ = datp->da_colno;
			*dataindexp++ = datp->da_posit;
		}
		cnt += mkp->mkr_cnt;

	}
	*rowp++ = cnt;

	if (cnt != sp->sp_nonzero)
		fatal("%s(): IMPOSSIBLE!! cnt != sp->sp_nonzero", __F);

	/*
	 * All is over
	 * Let's free the temposary allocations
	 */

	for (i=0; i< sp->sp_rown; i++)
		free(mkroot[i]->mkr_datas);

	free((void *) mkroot[0]);
	free((void *) mkroot);
	free((void *) datroot[0]);
	free((void *) datroot);

	return;
}
