#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <values.h>	/* for MAXLONG, MAXFLOAT, MAXDOUBLE, etc */
#include <math.h>

#include "common.h"
#include "vector.h"
#define vec_ops

/*
 * Check to see whether or not the three vector_t *ap, *bp, *cp are really
 * vector_t pointer and these are the same type and have the same size.
 * DO NOT call this function directly.
 */
int
vec_istriplets(const struct vector_t *ap, const struct vector_t *bp,
	const struct vector_t *cp)
{
	if (ap == NULL || bp == NULL || cp == NULL)
		fatal("%s(A, B, C): one of these is NULL", __F);

	if (ap->vt_size != bp->vt_size || ap->vt_size != cp->vt_size)
		return 1;
	if (ap->vt_type != bp->vt_type || ap->vt_type != cp->vt_type)
		return 1;
	return 0;
}

/*
 * Check to see whether or not the three vector_t *ap, *bp are really
 * vector_t pointer and these are the same type and have the same size.
 * DO NOT call this function directly.
 */
int
vec_isdouplets(const struct vector_t *ap, const struct vector_t *bp)
{
	if (ap == NULL || bp == NULL)
		fatal("%s(A, B): one of these is NULL", __F);

	if (ap->vt_size != bp->vt_size)
		return 1;
	if (ap->vt_type != bp->vt_type)
		return 1;
	return 0;
}

/*
 * returns the \Sigma |a_i|
 */
double
vec_getnorm1(const struct vector_t *ap)
{
	double	height;
	int	i;
	int	*idata;
	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t  *zdata;

	vec_check(__F, ap, -1); /* -1 means I dont care about its type */

	height = 0.0;	/* initialized */
	switch (ap->vt_type) {
		case SP_INTEGER:
			idata = ivec(ap);
			for (i=0; i< ap->vt_size; i++)
				height += (double) abs(idata[i]);
			break;
		case SP_SINGLE:
			sdata = svec(ap);
			for (i=0; i< ap->vt_size; i++)
				height += fabs(sdata[i]);
			break;
		case SP_DOUBLE:
			ddata = dvec(ap);
			for (i=0; i< ap->vt_size; i++)
				height += fabs(ddata[i]);
			break;
		case SP_SIMPLEX:
			cdata = cvec(ap);
			for (i=0; i< ap->vt_size; i++)
				height += cp_absolute(&cdata[i]);
			break;
		case SP_DUPLEX:
			zdata = zvec(ap);
			for (i=0; i< ap->vt_size; i++)
				height += zp_absolute(&zdata[i]);
			break;
		default:
			fatal("%s(\"%s\"): Hmmm", __F);
	}
	return height;
}

double
vec_getmeannorm2(const struct vector_t *ap)
{
	return (vec_getnorm2(ap)/(double) ap->vt_size);
}

/*
 * returns the \sqrt(\Sigma (a_i * a_i))
 */
double
vec_getnorm2(const struct vector_t *ap)
{
	double	height;
	int	i;
	int	*idata;
	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t  *zdata;

	vec_check(__F, ap, -1); /* -1 means I dont care about its type */

	height = 0.0;	/* set to ZERO */
	switch (ap->vt_type) {
		case SP_INTEGER:
			idata = ivec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += (double) idata[i]*idata[i];
			break;
		case SP_SINGLE:
			sdata = svec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += sdata[i]*sdata[i];
			break;
		case SP_DOUBLE:
			ddata = dvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += ddata[i]*ddata[i];
			break;
		case SP_SIMPLEX:
			cdata = cvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += cp_square(&cdata[i]);
			break;
		case SP_DUPLEX:
			zdata = zvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += zp_square(&zdata[i]);
			break;
		default:
			fatal("vec_getnorm(\"%s\"): Hmmm", ap->vt_name);
	}
	return sqrt(height);
}

/*
 * returns the \Sigma (a_i * a_i))
 */
double
vec_getnormsq(const struct vector_t *ap)
{
	double	height;
	int	i;
	int	*idata;
	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t  *zdata;

	vec_check(__F, ap, -1); /* -1 means I dont care about its type */

	height = 0.0;	/* set to ZERO */
	switch (ap->vt_type) {
		case SP_INTEGER:
			idata = ivec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += (double) idata[i]*idata[i];
			break;
		case SP_SINGLE:
			sdata = svec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += sdata[i]*sdata[i];
			break;
		case SP_DOUBLE:
			ddata = dvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += ddata[i]*ddata[i];
			break;
		case SP_SIMPLEX:
			cdata = cvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += cp_square(&cdata[i]);
			break;
		case SP_DUPLEX:
			zdata = zvec(ap);
			for (i=0; i < ap->vt_size; i++)
				height += zp_square(&zdata[i]);
			break;
		default:
			fatal("vec_getnorm(\"%s\"): Hmmm", ap->vt_name);
	}
	return height;
}

/*
 * Returns the entry index whose entry has the largest norm
 */
int
vec_getlargestindex(const struct vector_t *ap)
{
	double	dlarge, dtmp;
	int	ilarge;
	int	i, *idata, retval;
	float	*sdata;
	double	*ddata;
	struct simplex_t *cdata;
	struct duplex_t  *zdata;

	vec_check(__F, ap, -1); /* -1 means I dont care about its type */

	dlarge = -1.0 * MAXDOUBLE;
	ilarge = (-1   * 0x00ffffff);
	switch (ap->vt_type) {
		case SP_INTEGER:
			idata = ivec(ap);
			for (i=0; i<ap->vt_size; i++) {
				if (ilarge < idata[i]) {
					retval = i;
					ilarge = idata[i];
				}
			}
			break;
		case SP_SINGLE:
			sdata = svec(ap);
			for (i=0; i<ap->vt_size; i++) {
				if (dlarge < sdata[i]) {
					retval = i;
					dlarge = sdata[i];
				}
			}
			break;
		case SP_DOUBLE:
			ddata = dvec(ap);
			for (i=0; i<ap->vt_size; i++) {
				if (dlarge < ddata[i]) {
					retval = i;
					dlarge = ddata[i];
				}
			}
			break;
		case SP_SIMPLEX:
			cdata = cvec(ap);
			for (i=0; i<ap->vt_size; i++) {
				dtmp = cp_square(&cdata[i]);
				if (dlarge < dtmp) {
					retval = i;
					dlarge = dtmp;
				}
			}
			break;
		case SP_DUPLEX:
			zdata = zvec(ap);
			for (i=0; i<ap->vt_size; i++) {
				dtmp = zp_square(&zdata[i]);
				if (dlarge < dtmp) {
					retval = i;
					dlarge = dtmp;
				}
			}
			break;
		default:
			fatal("vec_getlargest(\"%s\"): Unkown vector_type %d",
				ap->vt_name, ap->vt_type);
	}
	return retval;
}


/*
 * vector plus operations
 * A = B + C
 */
struct vector_t *
vec_plus(struct vector_t *ap, const struct vector_t *bp,
	const struct vector_t *cp)
{
	int	i, size;
	int	*ipp[3];
	float	*spp[3];
	double	*dpp[3];
	struct simplex_t *cpp[3];
	struct duplex_t  *zpp[3];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	vec_check(__F, cp, -1);

	if (vec_istriplets(ap, bp, cp))
		fatal("vec_plus(A, B, C): Three vector_t's are not triplets");

	size = ap->vt_size;
	/*
	 * there is no need to null vector_t *ap
	 */
	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp[0] = ivec(ap); ipp[1] = ivec(bp); ipp[2] = ivec(cp);
			for (i=0; i< size; i++)
				ipp[0][i] = ipp[1][i] + ipp[2][i];
			break;
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp); spp[2] = svec(cp);
			for (i=0; i< size; i++)
				spp[0][i] = spp[1][i] + spp[2][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp); dpp[2] = dvec(cp);
			for (i=0; i< size; i++)
				dpp[0][i] = dpp[1][i] + dpp[2][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp); cpp[2] = cvec(cp);
			for (i=0; i< size; i++)
				cp_plus(&cpp[0][i], &cpp[1][i], &cpp[2][i]);
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp); zpp[2] = zvec(cp);
			for (i=0; i< size; i++)
				zp_plus(&zpp[0][i], &zpp[1][i], &zpp[2][i]);
			break;
	}
	return ap;
}


/*
 * vector pplus operations
 * A += B, ie,  A = A + B
 */
struct vector_t *
vec_pplus(struct vector_t *ap, const struct vector_t *bp)
{
	int	i, size;
	int	*ipp[2];
	float	*spp[2];
	double	*dpp[2];
	struct simplex_t *cpp[2];
	struct duplex_t  *zpp[2];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("vec_pplus(A, B): Two vector_t's are not douplets");

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp[0] = ivec(ap); ipp[1] = ivec(bp);
			for (i=0; i< size; i++)
				ipp[0][i] += ipp[1][i];
			break;
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp);
			for (i=0; i< size; i++)
				spp[0][i] += spp[1][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp);
			for (i=0; i< size; i++)
				dpp[0][i] += dpp[1][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp);
			for (i=0; i< size; i++)
				cp_pplus(&cpp[0][i], &cpp[1][i]);
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp);
			for (i=0; i< size; i++)
				zp_pplus(&zpp[0][i], &zpp[1][i]);
			break;
	}
	return ap;
}

/*
 * vector minus operations
 * A = B - C
 */
struct vector_t *
vec_minus(struct vector_t *ap, const struct vector_t *bp,
	const struct vector_t *cp)
{
	int	i, size;
	int	*ipp[3];
	float	*spp[3];
	double	*dpp[3];
	struct simplex_t *cpp[3];
	struct duplex_t  *zpp[3];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	vec_check(__F, cp, -1);

	if (vec_istriplets(ap, bp, cp))
		fatal("vec_minus(A, B, C): Three vector_t's are not triplets");

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp[0] = ivec(ap); ipp[1] = ivec(bp); ipp[2] = ivec(cp);
			for (i=0; i< size; i++)
				ipp[0][i] = ipp[1][i] - ipp[2][i];
			break;
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp); spp[2] = svec(cp);
			for (i=0; i< size; i++)
				spp[0][i] = spp[1][i] - spp[2][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp); dpp[2] = dvec(cp);
			for (i=0; i< size; i++)
				dpp[0][i] = dpp[1][i] - dpp[2][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp); cpp[2] = cvec(cp);
			for (i=0; i< size; i++)
				cp_minus(&cpp[0][i], &cpp[1][i], &cpp[2][i]);
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp); zpp[2] = zvec(cp);
			for (i=0; i< size; i++)
				zp_minus(&zpp[0][i], &zpp[1][i], &zpp[2][i]);
			break;
		default:
			fatal("vec_minus(\"%s\"):undotprod-able type",
				ap->vt_name);
	}
	return ap;
}


/*
 * vector mminus operations
 * A -= B, i.e., A = A - B
 */
struct vector_t *
vec_mminus(struct vector_t *ap, const struct vector_t *bp)
{
	int	i, size;
	int	*ipp[2];
	float	*spp[2];
	double	*dpp[2];
	struct simplex_t *cpp[2];
	struct duplex_t  *zpp[2];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("vec_mminus(A, B): Two vector_t's are not douplets");

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp[0] = ivec(ap); ipp[1] = ivec(bp);
			for (i=0; i< size; i++)
				ipp[0][i] -= ipp[1][i];
			break;
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp);
			for (i=0; i< size; i++)
				spp[0][i] -= spp[1][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp);
			for (i=0; i< size; i++)
				dpp[0][i] -= dpp[1][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp);
			for (i=0; i< size; i++)
				cp_mminus(&cpp[0][i], &cpp[1][i]);
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp);
			for (i=0; i< size; i++)
				zp_mminus(&zpp[0][i], &zpp[1][i]);
			break;
	}
	return ap;
}

/*
 * Vector copy operation; ap is the target and bp is source
 */
void
vec_copy(struct vector_t *ap, const struct vector_t *bp)
{
	int	size;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("vec_copy(A, B): A and B are not douplets; can not copy");

	switch (ap->vt_type) {
		case SP_INTEGER:
			size = sizeof(int) * ap->vt_size; break;
		case SP_SINGLE:
			size = sizeof(float) * ap->vt_size; break;
		case SP_DOUBLE:
			size = sizeof(double) * ap->vt_size; break;
		case SP_SIMPLEX:
			size = sizeof(struct simplex_t) * ap->vt_size; break;
		case SP_DUPLEX:
			size = sizeof(struct duplex_t) * ap->vt_size; break;
	}
	memcpy(ap->vt_data, bp->vt_data, (size_t) size);
	return;
}

/*
 * vector_t pure duplication operation
 * DOES NOT copy its contents, instead it makes same sized, same typed vector_t
 */
struct vector_t *
vec_puredup(const struct vector_t *ap)
{
	struct vector_t *retap;
	char	name[256];

	vec_check(__F, ap, -1);

	sprintf(name, "puredup'ed of <%s>", ap->vt_name);
	retap = vector(ap->vt_size, name, ap->vt_type);

	return retap;
}

/*
 * vector_t duplication operation. duplicates the contents..
 */
struct vector_t *
vec_dup(const struct vector_t *ap)
{
	struct vector_t *retap;
	char	name[256];

	vec_check(__F, ap, -1);

	sprintf(name, "dup'ed of <%s>", ap->vt_name);
	retap = vector(ap->vt_size, name, ap->vt_type);

	vec_copy(retap, ap);
	return retap;
}

/*
 * Get the norm of the entry at..
 */
double
vec_getabsoluteat(const struct vector_t *ap, int where)
{
	double	retval;

	int	*ipp;
	float	*spp;
	double	*dpp;
	struct simplex_t *cpp;
	struct duplex_t *zpp;

	vec_check(__F, ap, -1);
	if (where > ap->vt_size)
		fatal("%s(\"%s\", %d):location exceed the size", __F,
			ap->vt_name, where);

	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp = ivec(ap);
			retval = fabs(ipp[where]);
			break;
		case SP_SINGLE:
			spp = svec(ap);
			retval = fabs(spp[where]);
			break;
		case SP_DOUBLE:
			dpp = dvec(ap);
			retval = fabs(dpp[where]);
			break;
		case SP_SIMPLEX:
			cpp = cvec(ap);
			retval = cp_absolute(&cpp[where]);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			retval = zp_absolute(&zpp[where]);
			break;
	}
	return retval;
}

/*
 * vector product for non-complex cases
 */
double
vec_ddotprod(const struct vector_t *ap, const struct vector_t *bp)
{
	int	i, size;
	int	*ipp[2];
	float	*spp[2];
	double	*dpp[2];
	double	retval;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("%s(A, B): Three vector_t's are not douplets", __F);

	size = ap->vt_size;
	retval = 0.0;

	switch (ap->vt_type) {
		case SP_INTEGER:
			ipp[0] = ivec(ap); ipp[1] = ivec(bp);
			for (i=0; i< size; i++)
				retval += (double) (ipp[0][i] * ipp[1][i]);
			break;
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp);
			for (i=0; i< size; i++)
				retval += (double) (spp[0][i] * spp[1][i]);
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp);
			for (i=0; i< size; i++)
				retval += dpp[0][i] * dpp[1][i];
			break;
		default:
			fatal("%s(\"%s\",\"%s\"):undotprod-able type", __F,
				ap->vt_name, bp->vt_name);
				
	}
	return retval;
}

/*
 * vector product for complex cases
 * Is this correct??
 */
struct duplex_t *
vec_zdotprod(struct duplex_t *ret, const struct vector_t *ap,
		const struct vector_t *bp)
{
	int	i, size;
	struct simplex_t *cpp[2];
	struct duplex_t  *zpp[2];
	struct simplex_t cretval, ccomp;
	struct duplex_t zretval, zcomp;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("%s(A, B): Two vector_t's are not douplets", __F);

	size = ap->vt_size;
	cretval = simplex_null;
	zretval = duplex_null;

	switch (ap->vt_type) {
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp);
			for (i=0; i< size; i++)
				cp_pplus(&cretval, cp_conjcross(&ccomp,
					&cpp[0][i], &cpp[1][i]));

			ret->real = cretval.real;
			ret->imag = cretval.imag;
			break;

		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp);
			for (i=0; i< size; i++)
				zp_pplus(&zretval, zp_conjcross(&zcomp,
					&zpp[0][i], &zpp[1][i]));
			ret->real = zretval.real;
			ret->imag = zretval.imag;
			break;
		default:
			fatal("%s(\"%s\",\"%s\"):undotprod-able type", __F,
				ap->vt_name, bp->vt_name);
	}
	return ret;
}

/*
 * int ind indicates whether we should conjugate the ap complex 
 */
struct duplex_t *
vec_zdotprod2nd(struct duplex_t *ret, int ind, const struct vector_t *ap,
		const struct vector_t *bp)
{
	int	i, size;
	struct simplex_t *cpp[2];
	struct duplex_t  *zpp[2];
	struct simplex_t cretval, ccomp;
	struct duplex_t zretval, zcomp;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);

	if (vec_isdouplets(ap, bp))
		fatal("%s(A, B): Two vector_t's are not douplets", __F);

	size = ap->vt_size;
	cretval = simplex_null;
	zretval = duplex_null;

	switch (ap->vt_type) {
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp);
			if (ind)
				for (i=0; i < size; i++)
					cp_pplus(&cretval, cp_conjcross(&ccomp,
						&cpp[0][i], &cpp[1][i]));
			else
				for (i=0; i < size; i++)
					cp_pplus(&cretval, cp_cross(&ccomp,
						&cpp[0][i], &cpp[1][i]));

			ret->real = cretval.real;
			ret->imag = cretval.imag;
			break;

		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp);
			if (ind)
				for (i=0; i< size; i++)
					zp_pplus(&zretval, zp_conjcross(&zcomp,
						&zpp[0][i], &zpp[1][i]));
			else
				for (i=0; i< size; i++)
					zp_pplus(&zretval, zp_cross(&zcomp,
						&zpp[0][i], &zpp[1][i]));

			ret->real = zretval.real;
			ret->imag = zretval.imag;
			break;
		default:
			fatal("%s(\"%s\",\"%s\"):undotprod-able type", __F,
				ap->vt_name, bp->vt_name);
	}
	return ret;
}

/*
 * multiplication of vector with a double constant
 * A = B.d
 */
struct vector_t *
vec_crossd(struct vector_t *ret, const struct vector_t *ap,
	const double multicant)
{
	int	i, size;
	float	*spp, *sd;
	double	*dpp, *dd;
	struct simplex_t *cpp, *cd;
	struct duplex_t  *zpp, *zd;

	vec_check(__F, ap, -1);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SINGLE:
			spp = svec(ap);
			sd = svec(ret);
			for (i=0; i< size; i++)
				sd[i] = spp[i] * multicant;
			break;
		case SP_DOUBLE:
			dpp = dvec(ap);
			dd = dvec(ret);
			for (i=0; i< size; i++)
				dd[i] = dpp[i] * multicant;
			break;
		case SP_SIMPLEX:
			cpp = cvec(ap);
			cd = cvec(ret);
			for (i=0; i< size; i++)
				cp_crossd(&cd[i], &cpp[i], multicant);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			zd = zvec(ret);
			for (i=0; i< size; i++)
				zp_crossd(&zd[i], &zpp[i], multicant);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ret;
}

/*
 * multiplication of vector with a duplex_t constant
 * A = z.A
 */
struct vector_t *
vec_crossz(struct vector_t *ret, const struct vector_t *ap,
	const struct duplex_t multicant)
{
	int	i, size;
	struct simplex_t *cpp, *cd, multi;
	struct duplex_t  *zpp, *zd;

	if  (vec_isdouplets(ret, ap))
		fatal("%s(\"%s\", d, \"%s\"): vec_isdouplets()", __F,
			ret->vt_name, ap->vt_name);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SIMPLEX:
			multi.real = multicant.real;
			multi.imag  = multicant.imag;
			cpp = cvec(ap);
			cd = cvec(ret);
			for (i=0; i< size; i++)
				cp_cross(&cd[i], &cpp[i], &multi);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			zd = zvec(ret);
			for (i=0; i< size; i++)
				zp_cross(&zd[i], &zpp[i], &multicant);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ret;
}

/*
 * multiplication of vector with a double constant
 * A = d.A
 */
struct vector_t *
vec_dcross(struct vector_t *ret,
	const double multicant, const struct vector_t *ap)
{
	int	i, size;
	float	*spp, *sd;
	double	*dpp, *dd;
	struct simplex_t *cpp, *cd;
	struct duplex_t  *zpp, *zd;

	if  (vec_isdouplets(ret, ap))
		fatal("%s(\"%s\", d, \"%s\"): vec_isdouplets()", __F,
			ret->vt_name, ap->vt_name);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SINGLE:
			spp = svec(ap);
			sd = svec(ret);
			
			for (i=0; i< size; i++)
				sd[i] = spp[i] * multicant;
			break;
		case SP_DOUBLE:
			dpp = dvec(ap);
			dd = dvec(ret);
			for (i=0; i< size; i++)
				dd[i] = dpp[i] * multicant;
			break;
		case SP_SIMPLEX:
			cpp = cvec(ap);
			cd = cvec(ret);
			for (i=0; i< size; i++)
				cp_crossd(&cd[i], &cpp[i], multicant);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			zd = zvec(ret);
			for (i=0; i< size; i++)
				zp_crossd(&zd[i], &zpp[i], multicant);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ret;
}

struct vector_t *
vec_dcrossme(struct vector_t *ap, const double multicant)
{
	int	i, size;
	float	*spp;
	double	*dpp;
	struct simplex_t *cpp;
	struct duplex_t  *zpp;

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SINGLE:
			spp = svec(ap);
			for (i=0; i< size; i++)
				spp[i] *= multicant;
			break;
		case SP_DOUBLE:
			dpp = dvec(ap);
			for (i=0; i< size; i++)
				dpp[i] *= multicant;
			break;
		case SP_SIMPLEX:
			cpp = cvec(ap);
			for (i=0; i< size; i++)
				cp_crossdme(&cpp[i], multicant);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			for (i=0; i< size; i++)
				zp_crossdme(&zpp[i], multicant);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ap;
}

/*
 * multiplication of vector with a struct duplex_t constant
 * B = z.A
 */
struct vector_t *
vec_zcross(struct vector_t *ret, const struct duplex_t multicant,
		const struct vector_t *ap)
{
	int	i, size;
	struct simplex_t *cpp, *cd, mult;
	struct duplex_t  *zpp, *zd;

	if  (vec_isdouplets(ret, ap))
		fatal("%s(\"%s\", d, \"%s\"): vec_isdouplets()", __F,
			ret->vt_name, ap->vt_name);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SIMPLEX:
			mult.real = multicant.real;
			mult.imag = multicant.imag;
			cpp = cvec(ap);
			cd = cvec(ret);
			for (i=0; i< size; i++)
				cp_cross(&cd[i], &mult, &cpp[i]);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			zd = zvec(ret);
			for (i=0; i< size; i++)
				zp_cross(&zd[i], &multicant, &zpp[i]);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ret;
}

struct vector_t *
vec_zcrossme(struct vector_t *ret, const struct duplex_t multicant)
{
	int	i, size;
	struct simplex_t *cd;
	struct duplex_t  *zd;

	size = ret->vt_size;
	switch (ret->vt_type) {
		case SP_SIMPLEX:
			cd = cvec(ret);
			for (i=0; i< size; i++)
				cp_crosszme(&cd[i], multicant);
			break;
		case SP_DUPLEX:
			zd = zvec(ret);
			for (i=0; i< size; i++)
				zp_crosszme(&zd[i], multicant);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ret->vt_name);
	}
	return ret;
}

struct vector_t *
vec_zcross2nd(struct vector_t *ret, const int indicator,
		const struct duplex_t multicant, const struct vector_t *ap)
{
	int	i, size;
	struct simplex_t *cpp, *cd, mult;
	struct duplex_t  *zpp, *zd;

	if  (vec_isdouplets(ret, ap))
		fatal("%s(\"%s\", d, \"%s\"): vec_isdouplets()", __F,
			ret->vt_name, ap->vt_name);


	mult.real = multicant.real;
	mult.imag = multicant.imag;

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SIMPLEX:
			cpp = cvec(ap);
			cd = cvec(ret);
			if (indicator)
				for (i=0; i< size; i++)
					cp_conjcross(&cd[i], &mult, &cpp[i]);
			else
				for (i=0; i< size; i++)
					cp_cross(&cd[i], &mult, &cpp[i]);
			break;
		case SP_DUPLEX:
			zpp = zvec(ap);
			zd = zvec(ret);
			if (indicator)
				for (i=0; i< size; i++)
					zp_conjcross(&zd[i],&multicant, &zpp[i]);
			else
				for (i=0; i< size; i++)
					zp_cross(&zd[i], &multicant, &zpp[i]);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ret;
}

/*
 * vector combination operations
 * A = aB + bC
 */
struct vector_t *
vec_dcombi(struct vector_t *ap, const double first, const struct vector_t *bp,
	const double second, const struct vector_t *cp)
{
	int	i, size;
	float	*spp[3];
	double	*dpp[3];
	struct simplex_t *cpp[3], ctmp[2];
	struct duplex_t  *zpp[3], ztmp[2];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	vec_check(__F, cp, -1);

	if (vec_istriplets(ap, bp, cp))
		fatal("%s(A, B, C): Three vector_t's are not triplets", __F);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp); spp[2] = svec(cp);
			for (i=0; i< size; i++)
				spp[0][i] = first * spp[1][i] + second * spp[2][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp); dpp[2] = dvec(cp);
			for (i=0; i< size; i++)
				dpp[0][i] = first * dpp[1][i] + second * dpp[2][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp); cpp[2] = cvec(cp);
			for (i=0; i< size; i++)
				cp_plus(&cpp[0][i],
					cp_crossd(&ctmp[0], &cpp[1][i], first),
					cp_crossd(&ctmp[1], &cpp[2][i], second));
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp); zpp[2] = zvec(cp);
			for (i=0; i< size; i++)
				zp_plus(&zpp[0][i],
					zp_crossd(&ztmp[0], &zpp[1][i], first),
					zp_crossd(&ztmp[1], &zpp[2][i], second));
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ap;
}

struct vector_t *
vec_zcombi(struct vector_t *ap,
	const struct duplex_t first, const struct vector_t *bp,
	const struct duplex_t sec, const struct vector_t *cp)
{
	int	i, size;
	struct simplex_t *cpp[3], ctmp[2], cfirst, csec;
	struct duplex_t  *zpp[3], ztmp[2];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	vec_check(__F, cp, -1);

	if (vec_istriplets(ap, bp, cp))
		fatal("%s(A, B, C): Three vector_t's are not triplets", __F);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp); cpp[2] = cvec(cp);
			cz_du2sim(&cfirst, &first);
			cz_du2sim(&csec, &sec);

			for (i=0; i< size; i++)
				cp_plus(&cpp[0][i],
					cp_cross(&ctmp[0], &cfirst, &cpp[1][i]),
					cp_cross(&ctmp[1], &csec, &cpp[2][i]));
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp); zpp[2] = zvec(cp);
			for (i=0; i< size; i++)
				zp_plus(&zpp[0][i],
					zp_cross(&ztmp[0], &first, &zpp[1][i]),
					zp_cross(&ztmp[1], &sec, &zpp[2][i]));
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ap;
}


/*
 * a_i = (1.0/b_i) + c_i
 *
 */
struct vector_t *
vec_inverseplus(struct vector_t *ap, const struct vector_t *bp,
	const struct vector_t *cp)
{
	int	i, size;
	float	*spp[3];
	double	*dpp[3];
	struct simplex_t *cpp[3];
	struct duplex_t  *zpp[3];

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	vec_check(__F, cp, -1);

	if (vec_istriplets(ap, bp, cp))
		fatal("%s(A, B, C): Three vector_t's are not triplets", __F);

	size = ap->vt_size;
	switch (ap->vt_type) {
		case SP_SINGLE:
			spp[0] = svec(ap); spp[1] = svec(bp); spp[2] = svec(cp);
			for (i=0; i< size; i++)
				spp[0][i] = spp[2][i]/spp[1][i];
			break;
		case SP_DOUBLE:
			dpp[0] = dvec(ap); dpp[1] = dvec(bp); dpp[2] = dvec(cp);
			for (i=0; i< size; i++)
				dpp[0][i] = dpp[2][i]/dpp[1][i];
			break;
		case SP_SIMPLEX:
			cpp[0] = cvec(ap); cpp[1] = cvec(bp); cpp[2] = cvec(cp);
			for (i=0; i< size; i++)
				cp_slash(&cpp[0][i], &cpp[2][i], &cpp[1][i]);
			break;
		case SP_DUPLEX:
			zpp[0] = zvec(ap); zpp[1] = zvec(bp); zpp[2] = zvec(cp);
			for (i=0; i< size; i++)
				zp_slash(&zpp[0][i], &zpp[2][i], &zpp[1][i]);
			break;
		default:
			fatal("%s(\"%s\"):undotprod-able type", __F,
				ap->vt_name);
	}
	return ap;
}
