#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "common.h"
#include "complex.h"
#include "vector.h"

void vec_check(const char *function, const struct vector_t *vp, int type);

struct vector_t *vec_iread(FILE *, int, const char *);
struct vector_t *vec_sread(FILE *, int, const char *);
struct vector_t *vec_dread(FILE *, int, const char *);
struct vector_t *vec_cread(FILE *, int, const char *);
struct vector_t *vec_zread(FILE *, int, const char *);

void	vec_plexprintatfiles(const struct vector_t *, const char *);

/*
 * VECTOR CREATION
 */
struct vector_t *
vector(int size, const char *name, int type)
{
	struct vector_t *vt;
	int	allocing;

	if (size < 1)
		fatal("int vector size less than 1?");
	if (!ATYPE(type))
		fatal("%s(): wrong type for \"%s\"", name);

	vt = (struct vector_t *)
		malloc((size_t) (sizeof(struct vector_t)));
	if (vt == NULL)
		memerr("vector(\"%s\"): mallocing vt");

	switch (type) {
		case SP_INTEGER:
			allocing = sizeof(int)*size;
			vt->vt_data = malloc(allocing);
			vt->vt_type = SP_INTEGER;
			break;
		case SP_SINGLE:
			allocing = sizeof(float)*size;
			vt->vt_data = malloc(allocing);
			vt->vt_type = SP_SINGLE;
			break;
		case SP_DOUBLE:
			allocing = sizeof(double)*size;
			vt->vt_data = malloc(allocing);
			vt->vt_type = SP_DOUBLE;
			break;
		case SP_SIMPLEX:
			allocing = sizeof(struct simplex_t)*size;
			vt->vt_data = malloc(allocing);
			vt->vt_type = SP_SIMPLEX;
			break;
		case SP_DUPLEX:
			allocing = sizeof(struct duplex_t)*size;
			vt->vt_data = malloc(allocing);
			vt->vt_type = SP_DUPLEX;
			break;
		default:
			fatal("sparse(\"%s\"): wrong type %d", name, type);
	}
	vt->vt_flag = VECTOR_TYPED;
	vt->vt_size = size;
	vt->vt_base = 0;	/* default is 0 */

	/*
	 * Initialied as zero's
	 */

        if (vt->vt_data == NULL)
                memerr("vector(\"%s\"): [2]", name ? name : UNKNOWN);

	memset((void *) vt->vt_data, 0x00, allocing); /* make all zeros */
	if (name && *name)
		vt->vt_name = strdup(name);
	else
		vt->vt_name = noname;

	return vt;
}

void
vec_check(const char *function, const struct vector_t *vp, const int type)
{
	if (vp == NULL)
		fatal("%s(): I received NULL vector_t pointer", function);
	if (!IS_VECTOR(vp))
		fatal("%s(): I received a pointer but this is not vector_t",
			function);
	if (!ATYPE(type))
		return;	/* only checked whether this pointer is vector_t */

	if (type == vp->vt_type)
		return;	/* type is matched */

	/*
	 * Now mismatched case
	 */
	switch (type) {
		case SP_INTEGER:
			fatal("%s(): vector %s is not SP_INTEGER type",
				function, vp->vt_name);
			break;
		case SP_SINGLE:
			fatal("%s(): vector %s is not SP_SINGLE type",
				function, vp->vt_name);
			break;
		case SP_DOUBLE:
			fatal("%s(): vector %s is not SP_DOUBLE type",
				function, vp->vt_name);
			break;
		case SP_SIMPLEX:
			fatal("%s(): vector %s is not SP_SIMPLEX type",
				function, vp->vt_name);
			break;
		case SP_DUPLEX:
			fatal("%s(): vector %s is not SP_DUPLEX type",
				function, vp->vt_name);
			break;
	}
	return;
}

/*
 * DATA ACCESS ROUTINES
 */

inline int *
ivec(const struct vector_t *vp)
{
	vec_check(__F, vp, SP_INTEGER);
	return ((int *) vp->vt_data);
}

inline float *
svec(const struct vector_t *vp)
{
	vec_check(__F, vp, SP_SINGLE);
	return (float *) vp->vt_data;
}


inline double *
dvec(const struct vector_t *vp)
{
	vec_check(__F, vp, SP_DOUBLE);
	return (double *) vp->vt_data;
}

inline struct simplex_t *
cvec(const struct vector_t *vp)
{
	vec_check(__F, vp, SP_SIMPLEX);
	return (struct simplex_t *) vp->vt_data;
}

inline struct duplex_t *
zvec(const struct vector_t *vp)
{
	vec_check(__F, vp, SP_DUPLEX);
	return (struct duplex_t *) vp->vt_data;
}

void
vec_free(struct vector_t *vec)
{
	free((void *) vec->vt_data);
	free((void *) vec);
	return;
}

void
vec_mfree(int  num,...)
{
	int     i;
	va_list ap;
	struct vector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct vector_t *);
		vec_free(p);
	}
	va_end(ap);
	return;
}


struct vector_t *
vec_read(const char *fname)
{
	FILE	*fp;
	int	size;
	int	readn;
	char	typename[256];
	char	line[256], *p;
	struct vector_t *vectp;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("vec_read(\"%s\"): while fopen()", fname);
	p = fgets(line, 256, fp);
	if (p == NULL)
		fatal("vec_read(\"%s\"): while fgets()", fname);
	readn = sscanf(line, "%d %s", &size, typename);
	if (readn != 2 || size < 1)
		fatal("vec_read(\"%s\"): while sscanf()", fname);

	switch (*typename) {
		case 'i': vectp = vec_iread(fp, size, fname); break;
		case 's': vectp = vec_sread(fp, size, fname); break;
		case 'd': vectp = vec_dread(fp, size, fname); break;
		case 'c': vectp = vec_cread(fp, size, fname); break;
		case 'z': vectp = vec_zread(fp, size, fname); break;
		default:
			fatal("vec_read(\"%s\"): strange char=%c: type = isdcz",
				fname, *typename);
	}
	fclose(fp);
	return vectp;
}

/*
 * A private function of vec_read();
 * DO NOT call this function directly
 */

struct vector_t *
vec_iread(FILE *fp, int	size, const char *fname)
{
	char	line[256], name[BUFSIZ];
	int	readno, cnt;
	int	data, *datap;
	struct vector_t *vp;
	
	sprintf(name, "read from \"%s\"", fname);
	vp = vector(size, name, SP_INTEGER);
	datap = ivec(vp);
	cnt = 0;
	while (fgets(line, 256, fp)) {
		readno = sscanf(line, "%d", &data);
		if (readno != 1)
			fatal("%s(\"%s\"): while sscanf()", __F, fname);
		cnt++;
		if (cnt > size)
			fatal("%s(\"%s\"): exceeds the specified no %d", __F,
				fname, size);
		*datap++ = data;
	}
		
	return vp;
}

/*
 * A private function of vec_read();
 * DO NOT call this function directly
 */
struct vector_t *
vec_sread(FILE *fp, int	size, const char *fname)
{
	char	line[256], name[BUFSIZ];
	int	readno, cnt;
	float	data, *datap;
	struct vector_t *vp;
	
	sprintf(name, "read from \"%s\"", fname);
	vp = vector(size, name, SP_SINGLE);
	datap = svec(vp);
	cnt = 0;
	while (fgets(line, 256, fp)) {
		readno = sscanf(line, "%f", &data);
		if (readno != 1)
			fatal("%s(\"%s\"): while sscanf()", __F, fname);
		cnt++;
		if (cnt > size)
			fatal("%s(\"%s\"): exceeds the specified no %d", __F,
				fname, size);
		*datap++ = data;
	}
		
	return vp;
}

/*
 * A private function of vec_read();
 * DO NOT call this function directly
 */
struct vector_t *
vec_dread(FILE *fp, int	size, const char *fname)
{
	char	line[256], name[BUFSIZ];
	int	readno, cnt;
	float	data;
	double	*datap;
	struct vector_t *vp;

	sprintf(name, "read from \"%s\"", fname);
	vp = vector(size, name, SP_DOUBLE);
	datap = dvec(vp);
	cnt = 0;
	while (fgets(line, 256, fp)) {
		readno = sscanf(line, "%f", &data);
		if (readno != 1)
			fatal("%s(\"%s\"): while sscanf()", __F, fname);
		cnt++;
		if (cnt > size)
			fatal("%s(\"%s\"): exceeds the specified no %d", __F,
				fname, size);
		*datap++ = data;
	}

	return vp;
}

/*
 * A private function of vec_read();
 * DO NOT call this function directly
 */
struct vector_t *
vec_cread(FILE *fp, int	size, const char *fname)
{
	char	line[256], name[BUFSIZ];
	int	readno, cnt;
	float	rea, imag;
	struct vector_t *vp;
	struct simplex_t *sxp;
	
	sprintf(name, "read from \"%s\"", fname);
	vp = vector(size, name, SP_SIMPLEX);
	sxp = cvec(vp);
	cnt = 0;
	while (fgets(line, 256, fp)) {
		readno = sscanf(line, "%f %f" , &rea, &imag);
		if (readno != 2)
			fatal("%s(\"%s\"): %s", __F, line);
		cnt++;
		if (cnt > size)
			fatal("%s(\"%s\"):%d exceeds the size no %d", __F,
				fname, cnt, size);
		sxp->real = rea;
		sxp->imag = imag;
		sxp++;
	}
	return vp;
}


/*
 * A private function of vec_read();
 * DO NOT call this function directly
 */
struct vector_t *
vec_zread(FILE *fp, int	size, const char *fname)
{
	char	line[256], name[BUFSIZ];
	int	readno, cnt;
	float	rea, imag;
	struct vector_t *vp;
	struct duplex_t *dxp;
	
	sprintf(name, "read from \"%s\"", fname);
	vp = vector(size, name, SP_DUPLEX);
	dxp = zvec(vp);
	cnt = 0;
	while (fgets(line, 256, fp)) {
		readno = sscanf(line, "%f %f" , &rea, &imag);
		if (readno != 2)
			fatal("%s(\"%s\"): %s", __F, line);
		cnt++;
		if (cnt > size)
			fatal("%s(\"%s\"):%d exceeds the size no %d", __F,
				fname, cnt, size);
		dxp->real = rea;
		dxp->imag = imag;
		dxp++;
	}
	return vp;
}

void
vec_mprint(int  num,...)
{
	int     i;
	va_list ap;
	struct vector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct vector_t *);
		vec_print(p);
	}
	va_end(ap);
	return;
}


/*
 * PRINT
 */
void
vec_print(const struct vector_t *vp)
{
	int i;

	int	*ip;
	float	*sp;
	double	*dp;
	struct simplex_t *cp;
	struct duplex_t *zp;

	vec_check(__F, vp, -1);
	printf("--vec_print(%s)------\n", vp->vt_name);
	switch (vp->vt_type) {
		case SP_INTEGER:
			ip = ivec(vp);
			for (i=0;i < vp->vt_size; i++)
				(void) printf("%10d\n", ip[i]);
			break;
		case SP_SINGLE:
			sp = svec(vp);
			for (i=0;i < vp->vt_size; i++)
				(void) printf("%13.5e\n", sp[i]);
			break;
		case SP_DOUBLE:
			dp = dvec(vp);
			for (i=0;i < vp->vt_size; i++)
				(void) printf("%13.5e\n", dp[i]);
			break;
		case SP_SIMPLEX:
			cp = cvec(vp);
			for (i=0;i < vp->vt_size; i++)
				(void) printf("%13.5e %13.5e\n",
					cp[i].real, cp[i].imag);
			break;
		case SP_DUPLEX:
			zp = zvec(vp);
			for (i=0;i < vp->vt_size; i++)
				(void) printf("%13.5e %13.5e\n",
					zp[i].real, zp[i].imag);
			break;
		default:
			fatal("vec_print(): I dont understand your type");
			break;
	}
	return;
}

/*
 * two-colmn-wise printing
 */

void
vec_pairprint(const struct vector_t *ap, const struct vector_t *bp)
{
	int i;

	int	*ip0, *ip1;
	float	*sp0, *sp1;
	double	*dp0, *dp1;
	struct simplex_t *cp0, *cp1;
	struct duplex_t *zp0, *zp1;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	if (vec_isdouplets(ap, bp))
		fatal("%s(A, B): A and B are not douplets;", __F);

	printf("--vec_print(%s)------\n", ap->vt_name);
	switch (ap->vt_type) {
		case SP_INTEGER:
			ip0 = ivec(ap); ip1 = ivec(bp);
			for (i=0;i < ap->vt_size; i++)
				(void) printf("%10d %10d\n", ip0[i], ip1[i]);
			break;
		case SP_SINGLE:
			sp0 = svec(ap); sp1 = svec(bp);
			for (i=0;i < ap->vt_size; i++)
				(void) printf("%15.7e %15.7e\n", sp0[i],sp1[i]);
			break;
		case SP_DOUBLE:
			dp0 = dvec(ap); dp1 = dvec(bp);
			for (i=0;i < ap->vt_size; i++)
				(void) printf("%15.7e %15.7e\n", dp0[i],dp1[i]);
			break;
		case SP_SIMPLEX:
			cp0 = cvec(ap); cp1 = cvec(bp);
			for (i=0;i < ap->vt_size; i++)
				(void) printf("%15.7e %15.7e--%15.7e %15.7e\n",
					cp0[i].real, cp0[i].imag,
					cp1[i].real, cp1[i].imag);
			break;
		case SP_DUPLEX:
			zp0 = zvec(ap); zp1 = zvec(bp);
			for (i=0;i < ap->vt_size; i++)
				(void) printf("%15.7e %15.7e--%15.7e %15.7e\n",
					zp0[i].real, zp0[i].imag,
					zp1[i].real, zp1[i].imag);
			break;
		default:
			fatal("%s(): I dont understand your type", __F);
			break;
	}
	return;
}

/*
 * returns the difference of the two vector per entry
 */
double
vec_unitdiff(const struct vector_t *ap, const struct vector_t *bp)
{
	int i;

	int	*ip0, *ip1;
	float	*sp0, *sp1;
	double	*dp0, *dp1;
	struct simplex_t *cp0, *cp1, ctmp;
	struct duplex_t *zp0, *zp1, ztmp;
	double	diff;

	vec_check(__F, ap, -1);
	vec_check(__F, bp, -1);
	if (vec_isdouplets(ap, bp))
		fatal("%s(A, B): A and B are not douplets;", __F);

	diff = 0.0;
	switch (ap->vt_type) {
		case SP_INTEGER:
			ip0 = ivec(ap); ip1 = ivec(bp);
			for (i=0;i < ap->vt_size; i++)
				diff += (double) abs(ip0[i]- ip1[i]);
			break;
		case SP_SINGLE:
			sp0 = svec(ap); sp1 = svec(bp);
			for (i=0;i < ap->vt_size; i++)
				diff += fabs((double)(sp0[i] - sp1[i]));
			break;
		case SP_DOUBLE:
			dp0 = dvec(ap); dp1 = dvec(bp);
			for (i=0;i < ap->vt_size; i++)
				diff += fabs(dp0[i] - dp1[i]);
			break;
		case SP_SIMPLEX:
			cp0 = cvec(ap); cp1 = cvec(bp);
			for (i=0;i < ap->vt_size; i++)
				diff += cp_absolute(cp_minus(&ctmp,
					&cp0[i], &cp1[i]));
			break;
		case SP_DUPLEX:
			zp0 = zvec(ap); zp1 = zvec(bp);
			for (i=0;i < ap->vt_size; i++)
				diff += zp_absolute(zp_minus(&ztmp,
					&zp0[i], &zp1[i]));
			break;
		default:
			fatal("%s(): I dont understand your type", __F);
			break;
	}
	return (diff/(double) ap->vt_size);
}

void
zvec_info(const struct vector_t *vp)
{
	int	i;
	struct duplex_t *dux;

	vec_check(__F, vp, SP_DUPLEX);
	(void) printf("--zvec_info() duplex_t vector %s--\n", vp->vt_name);
	dux = zvec(vp);
	for (i=0;i < vp->vt_size; i++, dux++) {
			(void) printf("%7.3f %7.3f\n", dux->real, dux->imag);
	}
	(void) printf("+++\n");
	return;
}
void
vec_mnull(int  num,...)
{
	int     i;
	va_list ap;
	struct vector_t *p;
	va_start(ap, num);

	for (i=0;i<num;i++)  {
		p = va_arg(ap, struct vector_t *);
		vec_null(p);
	}
	va_end(ap);
	return;
}

void
vec_null(struct vector_t *vp)
{
	int	size;

	vec_check(__F, vp, -1);
	switch (vp->vt_type) {
		case SP_INTEGER:
			size = sizeof(int) * vp->vt_size; break;
		case SP_SINGLE:
			size = sizeof(float) * vp->vt_size; break;
		case SP_DOUBLE:
			size = sizeof(double) * vp->vt_size; break;
		case SP_SIMPLEX:
			size = sizeof(struct simplex_t) * vp->vt_size; break;
		case SP_DUPLEX:
			size = sizeof(struct duplex_t) * vp->vt_size; break;
	}
	memset((void *) vp->vt_data, 0x00, (size_t) size);
	return;
}


void
vec_setbase(struct vector_t *vp, int base)
{
	vec_check(__F, vp, -1);
	vp->vt_base = base;
	return;
}

inline double *
dvec_at(const struct vector_t *vp, int no)
{
	double	*dp;
	int	off;

	vec_check(__F, vp, SP_DOUBLE);
	dp = dvec(vp);

	off = (no - vp->vt_base);

	if (off < 0 || off >= vp->vt_size)
		fatal("dvec_at(\"%s\",%3d):is out of range!!", vp->vt_name, no);

	dp += off;
	return dp;
}

inline struct duplex_t *
zvec_at(const struct vector_t *vp, int no)
{
	struct duplex_t	*duxp;
	int	off;

	vec_check(__F, vp, SP_DUPLEX);
	duxp = zvec(vp);

	off = (no - vp->vt_base);
	if (off < 0 || off >= vp->vt_size)
		fatal("zvec_at(\"%s\",%3d):is out of range!!", vp->vt_name, no);

	duxp += off;
	return duxp;
}


void
dvec_set(struct vector_t *vp, double set)
{
	double	*dp;
	int	i;

	vec_check(__F, vp, SP_DOUBLE);

	dp = dvec(vp);
	for (i=0;i<vp->vt_size; i++)
		dp[i] = set;

	return;
}

void
zvec_set(struct vector_t *vp, struct duplex_t set)
{
	struct duplex_t *dp;
	int	i;

	vec_check(__F, vp, SP_DUPLEX);
		
	dp = zvec(vp);
	for (i=0;i<vp->vt_size; i++)
		dp[i] = set;

	return;
}

void
vec_printatfile(const struct vector_t *vp, const char *fname)
{
	int	i;
	FILE	*fp;

	int	*ip;
	float	*sp;
	double	*dp;
	struct simplex_t *cp;
	struct duplex_t *zp;

	vec_check(__F, vp, -1);

	fp = fopen(fname, "w");
	if (fp == NULL)
		fatal("vec_printatfile(\"%s\", \"%s\"): can not fopen",
			vp->vt_name, fname);

	switch (vp->vt_type) {
		case SP_INTEGER:
			ip = ivec(vp);
			for (i=0;i<vp->vt_size;i++)
				fprintf(fp, "%10d\n", ip[i]);
			break;
		case SP_SINGLE:
			sp = svec(vp);
			for (i=0;i<vp->vt_size;i++)
				fprintf(fp, "%12.6e\n", sp[i]);
			break;
		case SP_DOUBLE:
			dp = dvec(vp);
			for (i=0;i<vp->vt_size;i++)
				fprintf(fp, "%12.6e\n", dp[i]);
			break;
		case SP_SIMPLEX:
			cp = cvec(vp);
			for (i=0;i<vp->vt_size;i++)
				fprintf(fp, "%12.6e %12.6e\n",
					cp[i].real, cp[i].imag);
			break;
		case SP_DUPLEX:
			zp = zvec(vp);
			for (i=0;i<vp->vt_size;i++)
				fprintf(fp, "%12.6e %12.6e\n",
					zp[i].real, zp[i].imag);
			break;
	}
	fclose(fp);
	return;
}

void
vec_printatfiles(const struct vector_t *vp, const char *fname)
{
	vec_check(__F, vp, -1);
	switch (vp->vt_type) {
		case SP_INTEGER:
			vec_printatfile(vp, fname);	/* same as .. */
			break;
		case SP_SINGLE:
			vec_printatfile(vp, fname);	/* same as .. */
			break;
		case SP_DOUBLE:
			vec_printatfile(vp, fname);	/* same as .. */
			break;
		case SP_SIMPLEX:
			vec_plexprintatfiles(vp, fname);
			break;
		case SP_DUPLEX:
			vec_plexprintatfiles(vp, fname);
			break;
		default:
			fatal("vec_printatfiles():I dont understand your type");
			break;
	}
	return;
}

/*
 * A private function of vec_printatfiles()
 * DONT call this function directly
 */
void
vec_plexprintatfiles(const struct vector_t *vp, const char *filename)
{
	FILE	*fp[3];
	char	fname[3][132];
	char	*tails[3]={"real", "imag", "inten"};
	int	i;
	struct simplex_t *cp;
	struct duplex_t *zp;

	for (i=0;i<3;i++) {
		sprintf(fname[i], "%s.%s", filename, tails[i]);
		fp[i] = fopen(fname[i], "w");
		if (fp[i] == NULL)
			fatal("vec_cprintatfiles(\"%s\"): can not open",
				fname[i]);
	}

	switch (vp->vt_type) {
		case SP_SIMPLEX:
			cp = cvec(vp);
			for (i=0; i < vp->vt_size; i++) {
				fprintf(fp[0], "%10.6e\n", cp[i].real);
				fprintf(fp[1], "%10.6e\n", cp[i].imag);
				fprintf(fp[2], "%10.6e\n", cp_absolute(&cp[i]));
			}
			break;
		case SP_DUPLEX:
			zp = zvec(vp);
			for (i=0;i<vp->vt_size;i++) {
				fprintf(fp[0], "%10.6e\n", zp[i].real);
				fprintf(fp[1], "%10.6e\n", zp[i].imag);
				fprintf(fp[2], "%10.6e\n", zp_absolute(&zp[i]));
			}
	}
	
	for (i=0;i<3;i++)
		fclose(fp[i]);
	
	return;
}
