#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "matrix.h"
#include "finite.h"
#include "sparse.h"

struct mesh *meshes;
struct integral_t *intp[10];
int	entrycmp(const int *one, const int *another);

/*
 * Followings are used at the time of generating mesh
 */
struct for3dimcase *m_for3dim;
int	m_for3dimed;
float	m_coormin[3], m_coormax[3];


/*
 * This function makes the global meshes structure. You should know that two
 * meshes structure should not exist, I mean, only one meshes can
 * exist, so that you should not call this function twice.
 *
 * This routine will initialize the protos[].  The argument of this
 * mesh_init() will determine the the datatype of the whole programme. The
 * possible one of this argument is SP_SINGLE for float version program,
 * SP_DOUBLE for double version program and SP_SIMPLEX for float typed
 * complex space, and SP_DUPLEX for double typed complex space.
 *
 * This argument will, by that, determine the data type of the global K
 * matrix, and determine the datatype of the b matrix. This information is
 * stored at meshes->m_datatype, and this information will be queried at
 * various times, I think.
 */

void
mesh_init(const int sp_type)
{
	int	i, matsz;
	struct pairs *pp;

	for (i=0;i<10;i++)
		intp[0] = NULL;

	matsz = proto_init();

	if (meshes != NULL)
		fatal("%s(): you should not try to call mesh_init twice!!",
			__F);

	meshes = (struct mesh *) malloc((size_t)(sizeof(struct mesh)));
	if (meshes == NULL)
		memerr("%s(): at alloc'ting meshes", __F);

	/*
	 * initialize the global meshes structure
	 */
	meshes->m_spacedim = -1;	/* space dimension */
	meshes->m_ementsno = -1;	/* elment number in this mesh */
	meshes->m_nodesno = -1;		/* number of node in this mesh */
	meshes->m_nodes = NULL;
	meshes->m_ements = NULL;
	meshes->m_kmatrix = NULL;
	meshes->m_bvector = NULL;

	if (!ATYPE(sp_type))
		fatal("%s(): you gave wrong type of meshes->m_datatype", __F);

	meshes->m_datatype = sp_type;

	meshes->m_pairs = (struct pairs **)
		malloc(sizeof(struct pairs *) * matsz);
	if (meshes->m_pairs == NULL)
		memerr("%s(): malloc of meshes->m_pairs", __F);

	pp = (struct pairs *) malloc(sizeof(struct pairs) * matsz * matsz);
	if (pp == NULL)
		memerr("%s(): malloc of meshes->m_pairs[0]", __F);

	/*
	 * this structure pairs will be used at assemble
	 */
	for (i=0; i < matsz; i++) {
		meshes->m_pairs[i] = pp;
		pp += matsz;
	}

	return;
}


/*
 * This will create sparse matrix which will be used as a K matrix
 * You should call this function after providing nodes and elements information
 * This function only provides the space that is necessary to solve, and does
 * not make real K-matrix, see Kassemble. I think if time had been sufficient,
 * I could have coded this routine more compact and efficient.
 */
#define MAXIMUN_HERE 100

struct cols {
	int	col_col;
	int	col_norows;
	int	col_rows[MAXIMUN_HERE];
};

struct sparse_t *
create_kmatrix(void)
{
	struct ement_t *ep;
	struct sparse_t *sp;
	int	*colp, *rowp;
	int	i, j, k, l;
	nodno_t *nnp, *nnq;
	int	locus;
	struct cols **allcols, *colsp;
	int	cnt;
	int	(*compare)() = entrycmp;

	if (meshes == NULL)
		fatal("%s(): you must have called mesh_init()", __F);

	allcols = (struct cols **)
		malloc((size_t) (sizeof(struct cols *) * meshes->m_nodesno));
	if (allcols == NULL)
		memerr("%s(): allocing **allcols", __F);
	colsp = (struct cols *)
		malloc((size_t) (sizeof(struct cols) * meshes->m_nodesno));
	if (colsp == NULL)
		memerr("%s(): allocing *colsp", __F);

	for (i=0;i<meshes->m_nodesno;i++) {
		allcols[i] = colsp++;
		allcols[i]->col_col = (i+1);
		allcols[i]->col_norows = 0;
		for (j=0;j < MAXIMUN_HERE;j++) { /* is this sufficicient */
			allcols[i]->col_rows[j] = -1;
		}
	}

	for (i=0;i<meshes->m_ementsno;i++) {
		ep = meshes->m_ements[i];

		for (j=0, nnp = ep->e_nodes; j < ep->e_nodesno; j++) {
			for (k=0, nnq=ep->e_nodes; k < ep->e_nodesno; k++) {
				/* to check already counted */
				for (l=0; l<allcols[nnp[j]]->col_norows;l++) {
				    if (nnq[k] == allcols[nnp[j]]->col_rows[l])
						break;
				}
				if (l == allcols[nnp[j]]->col_norows) {
					locus = allcols[nnp[j]]->col_norows;
					allcols[nnp[j]]->col_rows[locus] =
						nnq[k];
					allcols[nnp[j]]->col_norows++;
				}
			}
		}
	}

	/*
	 * Sort
	 */
	cnt = 0;
	for (i=0;i<meshes->m_nodesno;i++) {
		colsp = allcols[i];
		qsort((int *) colsp->col_rows, colsp->col_norows,
			(size_t) sizeof(int), compare);
		cnt += colsp->col_norows;
	}

	/* cnt is nonzero number */
	switch (meshes->m_datatype) {
		case SP_SINGLE:
			sp = sparse(meshes->m_nodesno, meshes->m_nodesno, cnt,
				"Kmatrix[c]", SP_SINGLE);
			break;
		case SP_DOUBLE:
			sp = sparse(meshes->m_nodesno, meshes->m_nodesno, cnt,
				"Kmatrix[d]", SP_DOUBLE);
			break;
		case SP_SIMPLEX:
			sp = sparse(meshes->m_nodesno, meshes->m_nodesno, cnt,
				"Kmatrix[c]", SP_SIMPLEX);
			break;
		case SP_DUPLEX:
			sp = sparse(meshes->m_nodesno, meshes->m_nodesno, cnt,
				"Kmatrix[z]", SP_DUPLEX);
			break;
		default:
			fatal("%s", __F);
	}
	
	rowp  = sp->sp_rowx;
	colp  = sp->sp_colx;

	/*
	 * sparse is column-based
	 */
	cnt = 0;
	for (i=0;i < meshes->m_nodesno;i++) {
		colsp = allcols[i];
		colp[i] = cnt;
                for (j=0;j<colsp->col_norows; j++) {
			*rowp++ = colsp->col_rows[j];
                }
		cnt += colsp->col_norows;
	}
	colp[i] = cnt;

	free((void *) allcols[0]);
	free((void *) allcols);
	sp_mkroworder(sp);
	printf("+++++++++++++END   of create_kmatrix()+++++++++++++++++\n");
	return sp;
}

/*
 * makes the global b matrix, the right hand side of the FEM equation. You had
 * better to know that this function does not make real b matrix, that instead
 * this function only provides the b-matrix spaces, so that the real b matrix
 * will be constructed at bassemble() time.
 */
struct vector_t *
create_bvector(void)
{
	struct vector_t *vp;

	if (meshes == NULL)
		fatal("%s(): you must have called mesh_init()", __F);

	switch (meshes->m_datatype) {
		case SP_SINGLE:
			vp = vector(meshes->m_nodesno,"bvector[s]", SP_SINGLE);
			break;
		case SP_DOUBLE:
			vp = vector(meshes->m_nodesno,"bvector[d]", SP_DOUBLE);
			break;
		case SP_SIMPLEX:
			vp = vector(meshes->m_nodesno,"bvector[c]", SP_SIMPLEX);
			break;
		case SP_DUPLEX:
			vp = vector(meshes->m_nodesno,"bvector[z]", SP_DUPLEX);
			break;
		default:
			fatal("%s(): wrong type!!", __F);
	}

	return vp;
}


int
entrycmp(const int *first, const int *second)
{
	int one, two;

	one = (int) *first;
	two = (int) *second;

	if (one > two)
		return 1;
	else if (one < two)
		return -1;
	printf("first=%3d second=%3d\n", one, two);
	fatal("entrycmp(): TWO ENTRIES ARE SAME!!");
	/* NOTREACHED */
	return 0;
}


void
create_smalls(const struct proto_t *proto, struct matrix_t **retmatf,
	struct matrix_t **bmatf)
{
	int	size;

	struct matrix_t *smallk, *smallb;
	if (proto == NULL)
		fatal("%s(): with null proto pointer!!", __F);

	size = proto->p_nodsno;	/* # of the surrounding nodes */
	switch (meshes->m_datatype) {
		case SP_SIMPLEX:
		case SP_DUPLEX:
			smallk = matrix(size, size, "smallk", SP_DUPLEX);
			smallb = matrix(size, 1,    "smallb", SP_DUPLEX);
			break;
		case SP_SINGLE:
		case SP_DOUBLE:
			smallk = matrix(size, size, "smallk", SP_DOUBLE);
			smallb = matrix(size, 1,    "smallb", SP_DOUBLE);
			break;
		default:
		    fatal("%s(): bizare data type",__F);
	}
	*retmatf = smallk;
	*bmatf   = smallb;
	return;
}


float
get_centercoor(const int elemno, const int whatcoor)
{
	float	tmp;
	struct node_t *np;
	struct ement_t *ep;
	nodno_t *nnp;
	int	i;

	ep = meshes->m_ements[elemno];
	nnp = ep->e_nodes;

	tmp = 0.0;
	for (i=0; i< ep->e_nodesno;i++) {
		np = meshes->m_nodes[nnp[i]];
		tmp += np->n_coor[whatcoor];
	}

	tmp /= (float) i;
	return tmp;
}
