#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "common.h"
#include "finite.h"

int	node_compare(const nodno_t *one, const nodno_t *two);
static int	tri_readnode(const char *filename);
static int	tri_readelement(const char *filename);
void		tri_reduce(void);
static void	tri_mmaxspan(void);
static double	nodespan(const struct node_t *np0, const struct node_t *np1);
double	cmp_minlen, cmp_maxlen, cmp_maxmin;


/*
 * read the nodes and emement information that are stored at the char *fname
 * file.  The files should have the format of the .grid and .ele format of the
 * well-known triangle package.  Therefore if you want to use the grids
 * information which was produced by triangle, you can use this function.
 * Remember that you must not provide whole file name. After reading .node and
 * .ele file, this function call tri_reduce() to shorten the band width of the
 * K matrix.
 */
int
tri_getmesh(const char *fname)
{
	char	nodef[64];
	char	elef[64];

	if (meshes == NULL)
		fatal("%s():You should call mesh_init() before!", __F);

	strcpy(nodef, fname);
	strcat(nodef, ".node");
	strcpy(elef, fname);
	strcat(elef, ".ele");

	tri_readnode(nodef);
	tri_readelement(elef);
	tri_reduce();
	mk_nearements();

	return 0;
}


static int
tri_readnode(const char *nodef)
{
	FILE	*fp;
	char	line[BUFSIZ];
	int	notused;
	struct node_t *np;
	static int nodcnt = 0;
	float	tfloat;
	int	nodeno;
	int	ind;
	char	*p;
	int	tmp, i;

	fp = fopen(nodef, "r");
	if (fp == NULL)
		fatal("%s(): getnode(%s)!!", __F, nodef);
	
	if (fgets(line, BUFSIZ, fp)) {
		sscanf(line, "%d %d %d %d", &meshes->m_nodesno,
			&meshes->m_spacedim, &notused,  &notused);
	} else {
		fatal("%s(): %d line", __F, __LINE__);
	}

	/*
	 * XXX
	 */
	meshes->m_nodes = mesh_nodesbuf(meshes->m_nodesno);

	while (fgets(line, BUFSIZ, fp)) {
		p = (char *) line;
		if (*p == '#')
			break;
		p = skipwhites(p);
		nodeno = atoi(p);/* if failed, return atoi() returns zero */
		if (nodeno == 0)
			fatal("%s():Nodeno is bad!! while reading %s",
				__F, line);

		nodeno--;	/* index starts with 0 */
		np = meshes->m_nodes[nodeno];
		np->n_no = nodeno;

		for (i = 0; i < 3; i++)	/* clear its coors */
			np->n_coor[i] = 0.0;

		p = skipnumer(p);
		for (i = 0; i < meshes->m_spacedim; i++) {
			p = skipwhites(p);
			tmp = sscanf(p, "%f", &tfloat); /* XXX */
			np->n_coor[i] = tfloat;
			if (tmp - 1)
			    fatal("%s():\"%s\":reading coors[%d] of node %d ",
					__F, line, i, nodeno);
			p = skipnumer(p);
		}
		/*
		 * End of reading of coordinates
		 */

		np->n_bdflag = 0;

		p = skipwhites(p);
		for (i = 0; i< 1; i++) {
			for (; !isdigit(*p) && *p; p++)	/* skip to a number */
				/* void */ ;
			tmp = sscanf(p, "%d", &ind);
			if (tmp - 1)
				fatal("%s(): Wrong format in bd indi..", __F);
			/*
			 * XXX ---------
			ind--;
			 */
			if (ind < 0 || ind > 31) {
				printf("%s(): less than 0 or greater ", __F);
				printf(" than max ind %d: while reading %s\n",
					31, line);
				exit(1);
			}
			if (ind != 0) {
				ind--;
				setbit(np->n_bdflag, ind);
			}
		}
		if (i != 1)
			fatal("%s(): node indicater is not sufficient!!", __F);
		nodcnt++;
	}
	fclose(fp);
	if (nodcnt != meshes->m_nodesno)
		fatal("%s(): nodcnt is not correct!!", __F);
	return 0;
}

static int
tri_readelement(const char *fname)
{
	FILE *fp;
	static int elemcnt=0;
	char	line[BUFSIZ];
	int	i;
	struct ement_t *ep;
	char	*p;
	int	eno;
	int	tmp;
	int	gnode;
	nodno_t	*glob;
	int	surround;
	int	notused;

	fp = fopen(fname, "r");
	if (fp == NULL)
		fatal("%s(): fopen(\"%s\")", __F, fname);

	p = fgets(line, BUFSIZ, fp);
	if (p == NULL)
		fatal("%s(): %d line in gettri.c", __F, __LINE__);
	
	tmp = sscanf(p, "%d %d %d", &meshes->m_ementsno,
		&surround, &notused);

	if ((tmp - 3))
		fatal("%s(): Wrong element number", __F);

	meshes->m_ements = mesh_ementsbuf(meshes->m_ementsno);


	while (fgets(line, BUFSIZ, fp)) {
		p = (char *) line;
		if (*p == '#')
			break;
		p = skipwhites(p);
		eno = atoi(p);

		eno--;	/* index starts at 0 */
		ep = meshes->m_ements[eno];
		ep->e_no = eno;
		ep->e_nodesno = surround;
		ep->e_proto = NULL;

		switch (surround) {
			case 3:
				ep->e_proto = getproto("ElmT3n2D");
				break;
			case 6:
				ep->e_proto = getproto("ElmT6n2D");
				break;
			default:
				fatal("%s(): %d line in gettri", __F, __LINE__);
		}

		for (;isdigit(*p); p++)	/* skip element no */
			/* void */;

		p = skipwhites(p);

		glob = (nodno_t *)
			malloc((size_t)(sizeof(nodno_t)*ep->e_proto->p_nodsno));
		if (glob == NULL)
			memerr("%s(): mallocing glob correspondants", __F);

		ep->e_nodesno = ep->e_proto->p_nodsno;
		ep->e_nodes = glob;
		for (i=0; i< ep->e_proto->p_nodsno; i++) {
			p = skipwhites(p);
			tmp = sscanf(p, "%d", &gnode);
			if (tmp != 1)
				fatal("%s(): reading global corres.", __F);
			gnode--;
			ep->e_nodes[i] = gnode;
			p = skipnumer(p);
		}
		elemcnt++;
	}
	if (meshes->m_ementsno != elemcnt)
		fatal("%s(): insufficient element info", __F);
	fclose(fp);
	return 0;
}

int
node_compare(const nodno_t *one, const nodno_t *two)
{
	struct node_t *none, *ntwo;
	float	x1, y1, x2, y2;
	int	xcomp, ycomp;
	double	delx, dely, tlen;

	none = meshes->m_nodes[*one];
	ntwo = meshes->m_nodes[*two];

	x1 = none->n_coor[0]; y1 = none->n_coor[1];
	x2 = ntwo->n_coor[0]; y2 = ntwo->n_coor[1];

	delx = x1-x2; dely = y1-y2;
	tlen = sqrt(delx*delx + dely*dely);

	xcomp = x1 > x2 ? 1 : x1 < x2 ? -1 : 0;
	ycomp = y1 > y2 ? 1 : y1 < y2 ? -1 : 0;
#if 1
	if (tlen < cmp_maxmin)
		return 0;
#endif
	if (xcomp)
		return xcomp;
	if (ycomp)
		return ycomp;
	return 0;
}

#if BACKUP
static int
node_compare(const nodno_t *one, const nodno_t *two)
{
	struct node_t *none, *ntwo;
	float	x1, y1, x2, y2;
	int	xcomp, ycomp;

	none = meshes->m_nodes[*one];
	ntwo = meshes->m_nodes[*two];

	x1 = none->n_coor[0]; y1 = none->n_coor[1];
	x2 = ntwo->n_coor[0]; y2 = ntwo->n_coor[1];

	xcomp = x1 > x2 ? 1 : x1 < x2 ? -1 : 0;
	ycomp = y1 > y2 ? 1 : y1 < y2 ? -1 : 0;
	if (xcomp)
		return xcomp;
	if (ycomp)
		return ycomp;
	return 0;
}
#endif

/*
 * Renumbers the nodes in order to decrease the band width of the global
 * K-matrix.  The used algorithm is not good, only is a simple one.
 * Please pay attention on the qsort():
 */

void
tri_reduce(void)
{
	struct node_t **newnodes;
	struct ement_t *ep;
	nodno_t	*nodesmap, *invermap, *nnp;
	int	(*compare)();
	int	i, j;

	if (meshes == NULL)
		fatal("%s(): meshes == NULL", __F);
	if (meshes->m_ements == NULL) 
		fatal("%s(): meshes->m_ements == NULL", __F);

	tri_mmaxspan();

	nodesmap = (nodno_t *) malloc(sizeof(nodno_t) * meshes->m_nodesno);
	invermap = (nodno_t *) malloc(sizeof(nodno_t) * meshes->m_nodesno);

	compare = node_compare;

	for (i=0; i < meshes->m_nodesno; i++)
		nodesmap[i] = i;

	qsort(nodesmap, meshes->m_nodesno, sizeof(nodno_t), compare);
	for (i=0; i < meshes->m_nodesno; i++)
		invermap[nodesmap[i]] = i;


	/*
	 * meshes->m_nodes[i] change
	 */

	newnodes = (struct node_t **)
		malloc(sizeof(struct node_t *) * meshes->m_nodesno);
	if (newnodes == NULL)
		memerr("%s(): newnodes == NULL", __F);

	for (i=0; i < meshes->m_nodesno; i++) {
		newnodes[i] = meshes->m_nodes[nodesmap[i]];
		newnodes[i]->n_no = i;
	}
	free(meshes->m_nodes);
	meshes->m_nodes = newnodes;

	for (i=0;i < meshes->m_ementsno; i++) {
		ep = meshes->m_ements[i];
		nnp = ep->e_nodes;
		for (j=0; j < ep->e_nodesno; j++)
			nnp[j] = invermap[nnp[j]];
	}

	free(nodesmap);
	free(invermap);
	printf("%s(): NODE NUMBERS ARE CHANGED\n", __F);
	return;
}

/*
 * find max and min lengths of the all the sides of the elements.
 */

static void
tri_mmaxspan(void)
{
	int	i, j, k;
	nodno_t *nnp;
	struct ement_t *ep;
	struct node_t *np0, *np1;
	double	tlen, tlenmin, tlenmax;

	tlenmin = +1.0e+10;
	tlenmax = -1.0e+10;
	for (k=0; k < meshes->m_ementsno;k++) {
		ep = meshes->m_ements[k];
		nnp = ep->e_nodes;
		for (i=0; i < ep->e_nodesno;i++) {
			np0 = meshes->m_nodes[nnp[i]];
			for (j=0; j < i; j++) {
				np1 = meshes->m_nodes[nnp[j]];
				tlen = nodespan(np0, np1);
				if (tlen > tlenmax)
					tlenmax = tlen;
				if (tlen < tlenmin)
					tlenmin = tlen;
			}
		}
	}
	cmp_maxlen = tlenmax;
	cmp_minlen = tlenmin;
	if (0)
		cmp_maxmin = (cmp_minlen+cmp_maxlen)/2.0;
	else if (0)
		cmp_maxmin = cmp_maxlen * 0.3;
	else
		cmp_maxmin = 0.0;
	printf("%s(); finish\n", __F);
	return;
}

static double
nodespan(const struct node_t *np0, const struct node_t *np1)
{
	double	x, y, len;

	x = np0->n_coor[0] - np1->n_coor[0];
	y = np0->n_coor[1] - np1->n_coor[1];

	len = sqrt(x*x+y*y);
	return len;
}
