#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include "common.h"
#include "options.h"

struct options *optroot = NULL;

struct options *opt_malloc(void);
void	 	opt_free(struct options *);
void	 	opt_err(int linecnt, const char *fmt, ...);
struct options *opt_findopt(const char *optionname, int optiontype);

static char *opttype_name[] = {
	NULL,
	"integer type option",
	"real type option",
	"string type option",
	"timestep type option",
	NULL
};

/*
 * Private function
 */
struct options *
opt_malloc(void)
{
	struct options *op;

	op = (struct options *) malloc((size_t)sizeof(struct options));
	if (op == NULL)
		memerr("opt_malloc():[0]");

	memset((void *) op, 0x00, (size_t)(sizeof(struct options)));
	return op;
}

void
opt_add(type, name, explain, deflt, range)
	const char *name, *explain, *deflt, *range;
	int	type;
{
	struct options *op, *oq;

	op = opt_malloc();
	if (optroot == NULL) {
		op->op_next = NULL;
		op->op_prev = op;	/* self */
		optroot = op;		/* root of the list */
	} else {
		for (oq = optroot; oq; oq = oq->op_next) {
			if (!strcmp(oq->op_name, name))
				fatal("%s():duplicated option \"%s\"",
					__F, name);
			if (oq->op_next == NULL)	/* We met last option */
				break;
		}
		oq->op_next = op;
		op->op_prev = oq;
	}
	op->op_type = type;

	if (explain)
		op->op_explain  = strdup(explain);
	if (name) {
		op->op_name	= strdup(name);
		op->op_namelen  = strlen(name);
	}

	if (deflt)
		op->op_default	= strdup(deflt);
	if (range)
		op->op_range	= strdup(range);
	op->op_timestep = NULL;
	return;
}

void
opt_free(struct options *op)
{
	struct options *opp, *opn;
	if (op == optroot) {
		printf("%s(): I can't free option root \"%s\"", __F,
			op->op_name);
		return;
	}
	if (op->op_name)
		free(op->op_name);
	if (op->op_default)
		free(op->op_default);
	if (op->op_range)
		free(op->op_range);
	if (op->op_explain)
		free(op->op_explain);

	opp = op->op_prev;
	opn = op->op_next;
	if (opp)
		opp->op_next = op->op_next;
	if (opn)
		opn->op_prev = op->op_prev;
	free(op);
	return;
}

void
opt_print(void)
{
	struct options *op;

	for (op=optroot; op; op = op->op_next) {
		printf("%s++++%s\n", op->op_default, op->op_name);
	}
	return;
}

void
opt_read(const char *filename)
{
	FILE	*fp;
	char	line[128], *p, *q, *target;
	struct options *op;
	int	len, linecnt;

	if (optroot == NULL)
		fatal("%s(\"%s\"): You have not setup options yet", __F,
			filename);

	fp = fopen(filename, "r");
	if (fp == NULL)
		fatal("%s(\"%s\"): Can not open this file", __F, filename);

	linecnt = 0;
	while (fgets(line, 128, fp)) {
		linecnt++;
		p = (char *) line;
		if (*p == '#' || *p == '!')	/* skip comment line */
			continue;
		if (!strncmp(line, "ok", 2))
			break;
		/*
		 * Let's find target name
		 */
		if (strncmp(line, "set ", 4) == 0)
			p = (char *) line + 4;

		target = skipwhites(p);	/* maybe start of target name */
		q = index(target, '=');
		if (q == NULL) {
			fprintf(stderr,
				"%s(\"%s\"):What is %s? at line %d\n", __F,
				filename, target, linecnt);
			continue;
		}

		/* back trace to non-space to meet the end of the target name */
		p = q;
		p++;
		for (--q; isspace(*q); --q)
			/* void */;
		*++q = 0;	/* terminates the target name */
		len = strlen(target);

		/* Let's find the corresponding option */
		for (op=optroot; op; op = op->op_next) {
			if (len == op->op_namelen
			    && strcmp(op->op_name, target) == 0)
				break;
		}

		if (op == NULL) {
			fprintf(stderr,
				"%s(\"%s\"):What is %s? at line %d\n", __F,
				filename, target, linecnt);
			continue;
		}
		if (op->op_already != 0)
			fprintf(stderr,
				"%s(\"%s\"):Duplicated %s at %d\n", __F,
					filename, op->op_name, linecnt);

		op->op_already = linecnt;
		p = skipwhites(p);
		q = p + strlen(p);
		for (--q; isspace(*q); --q)
			/* void */;
		*++q = 0;	/* terminates the value name */

		if (op->op_default)
			free((void *) op->op_default);
		op->op_default = strdup(p);
	}
	return;
}

void
opt_check(void)
{
	struct options *op;
	char	*p, *suspected;
	int	istart, iend, ival;
	float	fstart, fend, fval;
	int	readn;

	for (op = optroot; op; op = op->op_next) {
		p = suspected = op->op_range;

		if (p == NULL)
			continue;

		switch (op->op_type) {
		case OPT_NONE:
			break;
		case OPT_INTEGER:
			readn = sscanf(op->op_default, "%d", &ival);
			if (readn != 1) {
				opt_err(op->op_already, "value %s for %s",
					op->op_default, op->op_name);
				continue;
			}
			if (*p++ != 'I') {
				opt_err(op->op_already, "range %s for %s=%s",
					p, op->op_name, op->op_default);
				continue;
			} else if (*p != '[' || *p++ == 0)
				continue;

			readn = sscanf(p, "%d:%d", &istart, &iend);
			if (readn != 2 || istart > ival || ival > iend) {
				opt_err(op->op_already, "range %s for %s=%s",
					(p-2), op->op_name, op->op_default);
				continue;
			}
			break;
		case OPT_REAL:
			readn = sscanf(op->op_default, "%f", &fval);
			if (readn != 1) {
				opt_err(op->op_already, "value %s for %s=%s",
				    op->op_default, op->op_name,
					op->op_default);
				continue;
			}
			if (*p++ != 'R') {
				opt_err(op->op_already, "range %s for %s=%s",
					p, op->op_name, op->op_default);
				continue;
			} else if (*p != '[' || *p++ == 0)
				continue;

			readn = sscanf(p, "%f:%f", &fstart, &fend);
			if (readn != 2 || fstart > fval || fval > fend) {
				opt_err(op->op_already, "range %s for %s=%s",
					(p-2), op->op_name, op->op_default);
				break;
			}
			break;
		case OPT_STRING:
			if (op->op_default == NULL) {
				opt_err(op->op_already, "NULL string for %s",
					op->op_name);
				continue;
			}
			break;
		case OPT_TIMESTEP:
			break;
		default:
			opt_err(op->op_already, "--I dont know your option %s",
				op->op_name);
			exit(1);
		}
	}
	return;
}

void
opt_err(int linecnt, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);

	(void)fprintf(stderr, "opt_check() error: Wrong ");
	(void)vfprintf(stderr, fmt, ap);
	if (linecnt)
		(void) fprintf(stderr, "--read from %dth line\n", linecnt);
	else
		(void) fprintf(stderr, "\n");
	va_end(ap);
	return;
}

struct options *
opt_findopt(const char *optname, int type)
{
	struct options *op;

	if (type > OPT_MAX) {
		(void) fprintf(stderr,
			"%s(\"%s\"): The type of this option is OOB\n", __F,
			optname);
		exit(1);
	}
	for (op = optroot; op; op = op->op_next)
		if (!strcmp(op->op_name, optname))
			break;
	if (op == NULL) {
		(void) fprintf(stderr,
			"%s(\"%s\"):There is no such option\n", __F,
			optname);
		exit(1);
	}
	if (op->op_type != type) {
		(void) fprintf(stderr,
			"%s(\"%s\"):This option is not %s\n", __F,
				op->op_name, opttype_name[type]);
		exit(1);
	}

	return op;
}

int
opt_getint(const char *optname)
{
	struct options *op;
	int	retval, readn;

	op = opt_findopt(optname, OPT_INTEGER);

	readn = sscanf(op->op_default, "%d", &retval);
	if (readn != 1) {
		(void) fprintf(stderr,
			"%s(\"%s\"):What is this?  %s", __F,
				op->op_name, op->op_default);
		exit(1);
	}
	return retval;
}

float
opt_getreal(const char *optname)
{
	struct options *op;
	float	retval;
	int	readn;

	op = opt_findopt(optname, OPT_REAL);

	readn = sscanf(op->op_default, "%f", &retval);
	if (readn != 1) {
		(void) fprintf(stderr,
			"%s(\"%s\"):What is this?  %s", __F,
				op->op_name, op->op_default);
		exit(1);
	}
	return retval;
}

char *
opt_getstr(const char *optname)
{
	struct options *op;

	op = opt_findopt(optname, OPT_STRING);
	return op->op_default;
}

struct timestep *
opt_gettimestep(const char *optname)
{
	struct options *op;
	struct timestep *ts;
	char	*p;
	int	readn;

	op = opt_findopt(optname, OPT_TIMESTEP);

	if (op->op_timestep != NULL) {
		return op->op_timestep;
	}

	ts = op->op_timestep = (struct timestep *)
		malloc((size_t)sizeof(struct timestep));
	if (op->op_timestep == NULL) {
		fprintf(stderr, "%s(\"%s\"): mallocing timestep", __F,
			op->op_name);
		exit(1);
	}
	/* parsing of the string of op->op_default */
	for (p = op->op_default; *p && *p != '='; p++)
		/* void */ ;

	if (*p++ == 0)
		goto error;

	readn = sscanf(p, "%f", &ts->ts_step);
	if (readn != 1)
		goto error;
	for (p = op->op_default; *p && *p != '['; p++)
		/* void */ ;
	if (*p++ == 0)
		goto error;
	readn = sscanf(p, "%f, %f", &ts->ts_start, &ts->ts_end);
	if (readn != 2)
		goto error;
	ts->ts_number = 0;
	ts->ts_valid = 1;
	ts->ts_current = ts->ts_start;
	return ts;
error:
	fprintf(stderr, "%s(\"%s\"): wrong format %s", __F,
		optname, op->op_default);
	exit(1);
}
