#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "sparse.h"
#include "solver.h"

/*
 * Remember that the result term is located in the leftest position
 */

void	atimes(struct vector_t *xp, struct sparse_t *sp,
			const struct vector_t *vp, int indicator);
double	snrm(const struct vector_t *vp, int indicator);
/*
 * Bi-conjugate Gradient
 */

#define EPS 1.0e-14

void
sp_biconjgradient(struct sparse_t *sp, struct vector_t *b, struct vector_t *x,
	int itol, double tol, int itmax, int *iter, double *err)
{
	struct vector_t *p, *pp, *r, *rr, *z, *zz;
	double ak, akden, bk, bkden, bknum;
	double bnrm, dxnrm, xnrm, zm1nrm, znrm;

	if (vec_isdouplets(b, x))
		fatal("sp_biconjgradient(sp, b, x, ..): b and x are dif. type");

	p  = vec_puredup(b);
	pp = vec_puredup(b);
	r  = vec_puredup(b);
	rr = vec_puredup(b);
	z  = vec_puredup(b);
	zz = vec_puredup(b);

	*iter = 0;
	atimes(r, sp, x, 0);
	vec_minus(r, b, r);
	vec_copy(rr, r);
	
	znrm = 1.0;
	if (itol == 1) {
		bnrm = snrm(b, itol);
#if 0
		precond_simple(z, sp, r);
#endif
	} else if (itol == 2) {
		precond_simple(z, sp, b);
		bnrm = snrm(z, itol);
	} else if (itol == 3 || itol == 4) {
		precond_simple(z, sp, b);
		bnrm = snrm(z, itol);
		precond_simple(z, sp, r);
		znrm = snrm(z, itol);
	} else
		fatal("sp_biconjgradient(): Illegal itol number");

	precond_simple(z, sp, r);

	while (*iter <= itmax) {
		++(*iter);

		zm1nrm = znrm;
		precond_simple(zz, sp, rr);
		bknum = vec_ddotprod(z, rr);

		if (*iter == 1) {
			vec_copy(p, z);
			vec_copy(pp, zz);
		} else {
			bk = bknum / bkden;
			vec_dcombi(p, bk, p, 1.0, z);
			vec_dcombi(pp, bk, pp, 1.0, zz);
		}
		bkden = bknum;
		atimes(z, sp, p, 0);

		akden = vec_ddotprod(z, pp);

		ak = bknum/akden;

		atimes(zz, sp, pp, 1);

		vec_dcombi(x,  1.0, x, ak, p);
		vec_dcombi(r,  1.0, r, -1.0 * ak, z);
		vec_dcombi(rr, 1.0, rr, -1.0 * ak, zz);

		precond_simple(z, sp, r);

		if (itol == 1 || itol == 2) {
			znrm = 1.0;
			*err = snrm(r, itol)/bnrm;
		} else if (itol == 3 || itol == 4) {
			znrm = snrm(z, itol);
			if (fabs(zm1nrm-znrm) > EPS*znrm) {
				dxnrm = fabs(ak) * snrm(p, itol);
				*err = znrm/fabs(zm1nrm-znrm)*dxnrm;
			} else {
				*err = znrm/bnrm;
				continue;
			}

			xnrm = snrm(x, itol);
			if (*err <= 0.5*xnrm)
				*err /= xnrm;
			else {
				*err = znrm/bnrm;
				continue;
			}
		}
		printf("iter=%4d err=%12.6f\n", *iter, *err);
		if (*err <= tol)
			break;
	}
	vec_mfree(6, p, pp, r, rr, z, zz);
	return;
}

void
atimes(struct vector_t *xp,
	struct sparse_t *sp, const struct vector_t *vp, int indicator)
{
	if (indicator)
		sp_transdotvec(xp, 0, sp, vp);
	else
		sp_dotvec(xp, 0, sp, vp);
	return;
}

double
snrm(const struct vector_t *vp, int indicator)
{
	double	retval;
	int	idx;

	if (indicator <= 3) 
		retval = vec_getnorm2(vp);
	else {
		idx = vec_getlargestindex(vp);
		retval = vec_getabsoluteat(vp, idx);
	}
	return retval;
}

