#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "solver.h"
#include "rtc.h"

int	iter_check(int Iter, double rNorm, double bNorm);

/*
 * It can solve real-symmetric case
 */

struct vector_t *
it_congradreal(struct sparse_t *sp, struct vector_t *x, struct vector_t *b,
	int itno, double omega, int pcd_type)
{
	int	it;
	int	dimen;
	double	alpha, beta, rho, rho_old;
	double	bnorm;
	struct vector_t *r, *tmp0, *tmp1, *q, *p, *z;
	struct sparse_t *diag, *diaginv;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	p    = vec_puredup(x);
	q    = vec_puredup(x);
	z    = vec_puredup(x);
	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(x);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_copy(r, vec_minus(tmp0, b, sp_dotvec(tmp1, 0, sp, x)));
	else
		vec_copy(r, b);

	diag = sp_getdiagsp(sp);
	diaginv = sp_getinvdiag(diag);

	it = 0;
	rho_old = 0.0;

	switch (pcd_type) {
	case PCD_JACOB:
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;
			vec_dcross(z, omega, sp_dotvec(tmp0, 0, diaginv, r));
			rho = vec_ddotprod(r, z);

			if (it == 1) {
				/* sp_putusingrow(diaginv); */
				vec_copy(p, z);
			} else {
				beta = rho/rho_old;
				vec_plus(p, z, vec_dcross(tmp0, beta, p));
			}

			sp_dotvec(q, 0, sp, p);
			alpha = rho/vec_ddotprod(p, q);


			vec_pplus(x, vec_dcross(tmp0, alpha, p));
			vec_mminus(r, vec_dcross(tmp0, alpha, q));

			rho_old = rho;
		}
		break;
	default:
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;

			rho = pow(vec_getnorm2(r), 2.0);

			if (it == 1)
				vec_copy(p, r);
			else {
				beta = rho/rho_old;
				vec_plus(p, r, vec_dcross(tmp0, beta, p));
			}

			sp_dotvec(q, 0, sp, p);
			alpha = rho/vec_ddotprod(p, q);


			vec_pplus(x, vec_dcross(tmp0, alpha, p));
			vec_mminus(r, vec_dcross(tmp0, alpha, q));

			rho_old = rho;
		}
		break;
	}
	
	vec_mfree(3, r, tmp0, tmp1);
	sp_mfree(2, diag, diaginv);
	return x;
}

struct vector_t *
it_normalcongradreal(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega, int pcd_type)
{
	int	it;

	double	alpha, beta, rho, rho_old;
	double	bnorm;
	int	dimen;
	struct vector_t *r, *p, *q, *s, *z;
	struct vector_t *tmp0;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	p    = vec_puredup(x);
	q    = vec_puredup(x);
	s    = vec_puredup(x);
	z    = vec_puredup(x);
	tmp0 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(x);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	rho_old = 0.0;
	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		sp_transdotvec(z, 0, sp, r);
		rho = pow(vec_getnorm2(z), 2.0);

		if (it == 1)
			vec_copy(p, z);
		else {
			beta = rho/rho_old;
			vec_plus(p, z, vec_dcross(tmp0, beta, p));
		}

		sp_dotvec(q, 0, sp, p);
		alpha = rho/pow(vec_getnorm2(q), 2.0);

		vec_pplus(x, vec_dcross(tmp0, alpha, p));
		vec_mminus(r, vec_dcross(tmp0, alpha, q));

		rho_old = rho;
	}
	vec_mfree(6, r, p, q, s, z, tmp0);
	return x;
}

struct vector_t *
it_normalcongradcomplex(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno, double omega, int pcd_type)
{
	int	it;

	double	alpha, beta, rho, rho_old;
	double	bnorm;
	int	dimen;
	struct vector_t *r, *p, *q, *s, *z;
	struct vector_t *tmp0;

	sp_symcheck(sp);
	iter_bontriplets(__F, sp, x, b);

	r    = vec_puredup(x);
	p    = vec_puredup(x);
	q    = vec_puredup(x);
	s    = vec_puredup(x);
	z    = vec_puredup(x);
	tmp0 = vec_puredup(x);

	dimen = x->vt_size;
	bnorm = vec_getnorm2(x);

	if (!iszero(vec_getnorm1(x)/(double) dimen))
		vec_minus(r, b, sp_dotvec(tmp0, 0, sp, x));
	else
		vec_copy(r, b);

	it = 0;

	rho_old = 0.0;
	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;

		sp_transdotvec(z, 1, sp, r);
		rho = pow(vec_getnorm2(z), 2.0);

		if (it == 1)
			vec_copy(p, z);
		else {
			beta = rho/rho_old;
			vec_plus(p, z, vec_dcross(tmp0, beta, p));
		}

		sp_dotvec(q, 0, sp, p);
		alpha = rho/pow(vec_getnorm2(q), 2.0);

		vec_pplus(x, vec_dcross(tmp0, alpha, p));
		vec_mminus(r, vec_dcross(tmp0, alpha, q));

		rho_old = rho;
	}
	vec_mfree(6, r, p, q, s, z, tmp0);
	return x;
}
