#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "vector.h"
#include "sparse.h"
#include "solver.h"
#include "precond.h"
#include "rtc.h"

/*
 * SHOULD BE symmetric or hermitian
 */

struct vector_t *
it_congrad(struct sparse_t *sp, struct vector_t *x, struct vector_t *b, int no)
{
	double	gamma, alpha;
	struct duplex_t zgamma, zalpha;
	struct duplex_t ztmp0, ztmp1, ztmp2;

	struct vector_t *atmp, *btmp, *ctmp;
	struct vector_t *r, *p, *r_, *p_;
	int	i;

	sp_symcheck(sp);
	if ((sp->sp_symflag & (SP_SYMMETRIC | SP_HERMITIAN)) == 0)
		fatal("%s(\"%s\"): sp should have SYMMETRY!!",__F, sp->sp_name);

	r  = vec_puredup(b);
	r_ = vec_puredup(b);
	p  = vec_puredup(b);
	p_ = vec_puredup(b);

	atmp  = vec_puredup(b);
	btmp  = vec_puredup(b);
	ctmp  = vec_puredup(b);

	vec_copy(r, b);
	vec_copy(p, r);

	switch (sp->sp_type) {
	case SP_SINGLE:
	case SP_DOUBLE:
		for (i=0; i < no ;i++) {
			vec_copy(r_, r);
			vec_copy(p_, p);

			atmp = sp_dotvec(atmp, 0, sp, p);
			alpha = vec_ddotprod(r, r)
				/ vec_ddotprod(atmp, p);
			vec_pplus(x, vec_dcross(ctmp, alpha, p));
			vec_mminus(r, vec_dcross(ctmp, alpha, atmp));

			gamma = vec_ddotprod(r, r)/ vec_ddotprod(r_, r_);
			vec_plus(p, r, vec_dcross(atmp, gamma, p_));

			if (vec_getnorm2(r) < 1.0e-5)
				break;
			if (verbose > 1) {
				if ((i % 100) == 0)
					printf("Iter=%d\n", i);
			} else if (verbose)
				printf("Iter=%d\n", i);
		}
		break;
	case SP_SIMPLEX:
	case SP_DUPLEX:
		for (i=0; i < no ;i++) {
			vec_copy(r_, r);
			vec_copy(p_, p);

			atmp = sp_dotvec(atmp, 0, sp, p);
			zalpha = *zp_slash(&ztmp2, vec_zdotprod(&ztmp0, r, r),
				vec_zdotprod(&ztmp1, atmp, p));

			vec_pplus(x, vec_zcross(ctmp, zalpha, p));
			vec_mminus(r, vec_zcross(ctmp, zalpha, atmp));

			zgamma = *zp_slash(&ztmp2, vec_zdotprod(&ztmp0, r, r),
				 vec_zdotprod(&ztmp1, r_, r_));

			vec_plus(p, r, vec_zcross(atmp, zgamma, p_));

			if (vec_getmeannorm2(r) < 1.0e-5)
				break;
			if (verbose > 1) {
				if ((i % 10) == 0)
					printf("Iter=%d\n", i);
			} else if (verbose)
				printf("Iter=%d\n", i);
		}
		break;
	default:
		fatal("%s():-----------+++++++++", __F);
	}
	return x;
}

/*
 * Bi-Conjugate Gradient Method
 */

int
it_bicongrad(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno)
{
	struct vector_t *r, *r0, *r_, *r_0;
	struct vector_t *p, *p_;
	struct vector_t *tmp0, *tmp1, *orthos, *hetero;

	double	dalpha, dbeta;
	double	bnorm;
	struct duplex_t	zalpha, zbeta, ztm0, ztm1;
	int	it;

	sp_symcheck(sp);
#if 1
	if (sp->sp_type == SP_SIMPLEX || sp->sp_type == SP_DUPLEX)
		if ((sp->sp_symflag & (SP_SYMMETRIC | SP_HERMITIAN)) == 0)
			fatal("%s(\"%s\"): sp should have SYMMETRY!!",
				__F, sp->sp_name);
#endif

	r   = vec_puredup(x);
	r0  = vec_puredup(x);
	r_  = vec_puredup(x);
	r_0 = vec_puredup(x);
	p   = vec_puredup(x);
	p_  = vec_puredup(x);

	tmp0 = vec_puredup(x);
	tmp1 = vec_puredup(x);
	orthos  = vec_puredup(x);
	hetero = vec_puredup(x);

	/*
	 * initialization
	 */
	vec_null(x);
#if 1
	precond_simple(x, sp, b);	/* guess x roughly!! */
#endif

	sp_dotvec(tmp0, 0, sp, x);
	vec_minus(tmp1, b, tmp0);

	vec_copy(r,  tmp1);
	vec_copy(r_, tmp1);
	vec_copy(p,  tmp1);
	vec_copy(p_, tmp1);

	bnorm = vec_getnorm2(x);

	it = 0;
	switch (sp->sp_type ) {
	case SP_SINGLE:
	case SP_DOUBLE:
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;
			orthos = sp_dotvec(orthos, 0, sp, p);
			hetero = sp_transdotvec(hetero, 0, sp, p_);
			
			dalpha = vec_ddotprod(r_, r)/vec_ddotprod(p_, orthos);

			vec_copy(r_0, r_);
			vec_copy(r0, r);

			vec_mminus(r,  vec_dcross(tmp0,  dalpha, orthos));
			vec_mminus(r_, vec_dcross(tmp0,  dalpha, hetero));

			dbeta = vec_ddotprod(r_, r)/vec_ddotprod(r_0, r0);

			vec_dcross(tmp0, dalpha, p);
			vec_pplus(x, tmp0);

			vec_crossd(p,   p,  dbeta);
			vec_crossd(p_,  p_, dbeta);
			vec_pplus(p,  r);
			vec_pplus(p_, r_);
		}
		break;
	case SP_SIMPLEX:
	case SP_DUPLEX:
		while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
			it++;
			orthos = sp_dotvec(orthos, 0, sp, p);
			hetero = sp_transdotvec(hetero, 1, sp, p_);
			
			zp_slash(&zalpha, vec_zdotprod(&ztm0, r_, r),
				vec_zdotprod(&ztm1, p_, orthos));

			vec_copy(r_0, r_);
			vec_copy(r0, r);

			vec_mminus(r,  vec_zcross(tmp0,  zalpha, orthos));
			vec_mminus(r_, vec_zcross(tmp0,  zalpha, hetero));

			zp_slash(&zbeta, vec_zdotprod(&ztm0, r_, r),
				vec_zdotprod(&ztm1, r_0, r0));

			vec_zcross(tmp0, zalpha, p);
			vec_pplus(x, tmp0);

			vec_crossz(p,   p,  zbeta);
			vec_crossz(p_,  p_, zbeta);
			vec_pplus(p,  r);
			vec_pplus(p_, r_);

		}
		break;
	default:
		fatal("%s(): impossible case..sp->sp_type", __F);
	}
	printf("vec_getnorm2(r) = %20.15f\n", vec_getnorm2(r));

	/*
	 * the no of free() is correct????
	 */
	vec_mfree(10, r, r0, r_, r_0, p, p_, tmp0, tmp1, orthos, hetero);
	return it;
}

/*
 * Bi-Conjugate Gradient Method
 * from laspack
 */

int
it_bicongradcomplex(struct sparse_t *sp, struct vector_t *x,
	struct vector_t *b, int itno)
{
	struct vector_t *r, *p, *ax;
	struct vector_t *vtmp;

	double	bnorm;
	struct duplex_t alpha, gamma, rnorm;
	struct duplex_t	ztmp;
	int	it;

	sp_symcheck(sp);
	if (sp->sp_type != SP_SIMPLEX && sp->sp_type != SP_DUPLEX)
		fatal("%s(\"%s\"): You can use this function at real-world!!",
			__F, sp->sp_name);
	if ((sp->sp_symflag & SP_COMPLEXSYM) == 0)
		fatal("%s(\"%s\"): sp should have COMPLEXSYM!!",
			__F, sp->sp_name);

	r   = vec_puredup(x);
	p   = vec_puredup(x);
	ax  = vec_puredup(x);
	vtmp= vec_puredup(x);

	bnorm = vec_getnormsq(b);
	vec_null(x);
#if 0
	precond_simple2nd(x, sp, b);	/* guess x roughly!! */
#endif
	sp_dotvec(ax, 0, sp, x);

	vec_minus(r, b, ax);
	vec_copy(p,  r);

	rnorm = complex_norm(r, r);

	it = 0;
	while (iter_check(it, vec_getnorm2(r), bnorm) && it < itno) {
		it++;
		sp_dotvec(ax, 0, sp, p);

		ztmp = complex_norm(ax, p);
		zp_slash(&alpha, &rnorm, &ztmp);

		vec_zcross(vtmp, alpha, p);
		vec_pplus(x, vtmp); 
		vec_zcross(vtmp, alpha, ax);
		vec_mminus(r, vtmp);

		ztmp = complex_norm(r, r);
		zp_slash(&gamma, &ztmp, &rnorm);
		rnorm = complex_norm(r, r); /* XXX */

		vec_plus(p, r, vec_zcross(vtmp, gamma, p));
	}

	vec_mfree(4, p, r, ax, vtmp);
	if (verbose == 0)
		printf("%4d: rnorm=%10.5e\n", it, vec_getnorm2(r));
	return it;
}

struct duplex_t
complex_norm(struct vector_t *a, struct vector_t *b)
{
	int	i;
	struct simplex_t *cx, *cy, cret, ctmp;
	struct duplex_t *zx, *zy, zret, ztmp;

	if (vec_isdouplets(a, b))
		fatal("%s(): cat not complex_norm()", __F);

	cret = simplex_null;
	zret = duplex_null;
	switch (a->vt_type) {
		case SP_SIMPLEX:
			cx = cvec(a);
			cy = cvec(b);
			for (i=0; i<a->vt_size; i++) {
				cp_pplus(&cret, cp_cross(&ctmp, &cx[i],&cy[i]));
			}
			zret.real = cret.real;
			zret.imag = cret.imag;
			break;
		case SP_DUPLEX:
			zx = zvec(a);
			zy = zvec(b);
			for (i=0; i<a->vt_size; i++) {
				zp_pplus(&zret, zp_cross(&ztmp, &zx[i],&zy[i]));
			}
			break;
		default:
			fatal("%s(%s): is neither SIMPLEX nor DUPLEX type", __F,
				a->vt_name);
	}
	return zret;
}
