#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "common.h"
#include "vector.h"
#include "sparse.h"
#include "solver.h"

double	escape_iter = 1.0e-5;
int
iter_check(int iter, double rnorm, double bnorm)
{
	int result;

	if (rnorm < escape_iter * bnorm
	    || (iszero(bnorm) && isone(1.0 + rnorm)))
		result = STOPNOW;
	else
		result = CONTINUE;

	if (verbose)
		printf("%4d: rnorm=%10.5e\n", iter, rnorm);

	return result;
}

void
iter_bontriplets(const char *function, const struct sparse_t *sp,
	const struct vector_t *ap, const struct vector_t *bp)
{
	if (sp == NULL || ap == NULL || bp == NULL)
		fatal("%s()/%s(): one of the triplets..is NULL",
			function, __F);
	if (sp->sp_type != ap->vt_type || ap->vt_type != bp->vt_type)
		fatal("%s()/%s(): the triplets..are not same typed",
			function, __F);
	if (sp->sp_issquare == 0)
		fatal("%s()/%s(): sparse_t is not square!!", function, __F);
	if (sp->sp_coln != ap->vt_size)
		fatal("%s()/%s(): sparse_t and vector have different size",
			function, __F);

	if (ap->vt_size != bp->vt_size)
		fatal("%s()/%s(): two vectors have different size",
			function, __F);

	return;
}

void
precond_simple(struct vector_t *xp, const struct sparse_t *sp,
	struct vector_t *bp)
{
	int	i, locus;

	float	*sx, *sb, *sdata;
	double	*dx, *db, *ddata;
	struct simplex_t *cx, *cb, *cdata;
	struct duplex_t *zx, *zb, *zdata;

	switch (sp->sp_type) {
	case SP_SINGLE:
		sx = svec(xp);
		sb = svec(bp);
		sdata = (float *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			sx[i] = (locus != -1 ? sb[i]/sdata[locus] : sb[i]);
		}
		break;
	case SP_DOUBLE:
		dx = dvec(xp);
		db = dvec(bp);
		ddata = (double *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			dx[i] = (locus != -1 ? db[i]/ddata[locus] : db[i]);
		}
		break;
	case SP_SIMPLEX:
		cx = cvec(xp);
		cb = cvec(bp);
		cdata = (struct simplex_t *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1)
				cx[i] = cb[i];
			else {
				cp_slash(&cx[i], &cb[i], &cdata[locus]);
			}
		}
		break;
	case SP_DUPLEX:
		zx = zvec(xp);
		zb = zvec(bp);
		zdata = (struct duplex_t *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1)
				zx[i] = zb[i];
			else {
				zp_slash(&zx[i], &zb[i], &zdata[locus]);
			}
		}
		break;
	default:
		fatal("%s(): impossible case type", __F);
	}
	return;
}

/*
 * 2nd Edition
 */

void
precond_simple2nd(struct vector_t *xp, const struct sparse_t *sp,
	struct vector_t *bp)
{
	int	i, locus;

	float	*sx, *sb, *sdata;
	double	*dx, *db, *ddata;
	struct simplex_t *cx, *cb, *cdata;
	struct duplex_t *zx, *zb, *zdata;

	static double limit, dlimit;

	limit = 0.001;
	dlimit = limit;

	switch (sp->sp_type) {
	case SP_SINGLE:
		sx = svec(xp);
		sb = svec(bp);
		sdata = (float *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1 || (fabs(sdata[locus])) < limit)
				sx[i] = sb[i];
			else
				sx[i] = sb[i]/sdata[locus];
		}
		break;
	case SP_DOUBLE:
		dx = dvec(xp);
		db = dvec(bp);
		ddata = (double *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1 || (fabs(ddata[locus])) < limit)
				dx[i] = db[i];
			else
				dx[i] = db[i]/ddata[locus];
		}
		break;
	case SP_SIMPLEX:
		cx = cvec(xp);
		cb = cvec(bp);
		cdata = (struct simplex_t *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1 || (cp_square(&cdata[locus])) < dlimit)
				cx[i] = cb[i];
			else
				cp_slash(&cx[i], &cb[i], &cdata[locus]);
		}
		break;
	case SP_DUPLEX:
		zx = zvec(xp);
		zb = zvec(bp);
		zdata = (struct duplex_t *) sp->sp_data;

		for (i=0;i < xp->vt_size; i++) {
			locus = sp_getwhereis(sp, i, i);
			if (locus == -1 || (zp_square(&zdata[locus])) < dlimit)
				zx[i] = zb[i];
			else
				zp_slash(&zx[i], &zb[i], &zdata[locus]);
		}
		break;
	default:
		fatal("%s(): impossible case type", __F);
	}
	return;
}

struct vector_t *
it_solve(struct sparse_t *sp,struct vector_t *x, struct vector_t *b)
{
	int	no;
	struct vector_t *ret;
	iter_bontriplets(__F, sp, x, b);	/* check */

	no = x->vt_size;

	sp_symcheck(sp);
	switch (sp->sp_type) {
	case SP_SINGLE:
	case SP_DOUBLE:
		ret = it_bicongrad(sp, x, b, no);
		break;
	case SP_SIMPLEX:
	case SP_DUPLEX:
		if (sp->sp_symflag & SP_COMPLEXSYM)
			ret = it_bicongradcomplex(sp, x, b, no);
		else if (sp->sp_symflag & SP_HERMITIAN)
			ret = it_bicongrad(sp, x, b, no);
		else
			fatal("%s(): I can not solve this matrix!!", __F);
		break;
	}
	return ret;
}
